/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include "GCB_errno.h"

// Global data structure
extern int _CB_initialized;
extern struct FDinfo *_CB_fdInfo;
extern int _GCB_max_no_fds;


int
GCB_accept(int fd, struct sockaddr * peer, socklen_t *addrlen)
{
	struct SockInfo *info;
	struct conn *conn;
	struct sockaddr_in *speer;
	int newfd;
	int val, ready;
	fd_set *rdfds = NULL;

	if (!_CB_initialized) { _CB_initialize(); }

    if (fd < 0 || fd >= _GCB_max_no_fds || _CB_fdInfo[fd].type == NOT_DEFINED) {
        LOG(D_FULL, "accept(%d[NOT_DEFINED], ...)\n", fd);
        newfd = accept(fd, peer, addrlen);
        _GCB_errno = (newfd < 0) ? GCB_ESYSCALLFAILED : 0;
        LOG(D_FULL, "GCB_accept() returning %d\n", newfd);
        return newfd;
	}

	if (_CB_fdInfo[fd].type == MANAGER || _CB_fdInfo[fd].type == HIDDEN) {
		LOG(D_FULL, "GCB_accept called with internal GCB socket\n");
		errno = EBADF;
		_GCB_errno = GCB_EINTERNALSOCK;
		return -1;
	}

	info = _CB_fdInfo[fd].info;
	// Do sanity checking
	if (info->status != LISTENING ) {
		LOG(D_BASIC, "GCB_accept() called for an fd not listening");
		_CB_dumpSockInfo(info, D_BASIC, "\t");
        _GCB_errno = GCB_ESOCKINVALIDSTATUS;
		return -1;
	}

	// Check if fd is nonblocking socket
	if ((val = fcntl(fd, F_GETFL, 0)) < 0) {
		ERROR("GCB_accept: fcntl() failed");
        _GCB_errno = GCB_ESYSCALLFAILED;
		return -1;
	}
	LOG(D_FULL, "GCB_accept(%d[GCB_SOCKET at %s] ...)...\n", fd, _ipport2str(info->cip, info->cport));
	while(1) {
		// Return the connection from the head of the queue
		if (info->head != NULL) {
			// get the 1st connection
			conn = info->head;
			info->head = info->head->next;
			if (info->head == NULL) {
				info->tail = NULL;
			}
			newfd = conn->fd;
			_CB_fdInfo[newfd].type = NOT_DEFINED;
			if(peer) {
				speer = (struct sockaddr_in *)peer;
				speer->sin_family = AF_INET;
				speer->sin_addr.s_addr = conn->ip;
				speer->sin_port = conn->port;
				LOG(D_FULL, "\tconnection from %s accepted. newfd = %d returned\n",
					_sin2str(speer), newfd);
			} else {
				LOG(D_FULL, "\tconnection accepted. newfd = %d returned\n",
					newfd);
			}
			free(conn);
            _GCB_errno = 0;
			return newfd;
		}

		// If fd is nonblocking socket, then return EAGIN
		if (val & O_NONBLOCK) {
			errno = EAGAIN;
            LOG(D_FULL, "GCB_accept() returning EAGAIN\n");
            _GCB_errno = 0;
			return -1;
		}

		// Wait until the connection queue becomes nonempty, while handling asynchronous events
		rdfds = _CB_mallocFDSet();
		FD_SET(fd, rdfds);
		ready = _CB_select(fd+1, rdfds, NULL, NULL, NULL);
		_CB_freeFDSet( rdfds );
		if (ready < 0 && errno != EINTR) {
			ERROR("GCB_accept() _CB_select() failed");
            if (!_GCB_errno) {
                _GCB_errno = GCB_ESYSCALLFAILED;
            }
			return -1;
		}
	}	// end of while(1)
}
