/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include "GCB_errno.h"

// Global data structure
extern int _CB_initialized;
extern struct FDinfo *_CB_fdInfo;
extern time_t _CB_lastHeartBeat;
extern int _GCB_max_no_fds;


/* Application can dynamically set socket as ACTIVE or PASSIVE by this function.
   By ACTIVE we mean that the socket always initiates the connection
   no matter whether the user calls "connect" or "listen". PASSIVE is defined
   similarily, but in the opposite way.

   For TCP, setting a socket ACTIVE or PASSIVe takes an effect on connections
   made afterward, i.e, "connect" or "listen" calls made after this function
   is called.

   For UDP, changing a socket to ACTIVE or PASSIVE takes effect immediatly,
   i.e., on "sendto" or "recvfrom" calls made after this function call.

   What if the user never specifies ACTIVE or PASSIVE? Then CB will determine
   which party, or both, of the communication will initiate the connection based
   on who is inside private network. Of course the party which calls "connect"
   will play active role when both parties are on public network and the user
   never calls "active".
 */
// int GCB_active (int fd, char act)
//		fd: file descriptor of the socket. The socket must have been created by
//			"CB_socket" call
//		act: ACTIVE--'A' or PASSIVE--'P'
//		@return: 0 if success, -1 if failed
//		errors:
//			ESOCKTNOSUPPORT - connection brokering is not supported for the type
//							  of this socket
//			EINVAL - invalid argument for "act"
int
GCB_active(int fd, char act)
{
	if (!useGCB()) {
		LOG(D_BASIC, "GCB_active: GCB is not being used");
		_GCB_errno = GCB_ESOCKTNOSUPPORT;
		return -1;
	}
	if (!_CB_initialized) { _CB_initialize(); }

    if (fd < 0 || fd >= _GCB_max_no_fds) {
        LOG(D_BASIC, "GCB_active(%d, %c): fd is invalid. Returning error.\n",
            fd, act);
        errno = EBADF;
        _GCB_errno = GCB_EBADARG;
        return -1;
    }

	if (_CB_fdInfo[fd].type != TCP && _CB_fdInfo[fd].type != UDP) {
		LOG(D_BASIC, "GCB_active: Not my type. Sorry");
		_GCB_errno = GCB_ESOCKTNOSUPPORT;
		return -1;
	}

	//LOG(D_FULL, "[GCB_active]\n");
	if (act == ACTIVE) {
		_CB_fdInfo[fd].info->active = ACTIVE;
	} else if (act == PASSIVE) {
		_CB_fdInfo[fd].info->active = PASSIVE;
	} else {
		LOG(D_BASIC, "GCB_active: Invalid act type");
		_GCB_errno = GCB_EBADARG;
		return -1;
	}

	// If fd is for UDP socket, flush _CB_fdInfo[fd].rcvers
	// so that the next sendto or recvfrom takes appropriate action.
	if (_CB_fdInfo[fd].type == UDP) {
		_CB_delRcvrs(_CB_fdInfo[fd].info);
	}
    _GCB_errno = 0;

	return 0;
}
