/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"

#define UDP_DUMMY 1
#define UDP_DUM_ACK 2
#define UDP_DATA 3

extern int _CB_maxfd;
extern fd_set _CB_mfds;
extern struct FDinfo *_CB_fdInfo;
extern time_t _CB_lastCheck;
extern time_t _CB_lastHeartBeat;
extern struct sockaddr_in _RelayServer;
extern time_t _Broker_down_max;
extern time_t _Broker_down_since;
extern time_t _Broker_down_last_callback;
extern time_t _CB_lastRecovery;
extern void (*_Broker_down_app_callback)(void);
extern void (*_Recovery_failed_app_callback)(void);

/*
  in this file, the places we want to use non-blocking recv() are all
  after we've already done a select().  so, it's just being extra
  paranoid to do it nonblocking, and it's 
*/
#if HAVE_MSG_DONTWAIT
static int msg_nonblock = MSG_DONTWAIT;
#else
static int msg_nonblock = 0;
#endif


// Do the followings according to the message type in "buf":
//	+ DUMMY msg - reply DUMMY_ACK and add the sender to receiver list
//	+ DUMMY_ACK - ignore
//	+ normal data msg - queue the msg overwriting the previously queued one
//
// @args:
//	in:
//		info - the record of the GCB socket which the UDP was delivered to
//		buf - buffer containing the UDP msg
//		nobytes - size of the UDP msg
//		from - sinaddr from which the UDP msg was received. NULL if fd is connected
//		fromlen - the length of "from"
//	out - meaningful only when the msg in 'buf' is either DUMMY or DUMMY_ACK
//		cip - cip of the sender of the msg
//		cport - cport of the sender of the msg
// @return:	UDP_DUMMY if the msg is DUMMY
//			UDP_DUM_ACK if the msg is DUMMY_ACK
//			UDP_DATA otherwise
//			0, if the msg is not for this socket
//			-1, if failed
static int
handleUDPmsg(
		int fd,
		char *buf,
		int nobytes,
		struct sockaddr_in *from,
		socklen_t fromlen,
		uint32_t *cip,
		uint16_t *cport)
{
	struct SockInfo *info;
    struct UDP_msg *msg;
	uint32_t dip;
	uint16_t dport;
	uint16_t size;
	char opcode;
	int rst;
	char str1[30];
	char str2[30];

	info = _CB_fdInfo[fd].info;
    // sanity check for short msgs
    if (nobytes < 3) {
        goto queue_it;
    }

	// unstripe the very first part of the message
	memcpy(&size, buf, 2);
	size = ntohs(size);
	memcpy(&opcode, &buf[2], 1);
    if (size != nobytes) {
        goto queue_it;
    }

	// if DUMMY or DUMMY_ACK received
    if (opcode == DUMMY) {
        if (nobytes != 15) {
            LOG(D_FULL, "handleUDPmsg() DUMMY msg is not 15 bytes. "
				"Take it as data packet\n");
            goto queue_it;
        }
        // DUMMY is meaningful only for unconnected UDP
        if (info->peer) {
            LOG(D_FULL, "handleUDPmsg() DUMMY received for connected UDP\n");
            return 0;
        }
        // Reply DUMMY_ACK
        memcpy(cip, &buf[3], 4);
        memcpy(cport, &buf[7], 2);
        memcpy(&dip, &buf[9], 4);
        memcpy(&dport, &buf[13], 2);
        sprintf(str1, "%s", _ipport2str(*cip, *cport));
        sprintf(str2, "%s", _ipport2str(dip, dport));
        LOG(D_FULL, "*recv* [DUMMY: %d] %s -> %s\n", fd, str1, str2);
        if (dip != info->cip || dport != info->cport) {
            LOG(D_FULL, "thandleUDPmsg() DUMMY to another guy\n");
            return 0;
        }
        size = _CB_stripeMsg(buf, DUMMY_ACK, dip, dport, 0, 0, 0, *cip, *cport, 0);
        if (size == 0) {
            LOG(D_BASIC, "handleUDPmsg() received [DUMMY: %d] %s -> %s\n",
				fd, str1, str2);
            LOG(D_BASIC, "handleUDPmsg(): _CB_stripMsg(DUMMY_ACK) failed\n");
            return -1;
        }
        rst = sendto(fd, buf, size, 0, (struct sockaddr *)from, fromlen);
        if (rst < 0) {
            ERROR("handleUDPmsg: received [DUMMY: %d] %s -> %s, "
				  "sendto(DUMMY_ACK) failed", fd, str1, str2);
            return -1;
        }
        // Add the sender to the receiver list
        if (_CB_addRcvr(info, *cip, *cport, from->sin_addr.s_addr, from->sin_port) < 0) {
            ERROR("handleUDPmsg(): _CB_addRcvr() failed: "
				  "Can't add the receiver");
            return -1;
        }
        return UDP_DUMMY;
    } else if (opcode == DUMMY_ACK) {
        if (nobytes != 15) {
            LOG(D_FULL, "handleUDPmsg(): DUMMY_ACK msg is not 15 bytes. "
				"Take it as data packet\n");
            goto queue_it;
        }
        memcpy(cip, &buf[3], 4);
        memcpy(cport, &buf[7], 2);
        memcpy(&dip, &buf[9], 4);
        memcpy(&dport, &buf[13], 2);
        sprintf(str1, "%s", _ipport2str(*cip, *cport));
        sprintf(str2, "%s", _ipport2str(dip, dport));
        LOG(D_FULL, "handleUDPmsg() *recv* [DUMMY_ACK: %d] %s -> %s\n",
			fd, str1, str2);
        return UDP_DUM_ACK;
    } else {
        LOG(D_FULL, "handleUDPmsg() size field of the msg matches the "
			"size of msg, but opcode is neither DUMMY nor DUMMY_ACK. "
			"Taking it as DATA\n");
    }

queue_it:
	// Normal UDP data packet received. Queue the packet unless rcv buffer is full
    if (info->nomsgs >= GCB_MAX_UDP_MSGS) {
        LOG(D_BASIC, "Warning!!! "
            "UDP msg just received was dropped b/c rcv buf is full\n");
        return UDP_DATA;
    }
    msg = (struct UDP_msg *) malloc(sizeof(struct UDP_msg));
    if (!msg) { ABORT("failed to malloc"); }
    msg->next = NULL;
    msg->bytes = nobytes;
    msg->data = (char *) malloc (nobytes);
    if (!msg->data) { ABORT("failed to malloc"); }
	memcpy(msg->data, buf, nobytes);
    if (info->rbuf_tail) {
        info->rbuf_tail->next = msg;
        info->rbuf_tail = msg;
    } else {
        info->rbuf_head = info->rbuf_tail = msg;
    }
    (info->nomsgs)++;
	if (from) {
        memcpy(&(msg->sender), from, sizeof(*from));
		LOG(D_FULL, "handleUDPmsg() [UDP packet to %s] Data(%d) received "
			"from %s\n", _ipport2str(info->cip, info->cport), nobytes,
			_sin2str(from));
	} else {
		LOG(D_FULL, "handleUDPmsg() [UDP packet to %s] Data(%d) received\n",
			_ipport2str(info->cip, info->cport), nobytes);
	}
	return UDP_DATA;
}

// Handle CONTACT msg from Broker. Note that this type of msg must be delivered
// to socket which is in state of LISTENING. Actions:
//	+ creat a new socket
//	+ connect to given ip:port
//	+ queue the connection to _CB_fdInfo[fd].info->tail
// @args
//	in:
//		- sock: pointer to the SockInfo of the CB socket
//		- ip, port: ip and port field of the msg received via mng sock
// @ret: 0 if succ, -1 if fail
static int
handleContact(struct SockInfo *sock, uint32_t ip, uint16_t port)
{
	int newfd;
	int result = -1;
	struct sockaddr_in sinAddr;
	struct conn *conn;

	if (sock->status != LISTENING) {
		ERROR( "handleContact: status is not LISTENING, "	
			   "but CONTACT received");
		result = -1;
		goto remote_log;
	}

	newfd = socket(AF_INET, SOCK_STREAM, 0);
	if (newfd <0) {
		ERROR("handleContact: socket()");
		result = -1;
		goto remote_log;
	}

	bzero(&sinAddr, sizeof(sinAddr));
	sinAddr.sin_family = AF_INET;
	sinAddr.sin_addr.s_addr = ip;
	sinAddr.sin_port = port;
	if (_CB_do_connect(newfd, (struct sockaddr *) &sinAddr, sizeof(sinAddr), NULL)) {
		ERROR("handleContact: connection to %s failed", _sin2str(&sinAddr));
		close(newfd);
		result = -1;
		goto remote_log;
	}
    // queue the connection
	_CB_fdInfo[newfd].type = HIDDEN;
	conn = (struct conn *)calloc(sizeof(struct conn), 1);
	conn->fd = newfd;
	conn->ip = ip;
	conn->port = port;
	if (sock->tail) {
		sock->tail->next = conn;
	} else {
		sock->head = conn;
	}
	sock->tail = conn;

	LOG(D_FULL, "handleContact: [connection accepted by CONTACT] %s <- %s\n",
			_ipport2str(sock->cip, sock->cport), _sin2str(&sinAddr));
	result = 0;

remote_log:
	if (_CBsend(sock->manager, REMOTE_REPORT, ip, port, REMOTE_REPORT_PASSIVE, result?1:0, 0, 0, 0, 0) < 0) {
		ERROR( "handleContact: _CBsend(REMOTE_REPORT) failed");
		/* Sending the log message is non-essential, so we don't return an
		 * error here */
	}
	return result;
}

// Handle TRIGGER msg from Broker. Note that this type of msg must be delivered
// to UDP. Procedure:
//	+ send DUMMY msg to every ip:port
//	+ recv DUMMY_ACK
// @args
//	in:
//		- sock: pointer to the SockInfo of the CB socket
//		- dip, dport: cip:cport of the peer
//		- ip, port: physical ip:port of the peer
// @ret: 0 if succ, -1 if failed
static int
handleTrigger(int fd, uint32_t dip, uint16_t dport, uint32_t ip, uint16_t port)
{
	uint32_t pcip;
	uint16_t pcport;
	char buf[65536], tstr[30];
	struct sockaddr_in peer;
	int size, tries, nobytes, maxfd, retval;
	struct timeval tv;
	fd_set *rdfds = _CB_mallocFDSet();
	int failed = 0;
	socklen_t addrlen = sizeof(struct sockaddr_in);
	struct SockInfo *sock = _CB_fdInfo[fd].info;

	// Since unconnecting UDP socket in Linux is not stable yet, we just comment out
	// the part that connects the socket to the peer and unconnects it again
	/*
	// connect the socket to the peer
	bzero(&peer, sizeof(struct sockaddr_in));
	peer.sin_family = AF_INET;
	peer.sin_addr.s_addr = ip;
	peer.sin_port = port;
	if (connect(fd, (struct sockaddr *)&peer, sizeof(struct sockaddr_in)) < 0) {
		ERROR("handleTrigger: connect() failed");
		return -1;
	}
	*/
	tries = 0;
	while(tries++ < GCB_MAX_TRY) {
		// send dummy msg to the peer
		size = _CB_stripeMsg(buf, DUMMY, sock->cip, sock->cport, 0, 0, 0, dip, dport, 0);
		if (size < 0) {
			ERROR("handleTrigger: _CB_stripMsg(DUMMY) failed");
			failed = 1;
			goto done;
		}
		// DUMMY should be sent by raw udp
		bzero(&peer, sizeof(struct sockaddr_in));
		peer.sin_family = AF_INET;
		peer.sin_addr.s_addr = ip;
		peer.sin_port = port;
		if (sendto(fd, buf, size, 0, (struct sockaddr *)&peer, sizeof(peer)) < 0) {
			ERROR("handleTrigger: sendto() failed");
		}
		// get DUMMY_ACK from the peer
		while(1) {
			_CB_FD_ZERO(rdfds);
			FD_SET(fd, rdfds);
			maxfd = fd + 1;
			tv.tv_sec = GCB_TRY_INTERVAL;
			tv.tv_usec = 0;
			retval = select(fd + 1, rdfds, NULL, NULL, &tv);
			if (retval < 0) {
				if (errno == EINTR)
					continue;
				else {
					ERROR("handleTrigger(%s): DUMMY sent, but "
						  "select() for DUMMY_ACK failed", 
						  _ipport2str(sock->cip, sock->cport));
					failed = 1;
					goto done;
				}
			} else if (retval == 0) {
				// Timeout! Send DUMMY again
				break;
			}

			// Receive a msg
			nobytes = recvfrom(fd,  buf, 65536, msg_nonblock,
							   (struct sockaddr *)&peer, &addrlen);
			if (nobytes < 0) {
				ERROR("handleTrigger: recvfrom() failed");
				failed = 1;
				goto done;
			}

			// If DUMMY_ACK msg received
			if (handleUDPmsg(fd, buf, nobytes, &peer, addrlen, &pcip, &pcport) == UDP_DUM_ACK) {
				// Check if the DUMMY_ACK is from the right guy
				if (dip == pcip && dport == pcport) {
					strcpy(tstr, _ipport2str(dip, dport));
					LOG(D_FULL, "handleTrigger: UDP channel opened by TRIGGER "
						"%s <-> %s Phys:%s\n", 
						_ipport2str(sock->cip, sock->cport), tstr,
						_sin2str(&peer));
					if (_CB_addRcvr(sock, dip, dport, ip, port) < 0) {
						ERROR("handleTrigger: _CB_addRcvr failed");
						failed = 1;
						goto done;
					}
					goto done;
				}
			}
		} // while(1)
	} // while(tries < 3)
		// should this be considered an error?
	LOG(D_FULL, "handleTrigger: could not get DUMMY_ACK\n");

done:
	/*
	// disconnect fd again
	peer.sin_family = AF_UNSPEC;
	peer.sin_port = 0;
	peer.sin_addr.s_addr = 0;
	if (connect(fd, (struct sockaddr *)&peer, sizeof(peer)) < 0) {
		ERROR("handleTrigger: connect() failed");
		return -1;
	}
	*/
	_CB_freeFDSet(rdfds);
	if (failed) {
		return -1;
	}
	return 0;
}

static int
rebind(struct SockInfo *info)
{
	int mfd;
	char opcode, errcode, active;
	uint32_t cip, dip, ip;
	uint16_t cport, dport, port;

    if (info->manager != NEED_RECOVER) {
		ERROR("rebind: info->manager is not NEED_RECOVER\n");
        return -1;
    }

	// create a new manager
	if ((mfd = _CB_newManager(info)) < 0) {
		ERROR("rebind: _CB_newManager() failed\n");
        info->manager = NEED_RECOVER;
		return -1;
	}

	// if UDP socket, delete mappings (cip:cport <-> ip:port)
	if (_CB_fdInfo[info->owner].type == UDP) {
		_CB_delRcvrs(info);
		opcode = UDP_BIND;
	} else {
		opcode = TCP_BIND;
	}

	// send BIND request
	if (_CBsend(mfd, opcode, info->cip, info->cport, info->active,
				info->ip, info->port, 0, 0, 0) < 0) {
		ERROR("rebind: _CBsend(BIND) failed\n");
		_CB_rmManager(mfd);
        info->manager = NEED_RECOVER;
		return -1;
	}

	// receive the result
	if (_CBrecv(mfd, &opcode, &cip, &cport, &active, &ip, &port, &dip, &dport, &errcode) < 0) {
		ERROR("rebind: BIND req. sent. Could not get the result\n");
		_CB_rmManager(mfd);
        info->manager = NEED_RECOVER;
		return -1;
	}
	if (opcode == BIND_OK) {
		return 0;
	} else {
        ERROR("rebind: Unexpected msg received. Not going to retry again\n");
		_CB_rmManager(mfd);
		return -2;
	}
}

static int
relisten(struct SockInfo *info)
{
	char opcode, errcode, active;
	uint32_t cip, dip, ip;
	uint16_t cport, dport, port;
	int received;

    // this routine is always called after rebind. we should not check if
    // info->manager == NEED_RECOVER
	if (_CBsend(info->manager, LISTEN, info->cip, info->cport, 0, 0, 0, 0, 0, 0) < 0) {
		ERROR( "relisten: _CBsend(LISTEN) failed");
		_CB_rmManager(info->manager);
        info->manager = NEED_RECOVER;
		return -1;
	}
	// Get the result from the Broker
	received = _CBrecv(info->manager, &opcode, &cip, &cport,
					   &active, &ip, &port, &dip, &dport, &errcode);
	if (received <= 0) {
		ERROR("relisten: LISTEN sent, but _CBCrecv failed");
        _CB_rmManager(info->manager);
        info->manager = NEED_RECOVER;
		return -1;
	}

	if (opcode == ACK) {
		return 0;
	} else {
        ERROR("Unexpected msg received. Not going to retry again");
        _CB_rmManager(info->manager);
		return -2;
	}
}

// Since any communication error with RelayServer often means the server crash,
// we need to recover every GCB sockets registed to the RelayServer
void
_CB_recoverGCBsockets()
{
	struct SockInfo *info;
	int i;
	int rc = 0;

	// Recover status of GCB sockets
	for (i =0; i < _CB_maxfd; i++) { 
		info = _CB_fdInfo[i].info;
        if (!info || info->manager != NEED_RECOVER) continue;

		switch(info->status) {
			case BOUND:
				if ((rc = rebind(info)) < 0) {
					ERROR("_CB_recoverGCBsockets() failed - bailing out");
                    // Any failure of recovery attempt by a socket generally means that 
                    // Broker has not been brought up yet and all the other sockets
                    // will fail too. So we give up at the moment.
                    goto report;
				}
				break;
			case LISTENING: // this must be TCP socket
				if ((rc = rebind(info)) < 0) {
					ERROR( "_CB_recoverGCBsockets() failed - bailing out");
                    // Any failure of recovery attempt by a socket generally means that 
                    // Broker has not been brought up yet and all the other sockets
                    // will fail too. So we give up at the moment.
                    goto report;
				}
				if ((rc = relisten(info)) < 0) {
					ERROR( "_CB_recoverGCBsockets() failed - bailing out");
                    // Any failure of recovery attempt by a socket generally means that 
                    // Broker has not been brought up yet and all the other sockets
                    // will fail too. So we give up at the moment.
                    goto report;
				}
				break;
			case CONNECTED: // this must be UDP socket
				if (rebind(info) < 0) {
					ERROR( "_CB_recoverGCBsockets() failed - bailing out");
                    // Any failure of recovery attempt by a socket generally means that 
                    // Broker has not been brought up yet and all the other sockets
                    // will fail too. So we give up at the moment.
                    goto report;
				}
				break;
			default:
				// NONE, CREATED, CONN_FAILED
				continue;
		}
	}
    // every socket has been recovered
    _Broker_down_since = _CB_lastRecovery = 0;
	_Broker_down_last_callback = 0;
    return;

report:
    _CB_lastRecovery = time(NULL);
    if (_Broker_down_app_callback) {
		if (_Broker_down_last_callback ?
			time(NULL) - _Broker_down_last_callback >= _Broker_down_max :
			time(NULL) - _Broker_down_since >= _Broker_down_max) {

            LOG(D_BASIC, "Haven't recovered from Broker failure after %d "
				"seconds, calling application's registered handler\n", 
				time(NULL) - _Broker_down_since);
            _Broker_down_app_callback();
			_Broker_down_last_callback = time(NULL);
        }
    }
	if (_Recovery_failed_app_callback && rc == -2 ) {
		LOG(D_BASIC, "Can't recover from Broker failure, "
			"calling application's registered handler\n");
		_Recovery_failed_app_callback();
	}
    return;
}

static int
handleGCBcommand (int mfd)
{
	int rst;
	struct SockInfo *info;
	char opcode, errcode;
	uint32_t cip, ip, dip;
	uint16_t cport, port, dport;
	char active;

	info = _CB_fdInfo[mfd].info;
	rst = _CBrecv(mfd, &opcode, &cip, &cport, &active, &ip, &port, &dip, &dport, &errcode);
	if ( rst <= 0 ) {
		ERROR("handleGCBcommand(%s): _CBrecv failed", 
			  _ipport2str(info->cip, info->cport)); 
        if (_Broker_down_since == 0) {
            _Broker_down_since = time(NULL);
        }
        _CB_rmManager(mfd);
        info->manager = NEED_RECOVER;
		return -1;
	}

	switch(opcode) {
		case BIND_OK:
		case ACK:
		case NAK:
		case ACTIVE_CONNECT:
		case PASSIVE_CONNECT:
		case ACTIVE_INIT:
		case PASSIVE_INIT:
			LOG(D_FULL, "handleGCBcommand(%s): Delayed synchronous msg "
				"received. Ignored\n", _ipport2str(info->cip, info->cport));
			return 0;

		case CONTACT:
			rst = handleContact(info, ip, port);
			return rst;

		case TRIGGER:
			if (info->peer) {
				LOG(D_FULL, "handleGCBcommand: TRIGGER received for "
					"connected UDP. Ignored!\n");
				return 0;
			}
			rst = handleTrigger(info->owner, cip, cport, ip, port);
			return rst;

		default:
			ERROR("handleGCBcommand(%s): Unknown type of message received",
				  _ipport2str(info->cip, info->cport));
			return -1;
	} // end of switch
}

// Add fds to given fd_sets so that caller can do select with augmented fd_set.
// The following set of fds are added to refd:
//  + MANAGER sockets
//  + UDP sockets
//  + listening TCP sockets
//  + blocked TCP connections
// Also remove sockets in PASSIVE_CONNECTING state from the write and
// except fdsets.
// @args:
//    + in & out: rdfds, wrfds, exfds
void
_CB_addFDs(int *maxfd, fd_set *rdfds, fd_set *wrfds, fd_set *exfds)
{
	int i;
	//char str[1000];

	//str[0] = '\0';
	for (i = 0; i < _CB_maxfd; i++) {
		if ((_CB_fdInfo[i].type == MANAGER) ||	// MANAGER
			(_CB_fdInfo[i].type == UDP && i == _CB_fdInfo[i].info->owner) || // UDP socket
			(_CB_fdInfo[i].type == TCP && i == _CB_fdInfo[i].info->owner &&
			 (_CB_fdInfo[i].info->status == LISTENING ||	// LISTENING TCP
			  _CB_fdInfo[i].info->status == PASSIVE_CONNECTING) // PASSIVE_CONNECTING TCP
			)
		) {
			if(rdfds) { FD_SET(i, rdfds); }
			if (_CB_fdInfo[i].info->status == PASSIVE_CONNECTING) {
				if(wrfds) { FD_CLR(i, wrfds); }
				if(exfds) { FD_CLR(i, exfds); }
			}
			//sprintf(&str[strlen(str)], "%d ", i);
			if (*maxfd <= i) {
				*maxfd = i + 1;
			}
		} else if (_CB_fdInfo[i].type == TCP && // TCP
				   i == _CB_fdInfo[i].info->owner && // onwer
				   _CB_fdInfo[i].info->status == ACTIVE_CONNECTING // PASSIVE_CONNECTING TCP)
		) {
			if(wrfds) { FD_SET(i, wrfds); } 
			if (*maxfd <= i) {
				*maxfd = i + 1;
			}
		}
	}
	/* if (str[0] != '\0') {
		LOG(D_FULL, "\tChecking UDP sockets: %s\n", str);
	} */
	return;
}

static int
handleFDasyncMsg(int fd)
{
	char buf[65536];
	struct sockaddr_in from;
	struct SockInfo *info;
	int nobytes;
    socklen_t fromlen;
	uint32_t cip;
	uint16_t cport;

	// receive a msg
	info = _CB_fdInfo[fd].info;
	if (info->peer) {
		nobytes = recv(fd,  buf, sizeof(buf), msg_nonblock);
	} else {
		fromlen = sizeof(from);
		nobytes = recvfrom(fd,  buf, sizeof(buf), msg_nonblock,
						   (struct sockaddr *)&from, &fromlen);
	}
	if (nobytes < 0) {
		ERROR("handleFDasyncMsg: fd(%d) for %s is ready, but recvfrom() "
			  "failed", fd, _ipport2str(info->cip, info->cport));
		return -1;
	}

	if (info->peer) {
		(void) handleUDPmsg(fd, buf, nobytes, info->peer, sizeof(struct sockaddr_in), &cip, &cport);
	} else {
		(void) handleUDPmsg(fd, buf, nobytes, &from, fromlen, &cip, &cport);
	}
	return 0;
}

int
_CB_handleDummy(int fd, uint32_t pcip, uint16_t pcport, struct sockaddr_in *phyTo)
{
	fd_set *rdfds = _CB_mallocFDSet();
	struct timeval tv;
	struct SockInfo *info = _CB_fdInfo[fd].info; 
	int ready;
	struct sockaddr_in from;
	int fromlen, nobytes;
	char buf[65536];
	uint32_t cip;
	uint16_t cport;
	int rst;
    time_t before, after;
	int result = 0;

	// this function should be called only for unconnected UDP socket
	if (_CB_fdInfo[fd].type != UDP || info->status == CONNECTED) {
		ERROR("_CB_handleDummy called for connected UDP");
        return -1;
	}

    before = after = time(NULL);
	while(1) {
		_CB_FD_ZERO(rdfds);
		FD_SET(fd, rdfds);
		tv.tv_sec = GCB_MAX_TRY * GCB_TRY_INTERVAL - (after - before);
		tv.tv_usec = 0;
		ready = select(fd+1, rdfds, NULL, NULL, &tv);
        after = time(NULL);
		if (ready < 0) {
			if (errno == EINTR) {
				continue;
			}
			ERROR("_CB_handleDummy: select() failed");
			result = -1;
			break;
		} else if (!ready) {
				// why is this an error?
			ERROR("_CB_handleDummy: select() timed-out");
			result = -1;
			break;
		}
		// receive a msg
		fromlen = sizeof(from);
		nobytes = recvfrom(fd,  buf, sizeof(buf), msg_nonblock,
						   (struct sockaddr *)&from, &fromlen);
		if (nobytes < 0) {
			if (errno == EAGAIN) {
				continue;
			}
			ERROR("_CB_handleDummy: recvfrom failed");
			result = -1;
			break;
		}

		// If DUMMY msg received
		rst = handleUDPmsg(fd, buf, nobytes, &from, fromlen, &cip, &cport);
		if (rst == UDP_DUMMY) {
			// Check if the DUMMY is from the peer we want to talk
			if (pcip == cip && pcport == cport) {
				memcpy(phyTo, &from , sizeof(from));
				result = 0;
				break;
			}
		} // if DUMMY msg
	} // while(1)

	_CB_freeFDSet( rdfds );
	return result;
}

static void
handleListener(struct SockInfo *info)
{
	struct sockaddr_in who;
	int newfd;
	socklen_t addrLen;
	struct conn * conn;

	// accept the incoming connection
	addrLen = sizeof(who);
	newfd = accept(info->owner, (struct sockaddr *) &who, &addrLen);
	if (newfd < 0) {
		ERROR("handleListeners: fd(%d) for %s is ready, but accept() failed",
			  info->owner, _ipport2str(info->cip, info->cport));
		return;
	} else {
		// queue up the established connection
		_CB_fdInfo[newfd].type = HIDDEN;
		conn = (struct conn *)calloc(sizeof(struct conn), 1);
		conn->fd = newfd;
		conn->ip = who.sin_addr.s_addr;
		conn->port = who.sin_port;
		if (info->tail) {
			info->tail->next = conn;
		} else {
			info->head = conn;
		}
		info->tail = conn;
		LOG(D_FULL, "handleListeners: %s(fd=%d) connection accepted from: "
			"%s(newfd=%d)\n", _ipport2str(info->cip, info->cport), 
			info->owner, _sin2str(&who), newfd);
	}
}

static void
handlePassiveBlkedConn(struct SockInfo *info)
{
	int newfd;
	struct sockaddr_in who;
	socklen_t len;

	// accept the incoming connection
	len = sizeof(who);
	newfd = accept(info->owner, (struct sockaddr *) &who, &len);
	if (newfd < 0) {
		if (errno == EAGAIN) {
			return;
		}
		ERROR("handlePassiveBlkedConn(%d): accept failed", info->owner);
		info->status = CONN_FAILED;
		info->err = errno;
		return;
	} else {
		// replace fd with newfd
		if (_CB_replaceFD(info->owner, newfd) < 0) {
			ERROR("handlePassiveBlkedConn(%d): connection accepted from %s"
				  "but _CB_replaceFD failed", info->owner, _sin2str(&who));
			info->status = CONN_FAILED;
			info->err = errno;
			close(newfd);
			return;
		} else {
			info->status = CONNECTED;
			LOG(D_FULL, "handleBlkConns: %s connection established from: %s\n",
				_ipport2str(info->cip, info->cport), _sin2str(&who));
			return;
		}
	}
}

static void
handleActiveBlkedConn(struct SockInfo *info)
{
	int error;
	socklen_t len = sizeof(error);

	// do getsockopt
	if (getsockopt(info->owner, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
		info->status = CONN_FAILED;
        info->err = errno;
		ERROR("handleActiveBlkedConn(%d): setting status=CONN_FAILED: "
			  "getsockopt() failed", info->owner);
		return;
	}

	if (error) {
        if (error == EINPROGRESS) return;
		// if error, status = CONN_FAILED
		info->status = CONN_FAILED;
        info->err = error;
		// TODO Should set errno=0 here, since it's not related to the
		//   message we're reporting and we don't want it printed out
		ERROR("handleActiveBlkedConn(%d): setting status=CONN_FAILED, "
			  "error %d %s", info->owner, error, strerror(error));
	} else {
		// if no error status = CONNECTED
		info->status = CONNECTED;
        info->err = 0;
		LOG(D_FULL, "handleActiveBlkedConn(%d): no error, "
			"status = CONNECTED\n", info->owner);
	}
	return;
}

void
_CB_handleAsyncEvent(int maxfd, fd_set *rdfds, fd_set *wrfds)
{
	int i;
	struct SockInfo *info;
	char str1[50], str2[50];

	sprintf(str1, "%s", _prtfds(maxfd,rdfds));
	sprintf(str2, "%s", _prtfds(maxfd,wrfds));
	LOG(D_FULL, "_CB_handleAsyncEvent(rd=%s, wr=%s)\n", str1, str2);
	for (i = 0; i < maxfd; i++) {
		if (!FD_ISSET(i, rdfds)) {
			if (_CB_fdInfo[i].type == TCP && _CB_fdInfo[i].info->status == PASSIVE_CONNECTING) {
				if (time(NULL) - _CB_fdInfo[i].info->blkedAt > GCB_CONN_TIMEOUT) {
					_CB_fdInfo[i].info->status = CONN_FAILED;
					_CB_fdInfo[i].info->err = ETIMEDOUT;
				}
			}
			if (!FD_ISSET(i, wrfds)) {
				continue;
			}
		}
		switch(_CB_fdInfo[i].type) {
			case MANAGER: // handle GCB command
				if (handleGCBcommand(i) < 0) {
					ERROR("_CB_handleAsyncEvent: "
						  "handleGCBcommand(%d) failed", i);
                    FD_CLR(i, rdfds);
                    FD_CLR(i, wrfds);
				}
				break;
			case TCP:
				info = _CB_fdInfo[i].info;
				if (i != info->owner) { break; }
				if (info->status == LISTENING) {
					handleListener(info);
				} else if (info->status == PASSIVE_CONNECTING) {
					handlePassiveBlkedConn(info);
				} else if (info->status == ACTIVE_CONNECTING) {
					handleActiveBlkedConn(info);
				}
				break;
			case UDP:
				info = _CB_fdInfo[i].info;
				if (i != info->owner) { break; }
				if (handleFDasyncMsg(i) < 0) {
					ERROR("_CB_handleAsyncEvent: "
						  "handleFDasyncMsg(%d) failed", i);
				}
                break;
			default:
                break;
		}
	}
	/*if (_Broker_down_since) {
		_CB_recoverGCBsockets();
	}*/
}

int
_CB_checkAsyncEvent()
{
	int maxfd, ready;
	fd_set *rdfds = _CB_mallocFDSet();
	fd_set *wrfds = _CB_mallocFDSet();
	struct timeval tv;
	time_t current;
	int result = 0;

	// trigger heartbeat
	current = time(NULL);
	if (current - _CB_lastCheck < GCB_KEEP_EYE_RATE) {
		result = 0;
		goto done;
    }
    if (_Broker_down_since) {
        _CB_recoverGCBsockets();
	}
    if (current - _CB_lastHeartBeat > GCB_HEARTBEAT_RATE) {
		_CB_triggerHeartBeat(current);
		_CB_lastHeartBeat = current;
	}

	// add mng fds, listeners, blocked connections, and UDP
	_CB_FD_ZERO( rdfds);
	_CB_FD_ZERO( wrfds );
	maxfd = 0;
	_CB_addFDs(&maxfd, rdfds, wrfds, NULL);

	// do nonblocking select
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	ready = select(maxfd, rdfds, wrfds, NULL, &tv);
	if (ready < 0) {
		ERROR("_CB_checkAsyncEvent: select() failed");
		result = -1;
		goto done;
	} else if (ready == 0) {
		_CB_lastCheck = time(NULL);
		result = 0;
		goto done;
	}

	// handle GCB commands
	_CB_handleAsyncEvent(maxfd, rdfds, wrfds);
	_CB_lastCheck = time(NULL);
	result = 0;

 done:
	_CB_freeFDSet( rdfds );
	_CB_freeFDSet( wrfds );
	return result;
}
