/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include "GCB_errno.h"

// Global data structure
extern int _CB_initialized;
extern struct FDinfo *_CB_fdInfo;
extern int _GCB_max_no_fds;

// Function prototypes for internal helper methods.
int _GCB_post_bind_save_info(int fd, struct SockInfo *info,
							 const char* func_name, int is_local);
int _GCB_pre_bind_init_info(int fd, struct sockaddr *my_addr,
							struct SockInfo *info, const char* func_name);


/* Register socket to broker */
//  + get RelayServer addr and connect mfd to the RelayServer
//	+ send (TCP | UDP)_BIND to the Broker
//	+ get cip:cport */
//	+ add the socket information to hash bucket: _CB_sockInfo[]
static int
registerSock(int fd, uint16_t port)
{
    // Note: to let the application be able to distinguish the error caused by the
    // Broker not running from other errors, this routine sets errno to EAGAIN for the
    // former cases
	char opcode, errcode, active;
	uint32_t cip, dip, rip;
	uint16_t cport, dport, rport;
	struct SockInfo *sock = _CB_fdInfo[fd].info;
	int mfd;
    int rcved;

	// create a manager
	if ((mfd = _CB_newManager(sock)) < 0) {
		LOG(D_BASIC, "Failed to establish a socket to the broker. (registerSock: _CB_newManager() failed)\n");
		return -1;
	}

	// send BIND to RelayServer
	opcode = (_CB_fdInfo[fd].type == TCP) ? TCP_BIND : UDP_BIND;
	if (_CBsend(mfd, opcode, 0, port, sock->active, sock->ip, sock->port, 0, 0, 0) < 0) {
		LOG(D_BASIC, "Failed to request bind() via broker. (GCB_bind: _CBsend(BIND) failed)\n");
        _GCB_errno = (errno == EPIPE) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
		_CB_rmManager(mfd);
		return -1;
	}
	// receive the result
    rcved = _CBrecv(mfd, &opcode, &cip, &cport, &active, &rip, &rport, &dip, &dport, &errcode);
    if (rcved <= 0) {
		LOG(D_BASIC, "Broker %s didn't respond to bind() request. Broker may not have replied, or may have given an invalid reply. (GCB_bind: _CBrecv failed)\n", _sin2str(getBroker()));
        _GCB_errno = (rcved == 0) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
		_CB_rmManager(mfd);
		return -1;
	}
	if (opcode == BIND_OK) {
		sock->cip = cip;
		sock->cport = cport;
		sock->status = BOUND;
		LOG(D_FULL, "\tOfficial: %s\n", _ipport2str(cip, cport));
		return 0;
	} else if (opcode == NAK) {
		LOG(D_BASIC, "Broker failed bind() attempt: %s (GCB_bind: _CBrecv returned NAK)\n", strerror(errcode)); 
		errno = errcode;
        _GCB_errno = GCB_ESYSCALLFAILED;
		_CB_rmManager(mfd);
		return -1;
	} else {
			// TODO: export printOP from ex_cmd.c and show opcode.
		LOG(D_BASIC, "Broker returns unexpected reponse to request for bind(). "
			"(GCB_bind: _CBrecv returned %d)\n", (int)opcode);
		_CB_rmManager(mfd);
        _GCB_errno = GCB_EBADMSGTYPE;
		return -1;
	}
}

/** Undo registerSock

For whatever reason we've decided to abandon a sock we've
registered with registerSock.  Try to undo any state
changes it might have done.
*/
static void
unregisterSock(struct SockInfo * info)
{
	if( ! info ) {
		LOG(D_BASIC, "unregisterSock(0): Called with null SockInfo.  "
				"Harmless, but suspicious\n");
		return;
	}
	if(info->manager) {
		_CB_rmManager(info->manager);
	}
	info->manager = NO_MNG;
	info->cip = 0;
	info->cport = 0;
	info->status = CREATED;
}

/* Bind the socket */
// #include <sys/types.h>
// #include <sys/socket.h>
// @args: same as "bind"
// @return: same as "bind"
// @errors: in addition to errors which are returned by "bind", error codes
//			from "socket" could be returned
int
GCB_bind(int fd, struct sockaddr *my_addr, socklen_t addrlen)
{
    int result;
	struct sockaddr_in tweakedAddr, *sinAddr;
	struct SockInfo *info;
	int registered_sock = 0;
	int num_collisions = 0;
	int MAX_COLLISIONS = 20; /* years of careful research says... 20 */

	if (!_CB_initialized) { _CB_initialize(); }

	if (fd < 0 || fd >= _GCB_max_no_fds || _CB_fdInfo[fd].type == NOT_DEFINED) {
        result = bind(fd, my_addr, addrlen);
        LOG(D_FULL, "GCB_bind(%d, %p, %d): Unknown file descriptor %d passed directly to bind() which returned %d\n", fd, my_addr, addrlen, fd, result);
        _GCB_errno = (result < 0) ? GCB_ESYSCALLFAILED : 0;
        return result;
	}

	// socket should be registered as TCP or UDP socket
	if (_CB_fdInfo[fd].type == MANAGER || _CB_fdInfo[fd].type == HIDDEN) {
		LOG(D_BASIC, "GCB_bind(%d, %p, %d): WARNING: A GCB internal file descriptor was passed to GCB_bind. This descriptor should never be visible to non-GCB code.\n", fd, my_addr, (int)addrlen);
        _GCB_errno = GCB_EINTERNALSOCK;
		return -1;
	}

	info = _CB_fdInfo[fd].info;
	sinAddr = (struct sockaddr_in *)my_addr;
	LOG(D_FULL, "GCB_bind(%d[GCB_SOCKET], %p->%s, %d)\n", fd, my_addr, _sin2str(sinAddr), addrlen);
	// Check if the socket has already been bound
	if (info->status != CREATED) {
		LOG(D_BASIC, "GCB_bind(%d[GCB_SOCKET], %p->%s, %d): WARNING: This socket is not in the CREATED state. It may have already been bound using GCB_bind.\n", fd, my_addr, _sin2str(sinAddr), addrlen);
		_CB_dumpSockInfo(info, D_BASIC, "  ");
		_GCB_errno = GCB_ESOCKINVALIDSTATUS;
		errno = EINVAL; // Matches Linux/glibc behavior. circa 2006.
		return -1;
	}

	if(info->port) {
		LOG(D_BASIC, "GCB_bind(%d[GCB_SOCKET], %p->%s, %d): Internal consistency error.  Socket has not been bound, but has a port (%d) associated with it.\n", fd, my_addr, _sin2str(sinAddr), addrlen, (int)(info->port));
           _GCB_errno = GCB_EMISC;
		return -1;
	}

		// We want to initialize the local IP address for this socket
		// *before* we call registerSock() to register it at the
		// broker, since the IP is sent to the broker and used for log
		// messages, etc.
	if (_GCB_pre_bind_init_info(fd, my_addr, info, "GCB_bind") < 0) {
		return -1;
	}

	/*
	To minimize surprise to users, we want the same port number
	on both sides.  This means both the broker and the client
	need to get the same port number.  Since a single broker may
	serve multiple clients, it seems safer to get the port on the
	broker first, then try to get the same one locally instead of
	the other way around.
	*/

		/*
		We may need to rewrite the port for the local bind.
		Thus, tweakedAddr (seems rude to rewrite the user's
		input)
		*/
	tweakedAddr = *sinAddr;

	RETRY_REGISTERSOCK:

	if(info->active != PASSIVE) {

		/*
		We're configured to and expecting to use a broker.  We
		should register the socket to the Broker and get a proxy
		ip:port from the Broker and use this proxy address as the
		official network identity
		*/

		if (registerSock(fd, sinAddr->sin_port) != 0) {
			return -1;
		}
		registered_sock = 1;
			/* Ensure we get the same port locally */
		tweakedAddr.sin_port = info->cport;
	}

	if (bind(fd, (struct sockaddr *)&tweakedAddr, addrlen) < 0) {
		if(registered_sock) { 
			unregisterSock(info);
			if( (errno == EADDRINUSE) && (sinAddr->sin_port == 0) ) {
				LOG(D_FULL, "GCB_bind(%d[GCB_SOCKET], %p->%s, %d): "
						"Collision attempting to allocate same port "
						"(%d) locally as on the broker.  Tried %d times "
						"out of %d.\n", fd, my_addr, _sin2str(sinAddr),
						addrlen, (int)ntohs(tweakedAddr.sin_port),
						num_collisions+1, MAX_COLLISIONS);
				num_collisions++;
				if(num_collisions < MAX_COLLISIONS) { goto RETRY_REGISTERSOCK; }
			}
		}
		ERROR("GCB_bind(%d[GCB_SOCKET], %p->%s, %d): binding the socket locally failed.", fd, my_addr, _sin2str(sinAddr), addrlen);
		_GCB_errno = GCB_ESYSCALLFAILED;
		return -1;
	}

	if (_GCB_post_bind_save_info(fd, info, "GCB_bind", 0) < 0) {
		if (registered_sock) {
			unregisterSock(info);
		}
		return -1;
	}

	LOG(D_FULL, "\tLocal: %s\n", _ipport2str(info->ip, info->port));
	return 0;
}

int
_CB_bind(int fd, struct sockaddr *my_addr, socklen_t addrlen)
{
    int result;
	struct sockaddr_in localAddr, *sinAddr;
	struct SockInfo *info;
	socklen_t length = sizeof(localAddr);
    int newfd;

	if (!_CB_initialized) { _CB_initialize(); }

	// socket should be registered as TCP or UDP socket
	if (_CB_fdInfo[fd].type == MANAGER || _CB_fdInfo[fd].type == HIDDEN) {
		LOG(D_BASIC, "_CB_bind is called for internal GCB socket\n");
        _GCB_errno = GCB_EINTERNALSOCK;
		return -1;
	} else if (_CB_fdInfo[fd].type == NOT_DEFINED) {
        result = bind(fd, my_addr, addrlen);
        _GCB_errno = (result < 0) ? GCB_ESYSCALLFAILED : 0;
        return result;
	}

	info = _CB_fdInfo[fd].info;
	sinAddr = (struct sockaddr_in *)my_addr;
	// Check if the socket has already been bound
	if (info->status != CREATED) {
		LOG(D_BASIC, "_CB_bind: status is not CREATED\n");
        _GCB_errno = GCB_ESOCKINVALIDSTATUS;
		return -1;
	}

	// GCB does not support bind with ip specified for machines behind firewall
	// or inside private network
    /*
	if (info->active == ACTIVE && sinAddr->sin_addr.s_addr != INADDR_ANY) {
		LOG(D_BASIC, "_CB_bind: does not support bind with ip specified for machines\
                      behind firewall or inside private network\n");
		return -1;
	}
    */

	// Get ip address of the machine
	if (!(info->ip)) {
		if (_myIP(&(info->ip)) < 0) {
			LOG(D_BASIC, "_CB_bind: _myIP failed\n");
            _GCB_errno = GCB_ESYSCALLFAILED;
			return -1;
		}
	}

	/* Bind the socket to the given address */
	// If the socket has already been locally bound and INADDR_ANY passed,
	// we do not bind the underlying socket and use the previous address
	// if the socket is bound locally already
	if (info->port) {
		if (info->port == sinAddr->sin_port || // trying to bind to the same address
			(sinAddr->sin_addr.s_addr == INADDR_ANY && sinAddr->sin_port == 0) // trying to bind to ephermeral
		) {
			// nothing need to be done
		} else {
            // unbound fd
            if (_CB_fdInfo[fd].type == TCP) {
                newfd = socket(AF_INET, SOCK_STREAM, 0);
            } else {
                newfd = socket(AF_INET, SOCK_DGRAM, 0);
            }
            if (newfd < 0) {
                ERROR("socket creation failed");
                _GCB_errno = GCB_ESYSCALLFAILED;
                return -1;
            }
            if (_CB_replaceFD(fd, newfd) < 0) {
                LOG(D_BASIC, "Can't replace socket\n");
                _GCB_errno = GCB_ESYSCALLFAILED;
                return -1;
            }
            info->port = 0;
		}
	}
    if (!(info->port)) {
	// if the socket is not bound locally, bind it now
		if (bind(fd, my_addr, addrlen) < 0) {
			ERROR("_CB_bind: binding the socket locally failed");
            _GCB_errno = GCB_ESYSCALLFAILED;
			return -1;
		}
		if (getsockname(fd, (struct sockaddr *) &localAddr, &length) < 0) {
			ERROR("_CB_bind: getsockname failed");
            _GCB_errno = GCB_ESYSCALLFAILED;
            return -1;
		}
		info->port = localAddr.sin_port;
	}
	LOG(D_FULL, "\tLocal: %s\n", _ipport2str(info->ip, info->port));

	/* If PASSIVE socket we do not register the socket to the Broker and use local ip:port
	   as the network identity */
	if (info->active == PASSIVE) {
		info->cip = info->ip;
		info->cport = info->port;
		info->status = BOUND;
		LOG(D_FULL, "\tOfficial: %s\n", _ipport2str(info->cip, info->cport));
        _GCB_errno = 0;
		return 0;
	}

	/* Here this must be a private machine and application tries to bind the socket to
	   INADDR_ANY. We should register the socket to the Broker and get a proxy ip:port
	   from the Broker and use this proxy address as the official network identity */
    result = registerSock(fd, ((struct sockaddr_in *)my_addr)->sin_port);
    _GCB_errno = (result < 0) ? GCB_ESYSCALLFAILED : 0;
    return result;
}

int
GCB_local_bind(int fd, struct sockaddr *my_addr, socklen_t addrlen)
{
	int rval;
	int last_wrapper_status;
	struct SockInfo *info;
	struct sockaddr_in *sin_addr;

	if( ! useGCB() ) {
		return bind(fd, my_addr, addrlen);
	}

	if (!_CB_initialized) { _CB_initialize(); }

	sin_addr = (struct sockaddr_in *)my_addr;

    if (fd < 0 || fd >= _GCB_max_no_fds) {
        LOG(D_BASIC, "GCB_outbound_bind(%d, %s, %d): File descriptor %d is outside of valid range from 0 through %d.  Returning error...\n", fd, _sin2str(sin_addr), (int)addrlen, fd, _GCB_max_no_fds);
        errno = EBADF;
        _GCB_errno = GCB_EBADARG;
        return -1;
    }

	info = _CB_fdInfo[fd].info;
	if (!info) {
        LOG(D_BASIC, "GCB_outbound_bind(%d, %s, %d): GCB has no information about file descriptor %d. Socket must have been created without GCB_socket(), which should be a fatal programmer error.\n", fd, _sin2str(sin_addr), (int)addrlen, fd);
        errno = EBADF;
        _GCB_errno = GCB_EBADARG;
        return -1;
	}

	if (info->status != CREATED) {
		LOG(D_BASIC, "GCB_outbound_bind(%d[GCB_SOCKET], %p->%s, %d): WARNING: This socket is not in the CREATED state. It may have already been bound using GCB_bind.\n", fd, my_addr, _sin2str(sin_addr), addrlen);
		_CB_dumpSockInfo(info, D_BASIC, "  ");
		_GCB_errno = GCB_ESOCKINVALIDSTATUS;
		errno = EINVAL; // Matches Linux/glibc behavior. circa 2006.
		return -1;
	}

	if (_GCB_pre_bind_init_info(fd, my_addr, info, "GCB_local_bind") < 0) {
		return -1;
	}

	last_wrapper_status = _GCB_disable_wrappers(1);
	rval = bind(fd, my_addr, addrlen);
	_GCB_disable_wrappers(last_wrapper_status);

	if (rval < 0 ) {
		LOG(D_BASIC, "\tFailed to bind socket %d (%s) locally for an outbound connection\n", fd, _sin2str(sin_addr));
		_GCB_errno = GCB_ESYSCALLFAILED;
		return -1;
	}

	if (_GCB_post_bind_save_info(fd, info, "GCB_local_bind", 1) < 0) {
		return -1;
	}

	LOG(D_FULL, "\tSuccessfully bound socket %d (%s) locally [%s] for an outbound connection\n", fd, _sin2str(sin_addr), _ipport2str(info->cip, info->cport));
	_GCB_errno = 0;
	return rval;
}


/**
 * Helper function to initialize the local IP inside a SockInfo record
 * for a given fd before calling bind().
 *
 * @param fd
 *   The fd we want to bind()
 * @param my_addr,
 *   The IP address the caller is trying to bind().
 * @param info
 *   Pointer to the struct SockInfo record associated with the fd.
 * @param func_name
 *   The top-level function calling this helper (used for log messages).
 */
int
_GCB_pre_bind_init_info(int fd, struct sockaddr *my_addr,
						struct SockInfo *info, const char* func_name)
{
	int last_wrapper_status;
	struct sockaddr_in *sinAddr;
	sinAddr = (struct sockaddr_in *)my_addr;

		// This function assumes local system calls, but is sometimes
		// called when the wrappers are enabled.  So, we manually
		// disable and restore internally so that we're always safe.
	last_wrapper_status = _GCB_disable_wrappers(1);

	// Get ip address of the machine
	if (!(info->ip)) {
        if (sinAddr->sin_addr.s_addr != INADDR_ANY) {
            info->ip = sinAddr->sin_addr.s_addr;
        }
		else if (_myIP(&(info->ip)) < 0) {
			LOG(D_BASIC, "%s(%d[GCB_SOCKET], %p->%s): Unable to determine local IP address (_myIP failed).\n", func_name, fd, my_addr, _sin2str(sinAddr));
            _GCB_errno = GCB_ESYSCALLFAILED;
			_GCB_disable_wrappers(last_wrapper_status);
			return -1;
		}
	}

	_GCB_disable_wrappers(last_wrapper_status);
	return 0;
}


/**
 * Helper function to update the SockInfo record for a given fd after a bind().
 *
 * @param fd
 *   The fd that was bound.
 * @param info
 *   Pointer to the struct SockInfo record associated with the fd.
 * @param func_name
 *   The top-level function calling this helper (used for log messages).
 * @param is_local
 *   Is this socket for local use only, or registered at the GCB broker?
 */
int
_GCB_post_bind_save_info(int fd, struct SockInfo *info,
						 const char* func_name, int is_local)
{
	int last_wrapper_status;
	struct sockaddr_in localAddr;
	socklen_t length = sizeof(localAddr);

		// This function assumes local system calls, but is sometimes
		// called when the wrappers are enabled.  So, we manually
		// disable and restore internally so that we're always safe.
	last_wrapper_status = _GCB_disable_wrappers(1);

	// Find the port.
	if (getsockname(fd, (struct sockaddr *) &localAddr, &length) < 0) {
		ERROR("%s(%d): Attempting to retrieve the socket's name with getsockname failed", fd);
		_GCB_errno = GCB_ESYSCALLFAILED;
		_GCB_disable_wrappers(last_wrapper_status);
		return -1;
	}
	info->port = localAddr.sin_port;

	// If it's a passive socket, there's nothing at the broker, so we
	// record the same IP/port in both fields.
	if (info->active == PASSIVE || is_local) {
		info->cip = info->ip;
		info->cport = info->port;
		// Why is this status only set in PASSIVE mode?!
		info->status = BOUND;
	}
	_GCB_disable_wrappers(last_wrapper_status);
	return 0;
}
