/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#ifndef _CB_CLIENT
#define _CB_CLIENT

#include "gcbstring.h"

#define GCB_HEARTBEAT_RATE 40
#define GCB_KEEP_EYE_RATE 15
#define GCB_FLUSH_RATE 250
#define GCB_CONN_TIMEOUT 120
#define GCB_CMD_TIMEOUT 40
#define GCB_MAX_TRY 5
#define GCB_TRY_INTERVAL 2
#define GCB_MAX_UDP_MSGS 50

// fd type
#define NOT_DEFINED 0
#define TCP 1
#define UDP 2
#define MANAGER 3
#define HIDDEN 4

// fd status
#define NONE 4
#define CREATED 5
#define BOUND 6
#define LISTENING 7
#define ACTIVE_CONNECTING 8
#define PASSIVE_CONNECTING 9
#define CONNECTED 10
#define CONN_FAILED 11

// mfd status
#define NO_MNG -1
#define NEED_RECOVER -2

#ifdef _cplusplus
extern "C" {
#endif

struct conn {
	int fd;
	uint32_t ip;	// network byte order
	uint16_t port;	// network byte order
	struct conn * next;
};

struct rcver {
	uint32_t cip;		// network byte order
	uint16_t cport;		// network byte order
	uint32_t sinIP;		// network byte order
	uint16_t sinPort;	// network byte order
	struct rcver * pre;
	struct rcver * next;
};

struct UDP_msg {
    int bytes;
    char *data;
    struct sockaddr_in sender;
    struct UDP_msg *next;
};

struct SockInfo {
	int owner;		// the smallest fd value that shares this socket info
	int manager;	// management fd. Only one mng fd is used even multiple fds share the sock info
	int refs;		// reference counts
    // official (leased) address
	uint32_t cip;	// network byte order
	uint16_t cport;	// network byte order
	char active;	// ACTIVE or PASSIVE
	int status;
	time_t blkedAt; // the time when status becomes PASSIVE_CONNECTING
    // physical ip:port in network byte order
	uint32_t ip; // network byte order
	uint16_t port; // network byte order
	int err;
	struct conn * head;		// the list of established connection is maintained
	struct conn * tail;		// by these two pointer. Pretty much like the listen
							// queue of TCP, but we don't have entries for pending
							// connections

	// Next 5 fields are valid only for UDP socket	
	struct sockaddr_in *peer; // for connected UDP socket, this points the connected peer info
	struct rcver *rcvers;	// mapping between receiver's cip:cport and sinIP:sinPort
    int nomsgs; // # of UDP msgs that get buffered
    struct UDP_msg *rbuf_head; // head of recv buffer
    struct UDP_msg *rbuf_tail; // tail of recv buffer
    /*
	char *rcv_buf;			// Receiving buffer
	int nobytes;			// The size of data buffered
	struct sockaddr_in lastSender;	// The sender of data that is buffered at 'rcv_buf'
    */
};

struct FDinfo {
	int type;	// TCP, UDP, MANAGER, NOT_DEFINED, HIDDEN
	struct SockInfo *info;	// the pointer to the socket information
							// Not NULL only for CB socket
};

int useGCB();
char * getRouteTbl();
struct sockaddr_in * getBroker();
struct sockaddr_in * getRelayServer();

/* The following 5 functions are wrapper of long socket calls. Since the process
 * should handle asynchronous events, it can't be blocked waiting for a long
 * system call being completed. Instead it should periodically check and handle
 * GCB async events. Those functions are all about this.
 */
// args: timeout: same meaning as the timeout of select
//       other args are exactly the same as its normal socket call counterpart
// ret: 0 if succ, -1 o/w
// ERROR: other than regular error values returned by underlying socket calls:
//    - ETIMEDOUT: connection can't be established within timeout occurs
int _CB_do_connect(
		int sock,
		const struct sockaddr *serv_addr,
		socklen_t addrlen,
		struct timeval *timeout);
// ret: newfd if succ, -1 o/w
// ERROR: other than regular error values returned by underlying socket calls:
//    - ETIMEDOUT: connection can't be established within timeout occurs
int _CB_do_accept(
		int sock,
		struct sockaddr *addr,
		socklen_t *addrlen,
		struct timeval *timeout);

int _CB_replaceFD(int target, int withfd);
int _CB_directConnect(uint32_t ip);
void _CB_initialize();
void _CB_freeSockInfo(struct SockInfo *sock);
void _CB_delRcvrs(struct SockInfo *sock);
int _CB_serializeSock(int fd, struct SockInfo *sock, GCBString str);
struct SockInfo *_CB_unserializeSock(char *str, int *fd);
void _CB_dumpSockInfo(struct SockInfo *sock, int dflag, char *pre);
int _BP_execv(const char *path, char *const argv[]);
int _BP_execl(const char *path, const char *arg, ...);
int _BP_execve(const char *path, char *const argv [], char *const envp[]);
void _CB_addFDs(int *maxfd, fd_set *rdfds, fd_set *wrfds, fd_set *exfds);

/* Returns the string of serialized GCB information passed from
 * the paent process to this one.
 *
 * You pass this string into _CB_inheritFromMaxFD and _CB_inheritFrom
 *
 * Caller must GCB_string_delete the returned string.  Returns
 * an empty string if there is no information to inherit. 
 * Can only be called once!
 */
GCBString _CB_inheritFromString();
/* What is the largest FD inherited from a parent GCB process?
 *
 * Pass in the output of _CB_inheritFromString().  Note that
 * you cn only call _CB_inheritFromString _once_.
 */
int _CB_inheritFromMaxFD(GCBString inheritFromString);
/* Restore FD information serialized by the parent GCB process.
 *
 * Pass in the output of _CB_inheritFromString().  Note that
 * you cn only call _CB_inheritFromString _once_.
 */
void _CB_inheritFrom(GCBString inheritFromString);
void _CB_inheritTo();
void _CB_inheritToBuffers(GCBString env_name_buf, GCBString env_value_buf);


/*************** Functions for Handling Asynchronous Events ********************/
// Handle asynchronous Broker command for each CB sockets
// @args
//    + in:
//		- maxfd: maxfd passed to select
//		- mngfds: fd_set whose members are read ready
// @return: 0, succ
//         -1, otherwise
void _CB_handleAsyncEvent(int maxfd, fd_set *rdfds, fd_set *wrfds);
int _CB_checkAsyncEvent();
int _CB_servMngSocks(int maxfd, fd_set *mngfds);

void _CB_servUDPsocks(int maxfd, fd_set *rdfds);

/* Handle a DUMMY msg from a specific peer.
   This function handles asynchronous UDP msg delivered to "fd", until a DUMMY msg
   from "to" is received. Upon receiving a DUMMY from "to", DUMMY_ACK will be replied
   to the peer, the peer gets registered as a receiver, and returns the peer's sin addres
   into "phyTo"
 */
// #include <CBclient.h>
// @args:
//    + in:
//       - fd: file descriptor of the UDP socket for which msgs should be handled
//       - pcip: the desired peer's cip
//       - pcport: the desired peer's cport
//    + out:
//       - phyTo: the peer's physical address
// @return: 0 if succeed, -1 if failed
int _CB_handleDummy(int fd, uint32_t pcip, uint16_t pcport, struct sockaddr_in *phyTo);

int _CB_addRcvr(struct SockInfo *sock, uint32_t cip,
		uint16_t cport, uint32_t ip, uint16_t sinPort);
struct rcver * _CB_getRcvr(struct SockInfo *sock, uint32_t ip, uint16_t port);
void _CB_triggerHeartBeat(time_t current);

/* Open a UDP channel using GCB protocol */
// args:
//   + in:
//     - info: the record of GCB socket.
//     - ip, port: cip:cport of the peer
//   + out:
//     - peer: address of the peer's socket that can be directly talked to
// ret: 1 if succ,
//      0 if failed but worth trying direct connection
//      -1 final verdict for failure
int _CB_openchannel(struct SockInfo *info, uint32_t ip, uint16_t port, struct sockaddr_in *peer);

// update _CB_maxfd
void _CB_setmaxfd(int upper);
// Create a MANAGER for 'sock'
int _CB_newManager(struct SockInfo *sock);
// close manager fd and clean related information
void _CB_rmManager(int mfd);

void _CB_recoverGCBsockets();

int _CB_bind(int fd, struct sockaddr *my_addr, socklen_t addrlen);
int _CB_select(int maxfd, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout);

/* non-blocking GCB helper-methods */
int _GCB_is_blocking( int fd, int cmd_flags, int* fd_flags );
int _GCB_nonblock_send( int fd, const void *msg, size_t len, int flags,
						int fd_flags );
int _GCB_nonblock_recv( int sock, void *buf, size_t count, int flags,
						struct timeval *timeout, int fd_flags );
int _GCB_nonblock_recvfrom( int sock, void *buf, size_t count, int flags,
							struct sockaddr *from, socklen_t *fromlen,
							struct timeval *timeout, int fd_flags );


/** Disable GNU linker trick. GCB INTERNAL USE ONLY!

This is intended ONLY for internal GCB use; client code should
NEVER call this function.

If you're using the GNU linker trick, this will force calls to
functions that GCB wraps to go through to the libc version
instead of the GCB_ version.  If you're not doing the trick,
this does nothing*.

* Strictly speaking it will fiddle with a meaningless variable.

For safety, save the return value.  When you're done call it
again with the return value you got previously.  Furthermore,
disable wrappers for the shortest time possible; you should never
exit from a function that disabled the wrappers without restoring
them to their previous state.

Example usage (for code that needs the "real" getsockname:

int last_wrapper_status;
last_wrapper_status = _GCB_disable_wrappers(1);
getsockname(fd, name, namelen);
_GCB_disable_wrappers(last_wrapper_status);

*/
int _GCB_disable_wrappers(int disable);

#ifdef _cplusplus
}
#endif

#endif
