/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"

/* Functions in this file have TCP/UDP system call counterparts and do the same things
   as their counterpart. The only difference is that they will give up when timeout occurs
 */
int
_CB_do_connect(
		int sock,
		const struct sockaddr *serv_addr,
		socklen_t addrlen,
		struct timeval *timeout)
{
	int val, err, rst, ready;
    	socklen_t len;
	fd_set *wrfds = NULL;
	fd_set *exfds = NULL;

	// if nonblocking fd, call connect and return the result 
	if ((val = fcntl(sock, F_GETFL, 0)) < 0) {
		ERROR("_CB_do_connect: fcntl failed");
		return -1;
	}
	if (val & O_NONBLOCK) {
		return connect(sock, serv_addr, addrlen);
	}

	// do nonblocking connect
	if (fcntl(sock, F_SETFL, val | O_NONBLOCK) < 0) {
		ERROR("_CB_do_connect: fcntl failed");
		return -1;
	}
	rst = connect(sock, serv_addr, addrlen);
	if (!rst) {
		goto succ;
	} else if (errno != EINPROGRESS) {
		goto error;
	}

	wrfds = _CB_mallocFDSet();
	exfds = _CB_mallocFDSet();

	while(1) {
		// wait until connection finished or failed
		_CB_FD_ZERO(wrfds);
		_CB_FD_ZERO(exfds);
		FD_SET(sock, wrfds);
		FD_SET(sock, exfds);
			/* XXX This looks non-portable! We should decrement timeout ourselves */
		ready = select(sock+1, NULL, wrfds, exfds, timeout);
		if (ready < 0) {
			if (errno == EINTR) {continue;}
			goto error;
		} else if (ready == 0) {
			errno = ETIMEDOUT;
			goto error;
		} else {
			break;
		}
	}

	// check if the connection established
	err = 0;
	len = sizeof(err);
	if (getsockopt(sock, SOL_SOCKET, SO_ERROR, &err, &len) < 0) {
		ERROR("_CB_do_connect: getsockopt failed");
		LOG(D_BASIC, "In Solaris, getsockopt returns fail with errno set appropriately"
                " if the pending connection has failed.\n");
		goto error;
	}
	if (err) {
		LOG(D_BASIC, "_CB_do_connect: getsockopt reported connection failed: %s\n",
                strerror(err));
		errno = err;
		goto error;
	} else {
		goto succ;
	}

error:
	_CB_freeFDSet( wrfds );
	_CB_freeFDSet( exfds );
	if (fcntl(sock, F_SETFL, val) < 0) {
		ERROR("_CB_do_connect: fcntl failed");
	}
	return -1;
succ:
	_CB_freeFDSet( wrfds );
	_CB_freeFDSet( exfds );
	if (fcntl(sock, F_SETFL, val) < 0) {
		ERROR("_CB_do_connect: fcntl failed");
        return -1;
	}
	return 0;
}

int
_CB_do_accept(int sock, struct sockaddr *addr, socklen_t *addrlen, struct timeval *timeout)
{
	int val, rst, ready;
	struct timeval tv;

	// if nonblocking fd, call accept and return the result
	if ((val = fcntl(sock, F_GETFL, 0)) < 0) {
		ERROR("_CB_do_accept: fcntl failed");
		return -1;
	}
	if (val & O_NONBLOCK) {
		//LOG(D_FULLDEBUG, "_CB_do_accept: socket is nonblocking\n");
		rst = accept(sock, addr, addrlen);
		//LOG(D_FULLDEBUG, "\taccpet returned %d\n", rst);
		return rst;
	}

	// do nonblocking accept and go back to blocking mode
	//LOG(D_FULLDEBUG, "_CB_do_accept: making socket nonblocking\n");
	if (fcntl(sock, F_SETFL, val | O_NONBLOCK) < 0) {
		ERROR("_CB_do_accept: fcntl failed");
		return -1;
	}
	//LOG(D_FULLDEBUG, "\tdoing nonblocking accept\n");
	rst = accept(sock, addr, addrlen);
	//LOG(D_FULLDEBUG, "\taccept returned %d:%s\n", rst, strerror(errno));
	//LOG(D_FULLDEBUG, "\tresuming socket's blocking option\n");
	if (rst > 0 || errno != EAGAIN) {
		if (fcntl(sock, F_SETFL, val) < 0) {
			ERROR("_CB_do_accept: fcntl failed");
            return -1;
		}
		return rst;
	}
	if (fcntl(sock, F_SETFL, val) < 0) {
		ERROR("_CB_do_accept: fcntl failed");
        return -1;
	}

	// wait until sock becomes ready for accept
	while(1) {
		fd_set *rdfds = _CB_mallocFDSet();
		_CB_FD_ZERO( rdfds );
		FD_SET(sock, rdfds);
		//LOG(D_FULLDEBUG, "\tselecting(%d, %s, ...)\n", sock+1, _prtfds(sock+1, &rdfds));
		if (timeout) {
			tv.tv_sec = timeout->tv_sec;
			tv.tv_usec = timeout->tv_usec;
			ready = select(sock+1, rdfds, NULL, NULL, &tv);
		} else {
			ready = select(sock+1, rdfds, NULL, NULL, NULL);
		}
		_CB_freeFDSet( rdfds );
		if (ready < 0) {
			if (errno == EINTR) {continue;}
			ERROR("_CB_do_accept: select returned error");
			return -1;
		} else if (ready == 0) {
			errno = ETIMEDOUT;
			LOG(D_BASIC, "_CB_do_accept: select return nothing within timeout\n");
			return -1;
		} else {
			rst = accept(sock, addr, addrlen);
			if (rst < 0) {
				ERROR("_CB_do_accept: accept failed");
				return -1;
			}
			return rst;
		}
	}
}

/* Replace target fd by another fd */
// #include <CBclient>
// @arg:	target: fd to be replaced. Socket options of target remain the same
//			withfd: fd by which "target" is replaced
// @return:	0, if succeed
//			-1, otherwise
// Note: Both fds are closed if failure occurs at some point
int
_CB_replaceFD(int target, int withfd)
{
	int val;

	// copy socket options
	if ((val = fcntl(target, F_GETFL, 0)) < 0) {
		ERROR("_CB_replaceFD: fcntl(GET) failed");
		return -1;
	}
	if (fcntl(withfd, F_SETFL, val) < 0) {
		ERROR("_CB_replaceFD: fcntl(SET) failed");
		return -1;
	}
	if (dup2(withfd, target) < 0) {
		close(withfd);
        ERROR("_CB_replaceFD: dup2 failed");
		return -1;
	}
	close(withfd);

	return 0;
}


int __inside_gcb_wrapper = 0;
int 
_GCB_disable_wrappers(int disable)
{
	int ret = __inside_gcb_wrapper;
	__inside_gcb_wrapper = disable;
	return ret;
}

