/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include "GCB_errno.h"

// Global data structure
extern int _CB_initialized;
extern int _CB_maxfd;
extern struct FDinfo *_CB_fdInfo;
extern struct SockInfo *_CB_sockInfo[];
extern time_t _CB_lastHeartBeat;
extern int _GCB_max_no_fds;


/*
   Since applications may blindly call GCB_close or close--for example
   child processes sometimes call close for every fd except a few designated ones,
   this close fuction should be very careful to do appropriate steps depending
   on the type and also status of fd being deleted.

   What steps should be done is best explained with the following table.
   -----------------------------------------------------------
        type    |                 action
   -----------------------------------------------------------
        TCP     | - close fd and clean
        UDP     |   fd's slot         
                | - decrease refs
                | - if refs == 0, clean everything
                |   else, clean _CB_fdInfo[fd] slot only
   -----------------------------------------------------------
      MANAGER   | - do nothing
                | - error: GCB_EINTERNALSOCK
   -----------------------------------------------------------
      hidden    | - do nothing
                | - error: GCB_EINTERNALSOCK
   -----------------------------------------------------------
   NOT_DEFINED  | - return close(fd)
   -----------------------------------------------------------
*/

int
GCB_close(int fd)
{
	int i;
	struct SockInfo *info;
    int result;

	if (!_CB_initialized) { _CB_initialize(); }

	if (fd < 0 || fd >= _GCB_max_no_fds || _CB_fdInfo[fd].type == NOT_DEFINED) {
		LOG(D_FULL, "close(%d[NOT_DEFINED])\n", fd);
		result = close(fd);
		_GCB_errno = (result < 0) ? GCB_ESYSCALLFAILED : 0;
		return result;
	}

	// Manager and Hidden sockets shouldn't be closed directly by the
	// application, but closing all fds is common practice on startup
	// and prior to exec()ing, so don't complain loudly.
	if (_CB_fdInfo[fd].type == MANAGER || _CB_fdInfo[fd].type == HIDDEN) {
		LOG(D_FULL, "GCB_close(%d): Ignoring close of an internal GCB socket\n", fd);
		errno = EBADF;
        _GCB_errno = GCB_EINTERNALSOCK;
		return -1;
	}

	// fd must be TCP or UDP socket
	info = _CB_fdInfo[fd].info;
	LOG(D_FULL, "[GCB_close(%d[GCB_SOCKET])\n", fd);
	// Decrease the reference count and free SockInfo if ref_count is zero
	if (--(info->refs) == 0) {
        LOG(D_FULL, "\tref = 0\n");
		// close manager
		if (info->manager != NO_MNG && info->manager != NEED_RECOVER) {
            LOG(D_FULL, "\tclosing manager[%d]\n", info->manager);
			_CB_rmManager(info->manager);
		}
        LOG(D_FULL, "\tfreeing data structure\n");
		// TODO Shouldn't we close all of the HIDDEN fds associated with
		//   this socket (those in the connection list)?
		_CB_freeSockInfo(info);
	} else if (info->owner == fd) {
        LOG(D_FULL, "\tref = %d\n", info->refs);
		// We need to change info->owner
		for (i=fd+1; i<_CB_maxfd; i++) {
			if (_CB_fdInfo[i].type == _CB_fdInfo[fd].type  && _CB_fdInfo[i].info == info) {
				info->owner = i;
                LOG(D_FULL, "\towner changed to [%d]\n", i);
			}
		}
	}

	// Clear slot for fd
	_CB_fdInfo[fd].type = NOT_DEFINED;
	_CB_fdInfo[fd].info = NULL;

	if (fd == _CB_maxfd - 1) {
		_CB_setmaxfd(fd);
	}

    if (close(fd) < 0) {
        _GCB_errno = GCB_ESYSCALLFAILED;
        return -1;
    } else {
        _GCB_errno = 0;
        return 0;
    }
}
