/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include "GCB_errno.h"

// Global data structure
extern int _CB_initialized;
extern struct FDinfo *_CB_fdInfo;
extern int _GCB_max_no_fds;


/* Open TCP channel using GCB protocol */
// args:
//     - info: the record of GCB socket. All the related fields of this record will be updated.
//     - serv_addr: peer's socket address
// ret: 1 if succ,
//      0 if failed but worth trying direct connection
//      -1 final verdict for failure
static int
GCB_TCP_connect(struct SockInfo *info, const struct sockaddr *serv_addr)
{
	struct sockaddr_in sinAddr, *ptr;
	int tempSock, received, newfd, val;
    socklen_t addrlen;
	struct timeval tv;
	char opcode, errcode;
	uint32_t ipAddr;
	uint16_t port;

	/* Send connect request to the destination's CB */
	// Create an ephemeral socket and connect it to the destination Broker
	if ((tempSock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		ERROR("GCB_TCP_connect(%d)-%s: socket failed",
                info->owner, _sin2str((struct sockaddr_in *)serv_addr));
        _GCB_errno = GCB_ESYSCALLFAILED;
        return -1;
	}
	ptr = (struct sockaddr_in *)serv_addr;
	bzero(&sinAddr, sizeof(sinAddr));
	sinAddr.sin_family = AF_INET;
	sinAddr.sin_addr.s_addr = ptr->sin_addr.s_addr;
	sinAddr.sin_port = htons(BROKER_PUBLIC_PORT);
	if (connect(tempSock, (struct sockaddr *)&sinAddr, sizeof(sinAddr)) < 0) {
        // connection to the dest Broker failed
        ERROR("GCB_TCP_connect: connection to the dest broker failed");
		close(tempSock);
        return 0;
	}
	// send connect request message
	if (_CBsend(tempSock, CONNECT, info->cip, info->cport, info->active, info->ip, info->port,
				ptr->sin_addr.s_addr, ptr->sin_port, 0) < 0) {
        LOG(D_BASIC, "GCB_TCP_connect: can't send CONNECT to the broker\n");
		close(tempSock);
        _GCB_errno = (errno == EPIPE) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
        return -1;
	}

	// Receive response
	received = _CBrecv(tempSock, &opcode, NULL, NULL, NULL, &ipAddr, &port, NULL, NULL, &errcode);
	if (received <= 0) {
        LOG(D_BASIC, "GCB_TCP_connect: can't recv resp to CONNECT from the broker\n");
		close(tempSock);
        _GCB_errno = (received == 0) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
        return -1;
	}
	switch(opcode) {
		case PASSIVE_CONNECT:
			if ( listen(info->owner, 5) < 0 ) {
					// TODO If this listen can actually fail, we need to
					//   handle it better.
				LOG(D_BASIC, "GCB_TCP_connect(%d):%s:"
					" CONNECT sent; PASSIVE_CONNECT received;"
					" but listen failed, errno=%d (%s)\n",
					info->owner,_sin2str((struct sockaddr_in *)serv_addr),
					errno, strerror(errno));
			}
			if (_CBsend(tempSock, ACK, 0, 0, 0, 0, 0, 0, 0, 0) < 0) {
				LOG(D_BASIC, "GCB_TCP_connect: Failed to send ACK for PASSIVE_CONNECT\n");
				close(tempSock);
                _GCB_errno = (errno == EPIPE) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
				info->status = CONN_FAILED;
				return -1;
			}
			close(tempSock);
			info->status = PASSIVE_CONNECTING;
			tv.tv_sec = GCB_CONN_TIMEOUT;
			tv.tv_usec = 0;
			addrlen = sizeof(struct sockaddr_in);
			newfd = _CB_do_accept(info->owner, (struct sockaddr *)&sinAddr, &addrlen, &tv);
			if (newfd < 0) {
				if (errno == EAGAIN) {
					info->blkedAt = time(NULL);
					errno = EINPROGRESS;
					info->err = EINPROGRESS;
					return -1;
				} else {
					LOG(D_BASIC, "GCB_TCP_connect(%d):%s:"
                            " CONNECT sent; PASSIVE_CONNECT received;"
                            " but _CB_do_accept failed, errno=%d (%s)\n",
                            info->owner, _sin2str((struct sockaddr_in *)serv_addr),
							errno, strerror(errno));
					info->status = CONN_FAILED;
                    _GCB_errno = GCB_ESYSCALLFAILED;
					return -1;
				}
			}
			// copy socket options
			if ((val = fcntl(info->owner, F_GETFL, 0)) < 0) {
				ERROR("GCB_TCP_connect: fcntl(F_GETFL) failed");
                _GCB_errno = GCB_ESYSCALLFAILED;
                close(newfd);
                return -1;
			}
			if (fcntl(newfd, F_SETFL, val) < 0) {
				ERROR("GCB_TCP_connect: fcntl(F_SETFL) failed");
                _GCB_errno = GCB_ESYSCALLFAILED;
                close(newfd);
                return -1;
			}
			if (dup2(newfd, info->owner) < 0) {
				ERROR("GCB_TCP_connect: dup2 failed");
                _GCB_errno = GCB_ESYSCALLFAILED;
                close(newfd);
                return -1;
			}
			close(newfd);
			info->status = CONNECTED;
			LOG(D_FULL, "\tPassively connected to %s\n", _sin2str(&sinAddr));
			return 1;
		case NAK: // should be "NAK(PEER_NOT_REGISTERED)"
			close(tempSock);
            return 0;
		case ACTIVE_CONNECT: // "contact(ipAddrs:port)"
			close(tempSock);
			bzero(&sinAddr, sizeof(sinAddr));
			sinAddr.sin_family = AF_INET;
			sinAddr.sin_addr.s_addr = ipAddr;
			sinAddr.sin_port = port;
			LOG(D_FULL, "\tconnecting actively to %s...\n", _sin2str(&sinAddr)); 
			tv.tv_sec = GCB_CONN_TIMEOUT;
			tv.tv_usec = 0;
			if (_CB_do_connect(info->owner, (struct sockaddr *) &sinAddr, sizeof(sinAddr), &tv) < 0) {
				if (errno == EINPROGRESS) {
					info->status = ACTIVE_CONNECTING;
					info->err = errno;
					return -1;
				}
				info->status = CONN_FAILED;
				info->err = errno;
                _GCB_errno = GCB_ESYSCALLFAILED;
				return -1;
			}
            info->status = CONNECTED;
            return 1;
		default:
			close(tempSock);
            LOG(D_BASIC, "GCB_TCP_connect(%d):%s: CONNECT sent"
                   " but unexpected message received\n",
                    info->owner, _sin2str((struct sockaddr_in *)serv_addr));
			return 0;
	} // end of switch(opcode)
}

// ret: 1 if succ,
//      0 if failed but worth trying direct connection
//      -1 final verdict for failure
int
_CB_openchannel(struct SockInfo *info, uint32_t dip, uint16_t dport, struct sockaddr_in *peer)
{
	struct sockaddr_in sinAddr;
	int tempSock;
	char opcode, errcode;
	uint32_t ip;
	uint16_t port;
    int rcved;

	LOG(D_FULL, "<_CB_openchannel> to %s\n", _ipport2str(dip, dport));
	/* Send INIT_COMM to destination CB */
	if ((tempSock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		ERROR("_CB_openchannel(%d):%s: socket failed",
                info->owner, _ipport2str(dip, dport));
        _GCB_errno = GCB_ESYSCALLFAILED;
        return -1;
	}
	bzero(&sinAddr, sizeof(sinAddr));
	sinAddr.sin_family = AF_INET;
	sinAddr.sin_addr.s_addr = dip;
	sinAddr.sin_port = htons(BROKER_PUBLIC_PORT);
	if (connect(tempSock, (struct sockaddr *)&sinAddr, sizeof(sinAddr)) < 0) {
        ERROR("_CB_openchannel: connection to the dest broker failed");
		close(tempSock);
		return 0;
	}
	if (_CBsend(tempSock, INIT_COMM, info->cip, info->cport, info->active,
				info->ip, info->port, dip, dport, 0) < 0) {
		LOG(D_BASIC, "_CB_openchannel: can't send INIT_COMM to %s\n", _sin2str(&sinAddr));
        _GCB_errno = (errno == EPIPE) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
		close(tempSock);
		return -1;
	}

	/* Get the reply from Broker */
    rcved = _CBrecv(tempSock, &opcode, NULL, NULL, NULL, &ip, &port, NULL, NULL, &errcode);
    if (rcved <= 0) {
		LOG(D_BASIC, "_CB_openchannel: sent INIT_COMM to %s, but can't get response\n",
                _sin2str(&sinAddr));
		close(tempSock);
        _GCB_errno = (rcved == 0) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
		return -1;
	}
	close(tempSock);
	switch(opcode) {
		case PASSIVE_INIT:
			LOG(D_FULL, "waiting for DUMMY...\n");
            if (_CB_handleDummy(info->owner, dip, dport, peer) < 0) {
                _GCB_errno = GCB_ETIMEOUT;
                return -1;
            } else {
                return 1;
            }
		case ACTIVE_INIT:
			peer->sin_family = AF_INET;
			peer->sin_addr.s_addr = ip;
			peer->sin_port = port;
			if (_CB_addRcvr(info, dip, dport, ip, port) < 0) {
                LOG(D_BASIC, "_CB_openchannel: _CB_addRcvr failed\n");
                return -1;
            }
            return 1;
		case NAK:
			return 0;
		default:
            LOG(D_BASIC, "_CB_openchannel: sent INIT_COMM to %s,"
                    " but invalid msg received from the dest Broker\n",
                    _sin2str(&sinAddr));
			return 0;
	}
}

/* Connect a TCP socket to the address of the peer.
   Note that we do not support UDP connect at this moment
 */
// #include <sys/types.h>
// #include <sys/socket.h>
// @args: same as "connect"
// @return: same as "connect"
// @errors: in addition to errors which are returned by "connect":
//			+ error codes from "_CB_bind"
//			+ error codes from "accept"
//			+ ECOMM - Failed to send "CONNECT" request to CB or failed
//					  to receive reply from CB
//			+ EPROTO - unexpected reply from CB received
int
GCB_connect(int fd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
    int result;
	struct SockInfo *info;
	struct timeval tv;
	struct sockaddr_in sinAddr, *peer, *brokeraddr;
	int rst;
	socklen_t t_addrlen;

	if (!_CB_initialized) { _CB_initialize(); }

    if (fd < 0 || fd >= _GCB_max_no_fds || _CB_fdInfo[fd].type == NOT_DEFINED) {
        LOG(D_FULL, "connect(%d[NOT_DEFINED], ...)\n", fd);
        result = connect(fd, serv_addr, addrlen);
        LOG(D_FULL, "returning %d\n", result);
        _GCB_errno = (result < 0) ? GCB_ESYSCALLFAILED : 0;
        return result;
	}

	if (_CB_fdInfo[fd].type == MANAGER || _CB_fdInfo[fd].type == HIDDEN) {
		LOG(D_BASIC, "GCB_connect called for internal GCB socket\n");
		errno = EBADF;
		_GCB_errno = GCB_EINTERNALSOCK;
		return -1;
	}

	info = _CB_fdInfo[fd].info;
	peer = (struct sockaddr_in *)serv_addr;
	if (peer->sin_family != AF_INET || addrlen != sizeof(struct sockaddr_in)) {
        LOG(D_FULL, "connect(%d[GCB_SOCKET], not sinaddr_in...)\n", fd);
		rst = connect(fd, serv_addr, addrlen);
        if (rst < 0) {
            if (errno == EINPROGRESS) {
			    info->status = ACTIVE_CONNECTING;
                _GCB_errno = 0;
            } else {
                info->err = errno;
                _GCB_errno = GCB_ESYSCALLFAILED;
            }
        } else {
            _GCB_errno = 0;
        }
        LOG(D_FULL, "returning %d(%s)\n", rst, strerror(errno));
		return rst;
	}

	LOG(D_FULL, "GCB_connect(%d[GCB_SOCKET], %s)\n", fd, _sin2str(peer));
	/* socket should be registered in _CB_fdInfo table if it passed all the tests
	 * when it is created  */
	tv.tv_sec = GCB_CONN_TIMEOUT;
	tv.tv_usec = 0;

	/*******************************************************************************
	 * 									UDP unconnect
	 *******************************************************************************/
	if (peer->sin_family == AF_UNSPEC) {
		if (_CB_fdInfo[fd].type == UDP) {
			if (info->status == CONNECTED) {
				if (connect(fd, serv_addr, addrlen) < 0) {
					ERROR("GCB_connect: unconnect failed");
					info->status = CONN_FAILED;
					free(info->peer);
					info->peer = NULL;
					info->err = errno;
                    _GCB_errno = GCB_ESYSCALLFAILED;
					return -1;
				}
				if (info->port) {
					info->status = BOUND;
				} else {
					info->status = CREATED;
				}
				free(info->peer);
				info->peer = NULL;
                _GCB_errno = 0;
				return 0;
			} else {
				// already disconnected
                _GCB_errno = 0;
				return 0;
			}
		} else {
			// invalid type
			LOG(D_BASIC, "[GCB_connect] unconnect call for nonUDP socket\n"); 
            _GCB_errno = GCB_ESOCKTNOSUPPORT;
			return -1;
		}
	}

	/*******************************************************************************
	 * 								TCP/UDP connect
	 *******************************************************************************/
	// 1. if TCP and registered to the RelayServer, unregister it
	if (_CB_fdInfo[fd].type == TCP) {
        if (info->manager == NEED_RECOVER) {
            info->manager = NO_MNG;
        } else if (info->manager != NO_MNG) {
            LOG(D_FULL, "deleting manager[%d]\n", info->manager);
		    _CB_rmManager(info->manager);
        }
		// Since the Broker uses cip to decide who should initiate connect, the cip
		// should be remained unchanged here
		info->cport = 0;
	}

	// 2. check the routing table
	rst = _CB_directConnect(peer->sin_addr.s_addr);
    if (rst > 0) { // direct
        goto direct_connection;
    } else if (rst < 0) { // error in reading the routing table
        LOG(D_BASIC, "Routing table must be in a bad shape."
                " Trying direct connection anyway...\n");
        goto direct_connection;
    }

	// 3. Here we try GCB connection procedure. Many Errors will result in trying
    // direct connection.
	// 3.1 if not bound, do it without registering to the RelayServer
	if (info->status == CREATED) {
		t_addrlen = sizeof(sinAddr);
		bzero(&sinAddr, addrlen);
		if (bind(fd, (struct sockaddr *)&sinAddr, t_addrlen) < 0) {
			ERROR("[GCB_connect(%d)]%s: bind failed", fd, _sin2str(peer));
            // Things gone really badly. Not worth trying direct connection
			info->err = errno;
            _GCB_errno = GCB_ESYSCALLFAILED;
            return -1;
		}
		if (_myIP(&(info->ip)) < 0) {
			LOG(D_BASIC, "[GCB_connect(%d)]%s: _myIP() failed", fd, _sin2str(peer));
            // Things gone really badly. Not trying direct connection
			info->err = errno;
            _GCB_errno = GCB_ESYSCALLFAILED;
            return -1;
		}
		bzero(&sinAddr, t_addrlen);
		if (getsockname(fd, (struct sockaddr *) &sinAddr, &t_addrlen) < 0) {
			ERROR("[GCB_connect(%d)]%s: getsockname failed", fd, _sin2str(peer));
            // Things gone really badly. Not trying direct connection
			info->err = errno;
            _GCB_errno = GCB_ESYSCALLFAILED;
            return -1;
		}
		info->port = sinAddr.sin_port;
		// If the client is being served by a Broker, the Broker's ip should be used
		brokeraddr = getBroker();
		if (brokeraddr) {
			info->cip = brokeraddr->sin_addr.s_addr;
		} else {
			info->cip = info->ip;
		}
		info->cport = info->port;
		info->status = BOUND;
	}
	// 3.2 Now do GCB connection
	if (_CB_fdInfo[fd].type == TCP) {
        rst = GCB_TCP_connect(info, serv_addr);
        if (rst < 0) {
            if (errno == EINPROGRESS) {
                _GCB_errno = 0;
                return -1;
            }
            LOG(D_BASIC, "[GCB_connect(%d)]%s: GCB_TCP_connect failed\n",
                    fd, _sin2str(peer));
            return -1;
        } else if (rst == 0) {
            goto direct_connection;
        }
        _GCB_errno = 0;
        return 0;
	} else {
		// open the udp channel with the peer
        rst = _CB_openchannel(info, peer->sin_addr.s_addr, peer->sin_port, &sinAddr);
        if (rst < 0) {
            LOG(D_BASIC, "[GCB_connect(%d)]%s: _CB_openchannel failed\n",
                    fd, _sin2str(peer));
            return -1;
        } else if (rst == 0) {
            goto direct_connection;
        }
        if (connect(fd, (struct sockaddr *)&sinAddr, sizeof(sinAddr)) < 0) {
			char peerstr[24]; /* Can't use _sin2str twice in one ERROR() */
			_sin2strbuf(peer, peerstr);
            ERROR("[GCB_connect(%d)]%s:"
                    " _CB_openchannel succeed. Now connecting to %s..."
                    " UDP connect failed",
                    fd, peerstr, _sin2str(&sinAddr));
            info->err = errno;
            _GCB_errno = GCB_ESYSCALLFAILED;
            return -1;
        }
		if (!(info->peer)) {
			info->peer = malloc(sizeof(struct sockaddr_in));
            if (!(info->peer)) { ABORT("malloc failed"); }
		}
		info->peer->sin_addr.s_addr = sinAddr.sin_addr.s_addr;
		info->peer->sin_port = sinAddr.sin_port;
		info->status = CONNECTED;
        _GCB_errno = 0;
		return 0;
	}

direct_connection:
    // One of the followings is true:
    //   - routing table says "direct" for this peer
    //   - routing table does not have an entry for this peer
    //   - routing table is not defined or does not exist
    //   - routing table is in a bad shape
    //   - connection attemp using the GCB procedure, above, failed for reasons that
    //     worth trying direct connection
    // We try direct connection any of these cases because direct connection is the
    // default.
    // we reset *peer to the original value
	peer = (struct sockaddr_in *)serv_addr;
    if (_CB_fdInfo[fd].type == UDP) {
        if (connect(fd, serv_addr, addrlen) < 0) {
            ERROR("[GCB_connect(%d)]%s: UDP connect failed", fd, _sin2str(peer));
            info->err = errno;
            _GCB_errno = GCB_ESYSCALLFAILED;
            return -1;
        }
        if (!(info->peer)) {
            info->peer = malloc(sizeof(struct sockaddr_in));
            if (!(info->peer)) { ABORT("malloc failed"); }
        }
        info->peer->sin_addr.s_addr = peer->sin_addr.s_addr;
        info->peer->sin_port = peer->sin_port;
    } else {
        if (_CB_do_connect(fd, serv_addr, addrlen, &tv) < 0) {
            if (errno == EINPROGRESS) {
                info->status = ACTIVE_CONNECTING;
                info->err = errno;
                LOG(D_FULL, "returning %d(%s)\n", rst, strerror(errno));
                return -1;
            }
            LOG(D_BASIC, "[GCB_connect(%d)]%s:"
                    " direct connect using _CB_do_connect failed\n",
                    fd, _sin2str(peer));
            info->status = CONN_FAILED;
            info->err = errno;
            _GCB_errno = GCB_ESYSCALLFAILED;
            LOG(D_FULL, "returning %d(%s)\n", rst, strerror(errno));
            return -1;
        }
    }
    info->status = CONNECTED;
    _GCB_errno = 0;
    return 0;
}
