/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include "GCB_errno.h"

int _CB_initialized = 0;
// max fd of TCP, UDP, or MANAGER
int _CB_maxfd = 0;
struct FDinfo *_CB_fdInfo = NULL;
time_t _CB_lastCheck = 0;
time_t _CB_lastHeartBeat = 0;
time_t _CB_lastFlush = 0;
void (*_Recovery_failed_app_callback)(void) = NULL;
void (*_Broker_down_app_callback)(void) = NULL;
time_t _Broker_down_max = 0;
time_t _Broker_down_since = 0;
time_t _Broker_down_last_callback = 0;
time_t _CB_lastRecovery = 0;
int _GCB_errno = 0;
int _GCB_max_no_fds = 0;

struct errno_2_str {
    int err_no;
    char *err_str;
};
// Error string must not be longer than MAX_ERROR_STRING characters
#define MAX_ERROR_STRING 150
struct errno_2_str GCB_errno_translation[] = {
    {0, "No GCB error" },
    {GCB_ESYSCALLFAILED, "errno explains the reason best"},
    {GCB_EBROKERNOTAVAIL, "Broker is not available"},
    {GCB_EBADMSGFORMAT, "GCB message in a bad format received"},
    {GCB_EBADMSGTYPE, "Invalid type of GCB message received"},
    {GCB_EINTERNALSOCK, "the socket is being internally used by GCB"},
    {GCB_EBADARG, "invalid argument is passed"},
    {GCB_ESOCKTNOSUPPORT, "operation not supported for fd"},
    {GCB_ESOCKINVALIDSTATUS, "the socket is in invalid status for the call"},
    {GCB_EPEERNOTREGISTERED, "the peer is not registered to the dest Broker"},
    {GCB_ETIMEOUT, "Can't hear from the dest in PASSIVE connection"},
    {GCB_EMISC, "Can't explain why"},
    // Caution!!!
    // New error numbers and its string representations must be added before
    // these lines
    {GCB_NO_ERROR_BEYOND_THIS, NULL}
};

char *
GCB_strerror(int GCB_err)
{
    static char msg[MAX_ERROR_STRING*2];
    int i;

    for(i=0; (GCB_errno_translation[i].err_no != GCB_NO_ERROR_BEYOND_THIS); i++) {
        if(GCB_errno_translation[i].err_no == GCB_err) {
            if (errno) {
                snprintf(msg, MAX_ERROR_STRING*2, "%s (errno %d: %s)",
                        GCB_errno_translation[i].err_str,
                        errno, strerror(errno));
            } else {
                snprintf(msg, MAX_ERROR_STRING*2, "%s",
                        GCB_errno_translation[i].err_str);
            }
            return msg;
        }
    }

    sprintf(msg, "Invalid error number passed");
    return msg;
}

void
GCB_Broker_down_callback_set(void (*app_callback)(void), int timeout)
{
    _Broker_down_max = timeout;
    _Broker_down_app_callback = app_callback;
}

void
GCB_Recovery_failed_callback_set(void (*app_callback)(void))
{
    _Recovery_failed_app_callback = app_callback;
}

void
_CB_cleanup(void)
{
    int i;

    for(i=0; i<_CB_maxfd; i++) {
        if (_CB_fdInfo[i].info) { GCB_close(i); }
    }
    _CB_maxfd = 0;
    _CB_lastCheck = 0;
    _CB_lastHeartBeat = 0;
    _CB_lastFlush = 0;
    _Broker_down_since = 0;
	_Broker_down_last_callback = 0;
}

void
_CB_setmaxfd(int upper)
{
	int i;

	if (_CB_maxfd == 0) return;

	for (i=upper-1; i>=0; i--) {
		if (_CB_fdInfo[i].type != NOT_DEFINED) {
			_CB_maxfd = i + 1;
			return;
		}
	}
	LOG(D_FULL, "_CB_setmaxfd: no GCB or mng socket remained\n");
    _Broker_down_since = 0;
	_Broker_down_last_callback = 0;
	_CB_maxfd = 0;
}

// Create a TCP socket and connect it to RelayServer
int
_CB_newManager(struct SockInfo *sock)
{
	int mfd;
	struct sockaddr_in *relayServer;

	relayServer = getRelayServer();
	if (relayServer == NULL) {
		ERROR("Failed to arrange a relay server through the broker. (_CB_newManager: getRelayServer() failed)");
		return -1;
	}

	if ((mfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		ERROR("Failed to create a new TCP socket for the relay server. (_CB_newManager: socket() failed)");
        _GCB_errno = GCB_ESYSCALLFAILED;
		return -1;
	}

	if (connect(mfd, (struct sockaddr *)relayServer, sizeof(struct sockaddr_in)) < 0) {
		ERROR("Failed to connect to the new relay server %s. (_CB_newManager: connect() failed)", _sin2str(relayServer));
        if (errno == ECONNREFUSED || errno == ETIMEDOUT || errno == ENETUNREACH) {
            _GCB_errno = GCB_EBROKERNOTAVAIL;
        }
		close(mfd);
		return -1;
	}

	sock->manager = mfd;
	_CB_fdInfo[mfd].type = MANAGER;
	_CB_fdInfo[mfd].info = sock;
	if (mfd >= _CB_maxfd) {
		_CB_maxfd = mfd + 1;
	}

	return mfd;
}

void
_CB_rmManager(int mfd)
{
	struct SockInfo *sock;

	sock = _CB_fdInfo[mfd].info;
	sock->manager = NO_MNG;
	_CB_fdInfo[mfd].type = NOT_DEFINED;
	_CB_fdInfo[mfd].info = NULL;
	if (mfd == _CB_maxfd-1) {
		_CB_setmaxfd(mfd);
	}
	close(mfd);
	return;
}

/* Free all connections of SockInfo pointed by 'sock' */
static void
deleteConns(struct SockInfo *sock)
{
	struct conn *ptr, *tptr;

	ptr = sock->head;
	while(ptr) {
		tptr = ptr->next;
		free(ptr);
		ptr = tptr;
	}
	sock->head = NULL;
	sock->tail = NULL;

	return;
}

void
_CB_freeSockInfo(struct SockInfo *sock)
{
    struct UDP_msg *ptr;
	deleteConns(sock);
	_CB_delRcvrs(sock);
    while (sock->rbuf_head) {
        ptr = sock->rbuf_head;
        sock->rbuf_head = ptr->next;
        if (ptr->data) {
            free(ptr->data);
        }
        free(ptr);
    }
	if (sock->peer) {
		free(sock->peer);
	}
	free(sock);
	return;
}

/* Add a receiver to the SockInfo pointed by 'sock' */
// #include <CBclient.h>
// @args:	sock - pointer to socket structure to which the receiver gets added
//			cip - cip of the receiver
//			cport - cport of the receiver
//			sinIP - real ip address of the receiver
//			sinPort - real port number of the receiver
//		note: ip and port numbers should be in network byte order
// @return:	0 if succeed, -1 o/w
int
_CB_addRcvr(struct SockInfo *sock, uint32_t cip,
		uint16_t cport, uint32_t sinIP, uint16_t sinPort)
{
	struct rcver *ptr;

	for (ptr = sock->rcvers; ptr; ptr = ptr->next) {
		if (ptr->cip == cip && ptr->cport == cport) {
			// already exist
			return 0;
		}
	}

	ptr = (struct rcver *) malloc(sizeof(struct rcver));
	if (!ptr) { ABORT("malloc failed"); }

	ptr->cip = cip;
	ptr->cport = cport;
	ptr->sinIP = sinIP;
	ptr->sinPort = sinPort;
	if (sock->rcvers) {
		sock->rcvers->pre = ptr;
	}
	ptr->next = sock->rcvers;
	ptr->pre = NULL;
	sock->rcvers = ptr;

	return 0;
}

/* Get the receiver with cip:cport */
// #include <CBclient.h>
// @args:	fd - fd whose _CB_fdInfo[fd].rcvers will be searched
//			cip, cport: cip and cport of the receiver which is wanted
// @return: the pointer to the receiver with matching cip:cport
//			NULL, if not found
struct rcver *
_CB_getRcvr(struct SockInfo *sock, uint32_t ip, uint16_t port)
{
	struct rcver *ptr = sock->rcvers;

	while (ptr) {
		if (ip == ptr->cip && port == ptr->cport) {
			return ptr;
		}
		ptr = ptr->next;
	}
	return NULL;
}

/* Flush receivers of SockInfo pointed by 'sock' */
void
_CB_delRcvrs(struct SockInfo *sock)
{
	struct rcver *ptr, *tptr;

	ptr = sock->rcvers;
	while (ptr) {
		tptr = ptr;
		ptr = ptr->next;
		free(tptr);
	}
	sock->rcvers = NULL;
	return;
}

void
_CB_dumpSockInfo(struct SockInfo *sock, int dflag, char *pre)
{
	char str[1024];
	struct rcver *rcvr;
	struct conn *cons;

	sprintf(str, "refs=%d ", sock->refs);
	sprintf(&str[strlen(str)], "owner=%d ", sock->owner);
	sprintf(&str[strlen(str)], "mnger=%d ", sock->manager);
	sprintf(&str[strlen(str)], "canon=%s ", _ipport2str(sock->cip, sock->cport));
	if (_CB_fdInfo[sock->owner].type == TCP) {
		strcat(str, "TCP ");
	} else {
		strcat(str, "UDP ");
	}
	if (sock->active == ACTIVE) {
		strcat(str, "ACTIVE ");
	} else {
		strcat(str, "PASSIVE ");
	}
	switch(sock->status) {
		case NONE:
			strcat(str, "NONE ");
			break;
		case CREATED:
			strcat(str, "CREATED ");
			break;
		case BOUND:
			strcat(str, "BOUND ");
			break;
		case LISTENING:
			strcat(str, "LISTENING ");
			break;
		case CONNECTED:
			strcat(str, "CONNECTED ");
			break;
		case CONN_FAILED:
			strcat(str, "CONN_FAILED ");
			break;
		case PASSIVE_CONNECTING:
			strcat(str, "PASSIVE_CONNECTING ");
			break;
		default:
			strcat(str, "Invalid ");
			break;
	}
	sprintf(&str[strlen(str)], "phyAddr=%s ", _ipport2str(sock->ip, sock->port));
	if (sock->peer) {
		sprintf(&str[strlen(str)], "connected=%s ", _sin2str(sock->peer));
	}
	strcat(str, "conn_q:");
	if (sock->head) {
		for (cons = sock->head; cons; cons = cons->next) {
            if (strlen(str) > sizeof(str) - 30) {
                strcat(str, "...");
                return;
            }
			sprintf(&str[strlen(str)], "%d|", cons->fd);
			sprintf(&str[strlen(str)], "%s*", _ipport2str(cons->ip, cons->port));
		}
		// delete the last "*"
		sprintf(&str[strlen(str)-1], " ");
	} else {
		strcat(str, "Empty ");
	}
	strcat(str, "Rcvers:");
	if (sock->rcvers) {
		for (rcvr = sock->rcvers; rcvr; rcvr = rcvr->next) {
            if (strlen(str) > sizeof(str) - 30) {
                strcat(str, "...");
                return;
            }
			strcat(str, _ipport2str(rcvr->cip, rcvr->cport));
			strcat(str, "*");
		}
		str[strlen(str)-1] = '\0';
	} else {
		strcat(str, "Empty");
	}

	LOG(dflag, "%s%s\n", pre, str);
	return;
}
