/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include "GCB_errno.h"

// Global data structure
extern int _CB_initialized;
extern struct FDinfo *_CB_fdInfo;
extern time_t _CB_lastHeartBeat;

void
inheritCBsockets()
{
	// Initialize global data structure if necessary
	if (!_CB_initialized) {
        _CB_initialize();
    }

	// Inherit CB sockets to child process
	_CB_inheritTo();
}

int
GCB_execv(const char *path, char *const argv[])
{
    int result;

	inheritCBsockets();
	result = execv(path, argv);
    return result;
}

int
GCB_execl(const char *path, const char *arg, ...)
{
	va_list ap;
	char *next_arg;
	char const **argv;
	int noargs = 0;
    int result, maxargs;

    // start with 20
    maxargs = 20;
    argv = (char const **) malloc(sizeof(char *) * maxargs);
    if (!argv) { ABORT("malloc failed"); }
	argv[noargs++] = arg;

	va_start(ap, arg);
	while(1) {
        if (noargs == maxargs) {
            // increase by 5
            maxargs += 5;
            argv = (char const **)realloc(argv, sizeof(char *) * maxargs);
            if (!argv) { ABORT("realloc failed"); }
        }
		next_arg = va_arg(ap, char *);
		argv[noargs++] = next_arg;
		if (!next_arg) {
			break;
		}
	}
	va_end(ap);

	inheritCBsockets();

	result = execv(path, (char **const)argv);
    return result;
}

int
GCB_execve(const char *path, char *const argv[], char *const envp[])
{
    int result, i = 0, size = 0;

	GCBString env_name = GCB_string_new_empty(32);
	GCBString env_value = GCB_string_new_empty(2048);
	GCBString gcb_var_buf = GCB_string_new_empty(2048);
	static char** gcb_envp = NULL;

	// Populate our stack buffers with the environment variable we need.
	_CB_inheritToBuffers(env_name, env_value);
	GCB_string_appendGG(gcb_var_buf, env_name);
	GCB_string_appendGP(gcb_var_buf, "=");
	GCB_string_appendGG(gcb_var_buf, env_value);
	GCB_string_delete(env_value);
	GCB_string_delete(env_name);

	// If we're running via clone(), make sure we only leak 1 of these.
	if (gcb_envp) {
		free(gcb_envp);
		gcb_envp = NULL;
	}

	// See how many valid entries are in the existing envp array.
	for (i=0; envp[i]; i++) {
		// NOOP, just counting the size.
	}
	size = i + 1;  // Need 1 extra for the new variable.

	// Now, allocate a new array with space for our extra variable.
	gcb_envp = (char**)malloc(size * sizeof(char*));
	if (!gcb_envp) {
		ERROR("GCB_execve: malloc() failed");
		_GCB_errno = GCB_ESYSCALLFAILED;
		GCB_string_delete(gcb_var_buf);
		return -1;
	}		
	
	// Copy over all the existing pointers
	for (i=0; envp[i]; i++) {
		gcb_envp[i] = envp[i];
	}

	// Finally, add our variable and terminate the array.
	gcb_envp[i] = GCB_string_c_str(gcb_var_buf);
	gcb_envp[i+1] = '\0';

	result = execve(path, argv, gcb_envp);

	// No sense leaking this if execve() failed.
	free(gcb_envp);
	gcb_envp = NULL;
	GCB_string_delete(gcb_var_buf);

    return result;
}
