/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"


char *useGCB();
int _priv_net(uint32_t ip);
void _CB_cleanup(void);

// This function is not used. Condor sets GCB_ENABLE and DPF_ENABLE based upon
// NET_REMAP_SERVICE and backend services--e.g. GCB and DPF--check
// their environment variables (i.e. GCB_ENABLE and DPF_ENABLE) directly w/o checking
// this variable.
int
useGeneric(void)
{
    char *answer;

    answer = getenv("NET_REMAP_ENABLE");
    if (!answer || strcasecmp(answer, "true")) {
        return 0;
    }
    return 1;
}

int
Generic_socket(int domain, int type, int protocol)
{
	if (useGCB()) {
		return GCB_socket(domain, type, protocol);
	}
	return socket(domain, type, protocol);
}

static void
calc_proxy(struct sockaddr_in *proxy, struct sockaddr_in *appaddr)
{
    proxy->sin_family = AF_INET;
    // set ip
    if (_priv_net(ntohl(appaddr->sin_addr.s_addr))) {
        // If Condor calls bind with private ip address, we must let GCB choose ip
        // address appropriately
        proxy->sin_addr.s_addr = htonl(INADDR_ANY);
        proxy->sin_port = appaddr->sin_port;
    } else {
        proxy->sin_addr.s_addr = appaddr->sin_addr.s_addr;
        proxy->sin_port = appaddr->sin_port;
    }
    return;
}
int
Generic_bind(int fd, struct sockaddr *my_addr, socklen_t addrlen)
{
    struct sockaddr_in proxy;
	if (useGCB()) {
        calc_proxy(&proxy, (struct sockaddr_in *)my_addr);
		return GCB_bind(fd, (struct sockaddr *)&proxy, sizeof(struct sockaddr_in));
	}
	return bind(fd, my_addr, addrlen);
}

int
Generic_listen(int fd, int backlog)
{
	if (useGCB()) {
		return GCB_listen(fd, backlog);
	}
	return listen(fd, backlog);
}

int
Generic_connect(int fd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
	if (useGCB()) {
		return GCB_connect(fd, serv_addr, addrlen);
	}
	return connect(fd, serv_addr, addrlen);
}

int
Generic_accept(int fd, struct sockaddr * peer, socklen_t *addrlen)
{
	if (useGCB()) {
		return GCB_accept(fd, peer, addrlen);
	}
	return accept(fd, peer, addrlen);
}

int
Generic_select(
		int maxfd,
		fd_set *readfds,
		fd_set *writefds,
		fd_set *exceptfds,
		struct timeval *timeout)
{
	if (useGCB()) {
		return GCB_select(maxfd, readfds, writefds, exceptfds, timeout);
	}
	return select(maxfd, readfds, writefds, exceptfds, timeout);
}

int
Generic_recv(int fd, void *buf, size_t len, int flags)
{
	if (useGCB()) {
		return GCB_recv(fd, buf, len, flags);
	}
	return recv(fd, buf, len, flags);
}

int
Generic_read(int fd, void *buf, size_t count)
{
	if (useGCB()) {
		return GCB_read(fd, buf, count);
	}
	return read(fd, buf, count);
}

int
Generic_recvfrom(
		int fd,  void *buf, size_t len, int flags,
		struct sockaddr *from, socklen_t *fromlen)
{
	if (useGCB()) {
		return GCB_recvfrom(fd, buf, len, flags, from, fromlen);
	}
	return recvfrom(fd, buf, len, flags, from, fromlen);

}

int
Generic_send(int fd, const void *msg, size_t len, int flags)
{
	if (useGCB()) {
		return GCB_send(fd, msg, len, flags);
	}
	return send(fd, msg, len, flags);
}

int
Generic_sendto(
		int fd,
		const void *msg, size_t len, int flags,
		const struct sockaddr *to, socklen_t tolen)
{
	if (useGCB()) {
		return GCB_sendto(fd, msg, len, flags, to, tolen);
	}
	return sendto(fd, msg, len, flags, to, tolen);
}

int
Generic_write(int fd, const void *buf, size_t count)
{
	if (useGCB()) {
		return GCB_write(fd, buf, count);
	}
	return write(fd, buf, count);
}

int
Generic_close(int fd)
{
	if (useGCB()) {
		return GCB_close(fd);
	}
	return close(fd);
}

int
Generic_dup(int fd)
{
	if (useGCB()) {
		return GCB_dup(fd);
	}
	return dup(fd);
}

int
Generic_dup2(int oldfd, int newfd)
{
	if (useGCB()) {
		return GCB_dup2(oldfd, newfd);
	}
	return dup2(oldfd, newfd);
}

int
Generic_getsockname(int fd, struct sockaddr * name, socklen_t *namelen)
{
	if (useGCB()) {
		return GCB_getsockname(fd, name, namelen);
	}
	return getsockname(fd, name, namelen);
}

int
Generic_getsockopt(
		int fd,
		int level,
		int optname,
		void *opt_val,
		socklen_t *optlen)
{
	if (useGCB()) {
		return GCB_getsockopt(fd, level, optname, opt_val, optlen);
	}
	return getsockopt(fd, level, optname, opt_val, optlen);
}

pid_t
Generic_fork(void)
{
    return fork();
}

int
Generic_execv(const char *path, char *const argv[])
{
	if (useGCB()) {
		return GCB_execv(path, argv);
	}
	return execv(path, argv);
}

int
Generic_execl(const char *path, const char *arg, ...)
{
	va_list ap;
	char *next_arg;
	char const *argv[20];
	int noargs = 0;

	argv[noargs++] = arg;

	va_start(ap, arg);
	while(1) {
		next_arg = va_arg(ap, char *);
		argv[noargs++] = next_arg;
		if (!next_arg) {
			break;
		}
	}
	va_end(ap);

	if (useGCB()) {
		return GCB_execv(path, (char **const)argv);
	}
	return execv(path, (char **const)argv);
}

int
Generic_execve(const char *path, char *const argv [], char *const envp[])
{
	if (useGCB()) {
		return GCB_execve(path, argv, envp);
	}
	return execve(path, argv, envp);
}

void
Generic_cleanup(void)
{
    return _CB_cleanup();
}

void
Generic_set_err_callback(void (*app_callback)(void), int timeout)
{
    if (useGCB()) {
        GCB_Broker_down_callback_set(app_callback, timeout);
    }
}
