/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "client.h"
#include "GCB_errno.h"

int
useGCB()
{
	char *answer;

	answer = getenv("GCB_ENABLE");
	if (!answer || strcasecmp(answer, "true")) {
		return 0;
	}
	return 1;
}

char *
getRouteTbl()
{
	return getenv("GCB_ROUTE");
}

struct sockaddr_in *
getBroker()
{
	static struct sockaddr_in sinAddr;
	struct in_addr inaddr;
	char *BrokerIP;
    uint32_t myips[10];
    int size = 10;
    int i;

	if ( (BrokerIP = getenv("GCB_INAGENT")) == NULL ) {
		return NULL;
	}
	if (!inet_aton(BrokerIP, &inaddr)) {
		ERROR("The environment variable %s is set to \"%s\", which may not be a valid IP address. (getBroker: inet_aton failed)", "GCB_INAGENT", BrokerIP);
        return NULL;
	}

    // if Broker is specified as local process, then act as if no Broker is required
    if (_all_myIP(myips, &size) < 0) {
        ERROR("Failed to enumerate local interfaces.  (getBroker: _all_myIP() failed)");
        return NULL;
    }
    for (i=0; i<size; i++) {
        if (myips[i] == inaddr.s_addr) {
            LOG(D_FULL, "getBroker: The broker is on the local machine. "
				"Acting as if no broker given.\n");
            return NULL;
        }
    }
    
	sinAddr.sin_family = AF_INET;
	sinAddr.sin_addr = inaddr;
	sinAddr.sin_port = htons(BROKER_LOCAL_PORT);

	return &sinAddr;
}

struct sockaddr_in *
getRelayServer()
{
	static char * brokerstr = NULL;
	char * newbrokerstr = 0;
	static struct sockaddr_in sinAddr;
	struct sockaddr_in *brokerAddr;
	int sock;
	char opcode;
	uint32_t cip;
	uint16_t cport;
	char active;
	uint32_t ip;
	uint16_t port;
	uint32_t dip;
	uint16_t dport;
	char errcode;
    int rcved;

	// get Broker's address
	if (!(brokerAddr = getBroker())) {
		return NULL;
	}

	newbrokerstr = _sin2str(brokerAddr);
	if( (brokerstr == NULL) || (strcmp(newbrokerstr, brokerstr) != 0) ) {
		/* First use of broker, or we changed brokers. Log it. */
		free(brokerstr);
		brokerstr = strdup(newbrokerstr);
		LOG(D_BASIC, "Attempting to use GCB broker %s.\n", brokerstr);
	}

	// connect to the Broker
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		ERROR("Failed to create an internal TCP socket. (getRelayServer: socket() failed)");
        _GCB_errno = GCB_ESYSCALLFAILED;
        return NULL;
	}
	if (connect(sock, (struct sockaddr *)brokerAddr, sizeof(struct sockaddr_in)) < 0) {
		char * fail_reason = "";;
		switch(errno) {
			case ECONNREFUSED:  fail_reason = "The connection was refused."; break;
			case ETIMEDOUT:  fail_reason = "The connection timed out."; break;
			case ENETUNREACH:  fail_reason = "The network is unreachable."; break;
		}
		ERROR("Failed to connect to the broker at %s. %s (getRelayServer: connect failed)", _sin2str(brokerAddr), fail_reason);
        if (errno == ECONNREFUSED || errno == ETIMEDOUT || errno == ENETUNREACH) {
            _GCB_errno = GCB_EBROKERNOTAVAIL;
        } else {
            _GCB_errno = GCB_ESYSCALLFAILED;
        }
		close(sock);
		return NULL;
	}

	// send NEW_SOCK
	if (_CBsend(sock, NEW_SOCK, 0, 0, 0, 0, 0, 0, 0, 0) < 0) {
		ERROR("Failed to send request to the broker for a new socket. "
			"(getRelayServer: failed to _CBsend(%d, NEW_SOCK)): ", sock);
        _GCB_errno = (errno == EPIPE) ? 
			GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
		close(sock);
		return NULL;
	}

	// receive RELAY_SERVER packet
    rcved = _CBrecv(sock, &opcode, &cip, &cport, &active, &ip, &port, &dip, &dport, &errcode);
    if (rcved <= 0) {
		ERROR("Failed to receive confirmation from broker. "
			"Remote side may have sent an invalid message, "
			"or may have failed to respond. "
			"(getRelayServer: _CBrecv(%d) failed)", sock );
        _GCB_errno = (rcved == 0) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
		close(sock);
		return NULL;
	}
	if (opcode != RELAY_SERVER) {
		// TODO: instead of printing opcode as an int, export 
		// a version of printOP from ex_cmd.c?
		ERROR("Broker returned invalid respones to request for a new "
			"socket. (getRelayServer: opcode %d returned)", (int)opcode);
        _GCB_errno = GCB_EBADMSGTYPE;
		close(sock);
		return NULL;
	}

	// return the address
	sinAddr.sin_family = AF_INET;
	sinAddr.sin_addr.s_addr = ip;
	sinAddr.sin_port = port;

	close(sock);

	return &sinAddr;
}
