/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

// Global data structure
extern int _CB_initialized;
extern fd_set _CB_fds;
extern fd_set _CB_mfds;
extern int _CB_maxfd;
extern struct FDinfo *_CB_fdInfo;
extern time_t _CB_lastHeartBeat;
extern int _GCB_max_no_fds;

#if ! HAVE_SETENV
int setenv( const char* name, const char* val, int overwrite );
#endif 

/* Caller must GCB_string_delete the returned string.  Returns
 * an empty string if there is no information to inherit. 
 * Can only be called once!  (You can call it again, but it will
 * return nothing. */
GCBString 
_CB_inheritFromString()
{
	GCBString result;
	char env_name[40];
	char * fd_str;
	int fd;
	FILE * fp;
	size_t didread;
	char buf[1024];

	snprintf( env_name, sizeof(env_name), "CB_INHERITFILE%ld", (long)getpid() );
	LOG(D_FULL, "_CB_inheritFromString: environment variable name: %s\n", env_name);
	fd_str = getenv(env_name);
	if( ! fd_str) {
		LOG(D_FULL, "_CB_inheritFromString: No environment string %s.  Parent process probably wasn't GCBized.\n", env_name);
		return GCB_string_new("");
	}
	LOG(D_FULL, "_CB_inheritFromString: %s=%s\n", env_name, fd_str);
	fd = atoi(fd_str);
	if(fd < 1) {
		ERROR("_CB_inheritFromString: No environment string %s contained %s, which appears to be invalid.\n", env_name, fd_str);
		return GCB_string_new("");
	}

	fp = fdopen(fd, "r");
	fseek(fp, 0, SEEK_SET);
	if(fp == NULL) {
		ERROR("_CB_inheritFromString: Unable to open FD %d, which should contain GCB inheritance information.\n", env_name, fd_str);
		return GCB_string_new("");
	}

	result = GCB_string_new_empty(16);

	while( (didread = fread(buf, 1, sizeof(buf)-1, fp)) == (sizeof(buf)-1) ) {
		buf[sizeof(buf)-1] = 0;
		GCB_string_appendGP(result, buf);
		/*fprintf(tmp,"%d reading: %s\n", getpid(), buf);*/
	}
	buf[didread] = 0;
	GCB_string_appendGP(result, buf);

	/*fprintf(tmp, "%d: %s=%s\n[[%s]]\n", getpid(), env_name, fd_str, GCB_string_c_str(result));*/

	fclose(fp);
#if HAVE_UNSETENV
	unsetenv(env_name);
#else
		// if we don't have setenv(), we implement a version below...
	setenv(env_name, "", 1 );
#endif 
	return result;
}

int
_CB_inheritFromMaxFD(GCBString inheritFromString)
{
	const char *str;
	int fd, mng;
	int max_fd = 0;
	int rc;

	str = GCB_string_c_str(inheritFromString);

	while ( (str = strchr( str, '$' )) != NULL ) {
		str++;
		rc = sscanf( str, "%d#%d", &fd, &mng );
		if ( rc >= 1 && fd > max_fd ) {
			max_fd = fd;
		}
		if ( rc >=2 && mng > max_fd ) {
			max_fd = mng;
		}
	}
	return max_fd;
}


// I will leave unpleasant literals in this file at the moment because this is one of
// the places that I have introduced many errors
void
_CB_inheritFrom(GCBString inheritFromString)
{
	const char * tmp;
	char *inheritbuf, *cur_str, *next_str;
	char str[30];
	int fd;
	struct SockInfo *sock;

	// Get environment variable
	tmp = GCB_string_c_str(inheritFromString);
	if (!tmp || *tmp == '\0') {
		LOG(D_FULL, "_CB_inheritFrom: Nothing to inherit\n");
		return;
	}

	inheritbuf = strdup(tmp);
	
	// For each CB_socket inherited
	next_str = inheritbuf;
	while(next_str) {
		// Unserialize the CB_socket
		cur_str = next_str;
		next_str = strchr(cur_str, '$');
		if (next_str) {
			*next_str = '\0';
			next_str++;
		}
		if (!(sock = _CB_unserializeSock(cur_str, &fd))) {
			ERROR("_CB_inheritFrom: failed to unserialize CB_socket "
				  "cur_str: %s", cur_str);
			free(inheritbuf);
            return;
		}
		sprintf(str, "\t[fd=%d] ", fd);
		_CB_dumpSockInfo(sock, D_FULL, str);

		cur_str = next_str;
	}
	free(inheritbuf);
}

/**
 * Populate our environment with information about GCB sockets that
 * should be inherited by a child process.
 *
 * This calls the _CB_inheritToBuffers() method that actually does the
 * work of serializing the datastructures, and then calls setenv() directly.
 *
 * @see _CB_inheritToBuffers()
 */
void
_CB_inheritTo()
{
	GCBString env_name = GCB_string_new_empty(32);
	GCBString inheritbuf = GCB_string_new_empty(2048);
	_CB_inheritToBuffers(env_name, inheritbuf);

	if (setenv(GCB_string_c_str(env_name), GCB_string_c_str(inheritbuf), 1) < 0) {
		ERROR("_CB_inheritTo: setenv() failed");
	}
	GCB_string_delete(inheritbuf);
}

/* Returned string must be GCB_string_delete()d. */
static GCBString 
get_temp_dir()
{
	static int initialized = 0;
	static GCBString tmpdir;
	if( ! initialized ) {
		tmpdir = GCB_string_new("/tmp");
		/* TODO: Respect TMP or TEMP environment variable?  Try fallbacks? */
	}
	return GCB_string_new(GCB_string_c_str(tmpdir));
}

/* Returns FD to a temporary file.  It has already been unlinked.
 * purposes should be a very short, filename-safe string, like
 * "inherit".  It's used to name the file, which may be helpful
 * in debugging in some problems. */
static int
get_temp_file(char * purpose)
{
	GCBString filename_build = get_temp_dir();
	mode_t oldmask;
	int fd;
	int errnotmp;
	char * filename = NULL;
	GCB_string_appendGP(filename_build, "/gcb-");
	GCB_string_appendGP(filename_build, purpose);
	GCB_string_appendGP(filename_build, "-XXXXXX");
	filename = strdup(GCB_string_c_str(filename_build));
	GCB_string_delete(filename_build);
	oldmask = umask(0077); /* Old glibc mkstemp uses the less safe 0666 */
	fd = mkstemp(filename);
	errnotmp = errno; /* If something goes wrong, we want to preserve this errno, not a later one. */
	umask(oldmask);
	if( (fd >= 0) && (unlink(filename) != 0) ) {
		LOG(D_BASIC, "Failed to remove temporary file %s. Will continue, but temporary file won't be cleaned up by GCB.\n", filename);
	}
	free(filename);
	if(errno == 0) { errno = errnotmp; }
	return fd;
}


/**
 * Populate the given buffers with a serialized copy of the data
 * structures about GCB sockets known to this process.
 *
 * @param env_name_buf
 *   GCBString to hold the name of the environment variable to use.
 * @param env_name_buf
 *   GCBString to hold the value of the environment variable to populate.
 */
void
_CB_inheritToBuffers(GCBString env_name_buf, 
					GCBString env_value_buf)
{
	GCBString tempbuf;
	char str[300];
	int i;
	int rst;
	int tempfile, tempfiledup;
	FILE * tempfilef;
	GCBString inheritbuf = GCB_string_new_empty(16);

	tempfile = get_temp_file("inherit");
	if(tempfile < 0) {
		ERROR("_CB_inheritTo: failed to create temporary file. (errno=%d)", errno);
		return;
	}

	snprintf(str, sizeof(str), "CB_INHERITFILE%ld", (long)getpid());
	GCB_string_appendGP(env_name_buf, str);
	LOG(D_FULL, "_CB_inheritTo (pid: %ld) environment variable name: %s\n",
		(long)getpid(), GCB_string_c_str(env_name_buf));

	// Scan _CB_fdInfo[] and close all GCB fds with close-on-exec bit set
	for (i = 0; i < _GCB_max_no_fds; i++) {
		if (_CB_fdInfo[i].info) {
			rst = fcntl(i, F_GETFD);
			if (rst == -1) {
				ERROR("_CB_inheritTo: fcntl(%d) failed", i);
				GCB_close(i);
			} else if (rst) {
				LOG(D_FULL, "_CB_inheritTo: deleting fd %d because "
					"close-on-exec tag is set\n", i);
				GCB_close(i);
			}
		}
	}

	snprintf(str, sizeof(str), "%d", tempfile);
	GCB_string_appendGP(env_value_buf, str);

	/*printf("%s=%s\n", GCB_string_c_str(env_name_buf), GCB_string_c_str(env_value_buf));*/


	// Serialize CB_sockets into file
	tempbuf = GCB_string_new_empty(512);
	for (i = 0; i < _CB_maxfd; i++) {
		if (_CB_fdInfo[i].type != TCP && _CB_fdInfo[i].type != UDP) {
			continue;
		}
		if (_CB_serializeSock(i, _CB_fdInfo[i].info, tempbuf)) {
			GCB_string_delete(tempbuf);
			ERROR("_CB_inheritTo: failed to serialize CB_socket");
            return;
		}
		GCB_string_appendGG(inheritbuf, tempbuf);
		GCB_string_truncate(tempbuf);
		GCB_string_appendGP(inheritbuf, "$");
		sprintf(str, "\t[fd=%d] ", i);
		_CB_dumpSockInfo(_CB_fdInfo[i].info, D_FULL, str);
	}
	GCB_string_delete(tempbuf);
	// delete the last "$"
	GCB_string_trim_last(inheritbuf);

	/* A FILE* is easier to deal with. But we want to fclose it
	 * to free up the associated memory, but that closes the FD
	 * as well. So we need to dupe it.  You know, this is a sad
	 * amount of code just to avoid having to deal with retrying
	 * write(). :-/  */
	tempfiledup = dup(tempfile);
	if(tempfiledup < 0) {
		ERROR("_CB_inheritTo: failed to duplicate temporary file descriptor");
		return;
	}
	tempfilef = fdopen(tempfiledup, "w");
	if(tempfilef == NULL) {
		ERROR("_CB_inheritTo: failed to duplicate temporary file descriptor");
		return;
	}
	fprintf(tempfilef, "%s", GCB_string_c_str(inheritbuf));
	/*printf("Into the temp file: %s\n", GCB_string_c_str(inheritbuf));*/
	fclose(tempfilef);
}


#if ! HAVE_SETENV
int
setenv( const char* name, const char* val, int overwrite ) 
{
	int len = 2;
	char* buf;

	if( ! name ) {
		return -1;
	}

	len += strlen(name);
	if( val ) {
		len += strlen(val);
	}
	buf = (char*)malloc(len*sizeof(char));
	sprintf( buf, "%s=%s", name, val ? val : "" );
		/* we do *NOT* want to free this buffer, since once we call
		   putenv() on it, it becomes part of the environment.
		*/
	return putenv( buf );
}
#endif
