/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include "GCB_errno.h"

// Global data structure
extern int _CB_initialized;
extern struct FDinfo *_CB_fdInfo;
extern int _CB_maxfd;
extern time_t _CB_lastHeartBeat;
extern int _GCB_max_no_fds;

int
GCB_listen(int fd, int backlog)
{
	struct SockInfo *info;
	int received;
	char opcode, errcode, active;
	uint32_t cip, dip, ip;
	uint16_t cport, dport, port;
    int rst;

	if (!_CB_initialized) { _CB_initialize(); }

    if (fd < 0 || fd >= _GCB_max_no_fds || _CB_fdInfo[fd].type == NOT_DEFINED) {
        rst = listen(fd, backlog);
        LOG(D_FULL, "GCB_listen(%d, %d): Unknown file descriptor %d passed directly to listen() which returned %d\n", fd, backlog, fd, rst);
        _GCB_errno = (rst < 0) ? GCB_ESYSCALLFAILED : 0;
        return rst;
	}

	if (_CB_fdInfo[fd].type == MANAGER || _CB_fdInfo[fd].type == HIDDEN) {
		LOG(D_BASIC, "GCB_listen(%d, %d): WARNING: A GCB internal file descriptor was passed to GCB_listen.  This descriptor should never be visible to non-GCB code.\n", fd, backlog);
		errno = EBADF;
        _GCB_errno = GCB_EINTERNALSOCK;
		return -1;
	}

	info = _CB_fdInfo[fd].info;

	LOG(D_FULL, "GCB_listen(%d[GCB_SOCKET], %d)\n", fd, backlog);
	/* socket should have been bounded */
	if (info->status != BOUND) {
		LOG(D_BASIC, "GCB_listen(%d, %d): This socket is not in the bound state.  bind() may not have been called on it, or listen() may have already been called on it.\n", fd, backlog); // TODO: D_MORE, probably human error.  Also: print actual state?
        _GCB_errno = GCB_ESOCKINVALIDSTATUS;
		return -1;
	}

	// If there's no manager socket, remove all our knowledge about this
	// socket. 
	if (info->manager == NO_MNG) {
		LOG(D_FULL, "GCB_listen(%d, %d): Deleting GCB info about manager-less file descriptor %d passed to listen()\n", fd, backlog, fd);

		// If there is more than one reference, we need to track down
		// the other fds.
		if (info->refs > 1) {
			int i;
			for (i = info->owner; i < _CB_maxfd; i++) {
				if (_CB_fdInfo[i].info == info) {
					_CB_fdInfo[i].type = NOT_DEFINED;
					_CB_fdInfo[i].info = NULL;
					if (i == _CB_maxfd - 1) {
						_CB_setmaxfd(i);
					}
				}
			}
		} else {
			_CB_fdInfo[fd].type = NOT_DEFINED;
			_CB_fdInfo[fd].info = NULL;
		}
		_CB_freeSockInfo(info);

		rst = listen(fd, backlog);
		_GCB_errno = (rst < 0) ? GCB_ESYSCALLFAILED : 0;
		return rst;
	}

	// Send LISTEN to Broker
	if (info->manager != NO_MNG && info->manager != NEED_RECOVER) {
		if (_CBsend(info->manager, LISTEN, info->cip, info->cport, 0, 0, 0, 0, 0, 0) < 0) {
            _GCB_errno = (errno == EPIPE) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
			LOG(D_BASIC, "GCB_listen: _CBsend(LISTEN) failed\n");
			_CB_rmManager(info->manager);
            info->manager = NEED_RECOVER;
			return -1;
		}
		// Get the result from the Broker
		received = _CBrecv(info->manager, &opcode, &cip, &cport, &active, &ip, &port, &dip, &dport, &errcode);
		if (received <= 0) {
			ERROR("GCB_listen: LISTEN sent, but _CBCrecv failed");
			_CB_rmManager(info->manager);
            info->manager = NEED_RECOVER;
            _GCB_errno = (received == 0) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
			return -1;
		}
		switch(opcode) {
			case ACK:
				break;
			case NAK:
				errno = errcode;
                _GCB_errno = GCB_ESYSCALLFAILED;
			    _CB_rmManager(info->manager);
				return -1;
			default:
				LOG(D_BASIC, "GCB_listen: unexpected msg received\n");
			    _CB_rmManager(info->manager);
                _GCB_errno = GCB_EBADMSGTYPE;
                return -1;
		}
	}

	if (listen(fd, backlog) < 0) {
        _GCB_errno = GCB_ESYSCALLFAILED;
        _CB_rmManager(info->manager);
		return -1;
	}
	info->status = LISTENING;
    _GCB_errno = 0;
	return 0;
}
