/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "client.h"
#include <assert.h>

/*
  if we can't use MSG_DONTWAIT on this platform, use fcntl()
  to set the socket into O_NONBLOCK mode... 
  this is an internal helper method to share code.
*/
#if ! HAVE_MSG_DONTWAIT
static int
_GCB_set_nonblock_flags( int fd, int* saved_flags )
{
	if( *saved_flags < 0 ) {
			// we don't already have the flags, so grab them now
		if( (*saved_flags = fcntl(fd, F_GETFL, 0)) < 0 ) {
			ERROR("_GCB_set_nonblock_flags: fcntl(%d, F_GETFL) failed", fd);
			return -1;
		}
	}
	if( *saved_flags | O_NONBLOCK ) {
			// nonblocking already set, return immediately that we
			// didn't need to touch the flags.
		return 0;
	}
	if( fcntl(fd, F_SETFL, *saved_flags | O_NONBLOCK) < 0 ) {
			ERROR("_GCB_set_nonblock_flags: fcntl(%d, F_SETFL) failed", fd);
		return -1;
	}
	return 1;
}
#endif /* HAVE_MSG_DONTWAIT */


/*
  this is an internal helper method to share code.  we want to return
  from one of the _GCB_nonblock_* methods, but we have to be careful
  about resetting the fcntl() flags on this fd if we set them
  ourselves. 
*/
static int
_GCB_return( int rval, int fd, int set_flags, int saved_flags )
{
	if( set_flags ) {
		if( fcntl(fd, F_SETFL, saved_flags) < 0 ) {
			ERROR("_GCB_return: fcntl(%d) failed to reset fd\n", fd);
			return -1;
		}
	}
	return rval;
}


/*
  determines if a command on the given fd with the given cmd flags
  will block or not.  if the fd_flags pointer is valid, we save the
  current fd flags for the caller.  returns 1 if we know it'll block,
  0 if we know it won't, and -1 if we hit an error finding out.
*/
int
_GCB_is_blocking( int fd, int cmd_flags, int* fd_flags )
{
	int cur_flags;

#if HAVE_MSG_DONTWAIT
	if( cmd_flags & MSG_DONTWAIT) {
		return 0;
	}
#endif

	if( (cur_flags = fcntl(fd, F_GETFL, 0)) < 0 ) {
		ERROR("_GCB_is_nonblocking: fcntl(%d, F_GETFL) failed", fd);
		return -1;
	}
	if( fd_flags ) {
		*fd_flags = cur_flags;
	}
	if( cur_flags & O_NONBLOCK ) {
		return 0;
	}
	return 1;
}


int
_GCB_nonblock_send( int fd, const void *msg, size_t len, int flags,
					int fd_flags )
{
	fd_set *wrfds;
	int ready;
	int rval = 0;
	int set_flags = 0;
	int saved_flags = fd_flags;

#if ! HAVE_MSG_DONTWAIT
	if( (set_flags = _GCB_set_nonblock_flags(fd, &saved_flags)) < 0 ) {
		LOG( D_BASIC, "_GCB_nonblock_send: no MSG_DONTWAIT and "
			 "_GCB_set_nonblock_flags(%d) failed\n", fd );
		return -1;
	}
#endif

	do {
		// do nonblocking send
#if HAVE_MSG_DONTWAIT
		rval = send( fd, msg, len, flags | MSG_DONTWAIT );
#else
		assert( set_flags );
		rval = send( fd, msg, len, flags );
#endif

		if (rval > 0) {
			LOG(D_FULL, "_GCB_nonblock_send: returning %d\n", rval);
			return _GCB_return( rval, fd, set_flags, saved_flags );
		} else if (errno != EAGAIN) {
			LOG(D_FULL, "_GCB_nonblock_send: returning -1, errno %d (%s)\n",
				errno, strerror(errno));
			return _GCB_return( -1, fd, set_flags, saved_flags );
		}

		// wait until fd becomes ready for write Here we call
		// GCB_select to get notified that fd is write ready, because
		// GCB_select will handle all async events while waiting
		wrfds = _CB_mallocFDSet();
		_CB_FD_ZERO( wrfds );
		FD_SET( fd, wrfds );
		ready = _CB_select( fd+1, NULL, wrfds, NULL, NULL );
		_CB_freeFDSet( wrfds );
		if( ready < 0 && errno != EINTR ) {
			LOG( D_BASIC, "_GCB_nonblock_send: _CB_select() failed\n" );
			return _GCB_return( -1, fd, set_flags, saved_flags );
		}
		// Here fd must be ready to write and the next send should succeed
	} while (1);
		
	LOG( D_BASIC, "_GCB_nonblock_send: IMPOSSIBLE: outside while(1) loop\n" );
	return _GCB_return( -1, fd, set_flags, saved_flags );
}



int
_GCB_nonblock_recv( int sock, void *buf, size_t count, int flags,
					struct timeval *timeout, int fd_flags )
{
	return _GCB_nonblock_recvfrom( sock, buf, count, flags, NULL, NULL,
								   timeout, fd_flags );
}


int
_GCB_nonblock_recvfrom( int sock, void *buf, size_t count, int flags,
						struct sockaddr *from, socklen_t *fromlen,
						struct timeval *timeout, int fd_flags )
{
	fd_set *rdfds;
	int ready;
	int rval = 0;
	int set_flags = 0;
	int saved_flags = fd_flags;

#if ! HAVE_MSG_DONTWAIT
	if( (set_flags = _GCB_set_nonblock_flags(sock, &saved_flags)) < 0 ) {
		LOG( D_BASIC, "_GCB_nonblock_send: no MSG_DONTWAIT and "
			 "_GCB_set_nonblock_flags(%d) failed\n", sock );
		return -1;
	}
#endif

	do {
		// nonblocking recv
#if HAVE_MSG_DONTWAIT
		if( from ) {
			rval = recvfrom( sock, buf, count, flags | MSG_DONTWAIT,
							 from, fromlen );
		} else {
			rval = recv( sock, buf, count, flags | MSG_DONTWAIT );
		}
#else
		assert( set_flags );
		if( from ) {
			rval = recvfrom( sock, buf, count, flags, from, fromlen );
		} else {
			rval = recv( sock, buf, count, flags );
		}
#endif

		if (rval >= 0) {
            LOG(D_FULL, "_GCB_nonblock_recvfrom: returning %d\n", rval);
			return _GCB_return( rval, sock, set_flags, saved_flags );
			return rval;
		} else if (errno != EAGAIN) {
            LOG(D_FULL, "_GCB_nonblock_recvfrom: returning -1, "
				"errno %d (%s)\n", errno, strerror(errno));
			return _GCB_return( -1, sock, set_flags, saved_flags );
		}

		assert( errno == EAGAIN );

		// If nonblocking recv requested, return EAGAIN
		if( (saved_flags & O_NONBLOCK)
#if HAVE_MSG_DONTWAIT
			|| (flags & MSG_DONTWAIT)
#endif
		  )
		{
            LOG(D_FULL, "_GCB_nonblock_recvfrom: returning errno EAGAIN\n");
			return _GCB_return( -1, sock, set_flags, saved_flags );
		}

		// wait until something is ready to recv

		// Here we call _CB_select() to get notified that fd is read
		// ready, because _CB_select() will handle all aynch event
		// while waiting

		rdfds = _CB_mallocFDSet();
		_CB_FD_ZERO( rdfds );
		FD_SET(sock, rdfds);
		ready = _CB_select( sock+1, rdfds, NULL, NULL, timeout );
		_CB_freeFDSet( rdfds );
		if (ready < 0 && errno != EINTR) {
			ERROR("_GCB_nonblock_recvfrom: _CB_select() failed");
				// should we return failure here?
		}
		// Here fd must be ready to read and the next recv should succeed
	} while(1);

	LOG( D_BASIC, "_GCB_nonblock_recv: IMPOSSIBLE: outside while(1) loop\n" );
	return _GCB_return( -1, sock, set_flags, saved_flags );
}

