/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include <GCB.h>
#include "client.h"
#include "GCB_errno.h"
#include "GCB_query.h"

int GCB_broker_query(const char *broker_ip, int query, int *result)
{
	int brokerfd;
	struct sockaddr_in brokeraddr;
	char opcode;
	uint32_t result_query;
	uint32_t result_data;
	char errcode;
	int received;

		/* If the caller gives us the IP address of a broker, then we
		 * query that broker. If they hand us a NULL, then we query
		 * our own broker.
		 */
	if ( broker_ip ) {

		struct in_addr inaddr;
		if ( !inet_aton( broker_ip, &inaddr ) ) {
			LOG(D_BASIC, "GCB_broker_query: Invalid IP address: '%s'\n",
				broker_ip);
			_GCB_errno = GCB_ESYSCALLFAILED;
			return -1;
		}
		brokeraddr.sin_family = AF_INET;
		brokeraddr.sin_addr = inaddr;
		brokeraddr.sin_port = htons(BROKER_LOCAL_PORT);

	} else {

		struct sockaddr_in *our_brokeraddr;
		our_brokeraddr = getBroker();
		if ( our_brokeraddr == NULL ) {
			_GCB_errno = GCB_EBROKERNOTAVAIL;
			return -1;
		}
		brokeraddr = *our_brokeraddr;
	}

	if ( query != GCB_DATA_QUERY_FREE_SOCKS ) {
		LOG(D_BASIC, "GCB_broker_query: Invalid query type %d (only %d is valid)\n", query, GCB_DATA_QUERY_FREE_SOCKS);
		_GCB_errno = GCB_EBADARG;
		return -1;
	}
	
	brokerfd = socket(AF_INET, SOCK_STREAM, 0);
	if ( brokerfd < 0 ) {
		LOG(D_BASIC, "GCB_broker_query: Failed to create socket. %d\n", errno);
		_GCB_errno = GCB_ESYSCALLFAILED;
		return -1;
	}

	if ( connect(brokerfd, (struct sockaddr*)&brokeraddr,
				 sizeof(struct sockaddr_in)) < 0 ) {
		LOG(D_BASIC, "GCB_data_query: Failed to connect to the broker. %d\n", errno);
		_GCB_errno = GCB_ESYSCALLFAILED;
		close( brokerfd );
		return -1;
	}

	if ( _CBsend(brokerfd, DATA_QUERY, query, 0, 0, 0, 0, 0, 0, 0) < 0 ) {
		LOG(D_BASIC, "GCB_broker_query: Can't send data query to the broker. %d\n", errno);
		close( brokerfd );
		_GCB_errno = (errno == EPIPE) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
		return -1;
	}

	received = _CBrecv(brokerfd, &opcode, &result_query, NULL, NULL,
					   &result_data, NULL, NULL, NULL, &errcode);
	close( brokerfd );
	if ( received <= 0 ) {
		LOG(D_BASIC, "GCB_broker_query: can't recv response to data query from the broker. %d\n", errno);
		_GCB_errno = (received == 0) ? GCB_EBROKERNOTAVAIL : GCB_ESYSCALLFAILED;
		return -1;
	}
	if ( opcode == NAK ) {
		LOG(D_BASIC, "GCB_broker_query: broker returned error %d (%s)\n",
			errcode, strerror(errcode)); 
		errno = errcode;
        _GCB_errno = GCB_ESYSCALLFAILED;
		return -1;
	}
	if ( opcode != DATA_RESPONSE ) {
		LOG(D_BASIC, "GCB_broker_query: broker returned unexpected response to data query (opcode %d)\n", (int)opcode);
		_GCB_errno = GCB_EBADMSGTYPE;
		return -1;
	}
	if ( result_query != query ) {
		LOG(D_BASIC, "GCB_broker_query: broker returned unexpected response to data query (query type %d instead of %d)\n", result_query, query);
		_GCB_errno = GCB_EBADMSGTYPE;
		return -1;
	}

	*result = result_data;

	return 0;
}
