/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include "GCB_errno.h"

// Global data structure
extern int _CB_initialized;
extern struct FDinfo *_CB_fdInfo;
extern int _GCB_max_no_fds;

static int
UDPrecv(int fd, void *buf, size_t len, int flags,
		 struct sockaddr *from, socklen_t *fromlen)
{
	int val, bytes, ready;
	fd_set *rdfds;
	struct sockaddr_in *peer;
    struct UDP_msg *msg;
	struct SockInfo *info = _CB_fdInfo[fd].info;

    LOG(D_FULL, "GCB_recv(%d[GCB, UDP])\n", fd);

	if ((val = fcntl(fd, F_GETFL, 0)) < 0) {
		ERROR("UDPrecv: fcntl(%d, F_GETFL) failed", fd);
        _GCB_errno = GCB_ESYSCALLFAILED;
		return -1;
	}
	
	while (info->rbuf_head == NULL) { // receive buffer is empty
		if( (val & O_NONBLOCK)
#if HAVE_MSG_DONTWAIT
			|| (flags & MSG_DONTWAIT)
#endif
		  )
		{
			errno = EAGAIN;
            _GCB_errno = 0;
            LOG(D_FULL, "UDPrecv: returning EAGAIN\n");
			return -1;
		}

		// Wait until the recv buffer is filled with a msg
		// GCB_select will buffer UDP msg to the receiving buffer while handling all async events
		rdfds = _CB_mallocFDSet();
		_CB_FD_ZERO( rdfds );
		FD_SET(fd, rdfds);
		ready = _CB_select(fd+1, rdfds, NULL, NULL, NULL);
		_CB_freeFDSet(rdfds);
		if (ready < 0 && errno != EINTR) {
			ERROR("UDPrecv: _CB_select() failed");
            _GCB_errno = GCB_ESYSCALLFAILED;
			return -1;
		}
	}

	// Here CB recv buf must have a msg
    msg = info->rbuf_head;
    info->rbuf_head = msg->next;
    (info->nomsgs)--;
    if (info->rbuf_head == NULL) { info->rbuf_tail = NULL; }
	bytes = (len < msg->bytes) ? len : msg->bytes;
	memcpy(buf, msg->data, bytes);
	if (from) {
		memcpy(from, &(msg->sender), sizeof(struct sockaddr_in));
		*fromlen = sizeof(struct sockaddr_in);
		if (_CB_addRcvr(info, msg->sender.sin_addr.s_addr, msg->sender.sin_port,
					msg->sender.sin_addr.s_addr, msg->sender.sin_port) < 0) {
			LOG(D_BASIC, "UDPrecv: Can't add NonGCB receiver, "
				"%d bytes copied from CB rcv_buf\n", bytes);
            _GCB_errno = GCB_EMISC;
            return -1;
		}
	}
    free(msg->data);
    free(msg);

    LOG(D_FULL, "UDPrecv: returning %d\n", bytes);
    _GCB_errno = 0;
	return bytes;
}


static int
TCPrecv(int fd, void *buf, size_t len, int flags)
{
	int fcntl_mode;
    int rcved;

    LOG(D_FULL, "GCB_recv(%d[GCB, TCP])\n", fd);

	if( ! _GCB_is_blocking(fd,flags,&fcntl_mode) ) {
			// if the socket is already non-blocking, we can just
			// directly call the real recv()
        rcved = recv(fd, buf, len, flags);
		_GCB_errno = (rcved < 0) ? GCB_ESYSCALLFAILED : 0;
		return rcved;
	}

    rcved = _GCB_nonblock_recv( fd, buf, len, flags, NULL, fcntl_mode );
	_GCB_errno = (rcved < 0) ? GCB_ESYSCALLFAILED : 0;
	return rcved;
}

int
GCB_recvfrom(int fd, void *buf, size_t len, int flags,
		 struct sockaddr *from, socklen_t *fromlen)
{
    int rst;
	struct sockaddr_in *peer;

	if (!_CB_initialized) { _CB_initialize(); }

    if (fd < 0 || fd >= _GCB_max_no_fds) {
		peer = (struct sockaddr_in *)from;
        LOG(D_BASIC,
            "GCB_recvfrom(%s): fd (%d) is invalid. Returning error.\n",
            _sin2str(peer), fd);
        errno = EBADF;
        _GCB_errno = GCB_EBADARG;
        return -1;
    }

	// If fd is not my type or TCP or connected UDP, call recvfrom directly
	if (_CB_fdInfo[fd].type == UDP && _CB_fdInfo[fd].info->status != CONNECTED) {
		return UDPrecv(fd, buf, len, flags, from, fromlen);
	} else {
        rst = recvfrom(fd, buf, len, flags, from, fromlen);
        _GCB_errno = (rst < 0) ? GCB_ESYSCALLFAILED : 0;
        return rst;
	}
}

int
GCB_recv(int fd, void *buf, size_t len, int flags)
{
    int rst;

	if (!_CB_initialized) { _CB_initialize(); }

    if (fd < 0 || fd >= _GCB_max_no_fds || _CB_fdInfo[fd].type == NOT_DEFINED) {
		rst = recv(fd, buf, len, flags);
		_GCB_errno = (rst < 0) ? GCB_ESYSCALLFAILED : 0;
		return rst;
	}

	switch(_CB_fdInfo[fd].type) {
		case UDP:
			return UDPrecv(fd, buf, len, flags, NULL, NULL);
		case TCP:
			return TCPrecv(fd, buf, len, flags);
		default:
			LOG(D_FULL, "GCB_recv called with internal GCB socket\n");
			errno = EBADF;
			_GCB_errno = GCB_EINTERNALSOCK;
			return -1;
	}
}

int
GCB_read(int fd, void *buf, size_t len)
{
    int rst;

	if (!_CB_initialized) { _CB_initialize(); }

    if (fd < 0 || fd >= _GCB_max_no_fds || _CB_fdInfo[fd].type == NOT_DEFINED) {
			rst = read(fd, buf, len);
            _GCB_errno = (rst < 0) ? GCB_ESYSCALLFAILED : 0;
            return rst;
	}

	switch(_CB_fdInfo[fd].type) {
		case UDP:
			return UDPrecv(fd, buf, len, 0, NULL, NULL);
		case TCP:
			return TCPrecv(fd, buf, len, 0);
		default:
			LOG(D_FULL, "GCB_read called with internal GCB socket\n");
			errno = EBADF;
			_GCB_errno = GCB_EINTERNALSOCK;
			return -1;
	}
}
