/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"

static unsigned int masks[33] = {
	0x00000000,
	0x80000000, 0xC0000000, 0xE0000000, 0xF0000000,
	0xF8000000, 0xFC000000, 0xFE000000, 0xFF000000,
	0xFF800000, 0xFFC00000, 0xFFE00000, 0xFFF00000,
	0xFFF80000, 0xFFFC0000, 0xFFFE0000, 0xFFFF0000,
	0xFFFF8000, 0xFFFFC000, 0xFFFFE000, 0xFFFFF000,
	0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00, 0xFFFFFF00,
	0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0, 0xFFFFFFF0,
	0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE, 0xFFFFFFFF
};

// @arg: ip (in network order) of the peer
// @ret: 0, if GCB should be used to contact "ip"
//       1, if the "ip" can be contacted directly
//      -1, if error
int
_CB_directConnect(uint32_t ip)
{
	char *path;
	char line[100], *sip, smask[3], route[10], *next_str;
	struct in_addr inaddr;
	int bits;
	uint32_t mask;
	FILE *rTable;

	// See if the machine has the routing table
	if ((path = getRouteTbl()) == NULL) {
			// no routing table, use "direct" by default...
		return 1;
	}

	// Open GCB routing table
	rTable = fopen(path, "r");
	if (!rTable) {
		ERROR("_CB_directConnect: fopen(%s) failed", path);
		return -1;
	}

	// File format of GCB routing table:
	//   - each line contains one routing rule
	//   - each routing rule is of ip/mask, where
	//     + ip: dotted notation of ip or "*"
	//     + network mask: number of bits to be checked. If "ip" is "*", this field is ignored
	//     + mask: "GCB" or "direct"
	while(fgets(line, sizeof(line), rTable) != NULL) {
		sip = line;
		if( line[0] == '#' ) {
				// skip comments
			continue;
		}
		if( line[0] == '\n' && line[1] == '\0' ) {
				// skip blank lines
			continue;
		}
		next_str = strchr(line, '/');
		if (!next_str) {
			goto error;
		}
		*next_str = '\0';
		next_str = next_str + 1;
		if (sscanf(next_str, "%s %s\n", smask, route) != 2) {
			goto error;
		}
		if (*sip == '*') {
			goto done;
		}
		// ip
		if (!inet_aton(sip, &inaddr)) {
			ERROR("_CB_directConnect: invalid ipaddr, inet_aton failed");
			goto error;
		}
		// network mask
		bits = atoi(smask);
		if (bits < 0 || bits > 32) {
			ERROR("_CB_directConnect: invalid ipaddr, atoi(%s) failed",
				  smask);
			goto error;
		}
		mask = masks[bits];
		// check if found
		if ((ntohl(inaddr.s_addr) & mask) == (ntohl(ip) & mask)) {
			goto done;
		}
	}
	if (!feof(rTable)) {
		ERROR("_CB_directConnect: GCB routing table read failed: "
			  "fgets() returned NULL but we're not at EOF" );
        return -1;
	}
	fclose(rTable);
		// no matching IP in the table, use "direct" by default...
	return 1;

done:
	fclose(rTable);
	if (!strcasecmp(route, "GCB")) {
		return 0;
	} else if (!strcasecmp(route, "direct")) {
		return 1;
	}

error:
	fclose(rTable);
	ERROR("routing table format error with line '%s'", line);
	return -1;
}
