/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include "GCB_errno.h"

// Global data structure
extern int _CB_initialized;
extern int _CB_maxfd;
extern struct FDinfo *_CB_fdInfo;
extern time_t _CB_lastCheck;
extern time_t _CB_lastHeartBeat;
extern time_t _CB_lastRecovery;
extern time_t _Broker_down_since;
extern int _GCB_max_no_fds;

static int
adjustFDs(int maxfd, fd_set *orgrfds, fd_set *orgwfds, fd_set *rfds, fd_set *wfds, fd_set *frdfds, fd_set *fwrfds)
{
	int i, ready = 0;
	struct SockInfo *info;

	for (i = 0; i < maxfd; i++) {
		switch(_CB_fdInfo[i].type) {
			case MANAGER:
			case HIDDEN:
				break;
			case TCP:
				info = _CB_fdInfo[i].info;
				switch(info->status) {
					case NONE:
					case CREATED:
					case BOUND:
					case PASSIVE_CONNECTING:
						break;
					case LISTENING:
						if (orgrfds && FD_ISSET(i, orgrfds) && info->head) {
							FD_SET(i, frdfds);
							ready++;
						}
						break;
					case ACTIVE_CONNECTING:
						if (orgwfds && FD_ISSET(i, orgwfds) && FD_ISSET(i, wfds)) {
							FD_SET(i, fwrfds);
							ready++;
						}
						break;
					case CONNECTED:
						if (orgrfds && FD_ISSET(i, orgrfds) && FD_ISSET(i, rfds)) {
							FD_SET(i, frdfds);
							ready++;
						}
						if (orgwfds && FD_ISSET(i, orgwfds) && FD_ISSET(i, wfds)) {
							FD_SET(i, fwrfds);
							ready++;
						}
						break;
					case CONN_FAILED:
						if (orgwfds && FD_ISSET(i, orgwfds)) {
							FD_SET(i, fwrfds);
							ready++;
						}
                        break;
					default:
						ABORT("adjustFDs: _CB_fdInfo[%d].info->status staled", i);
				}
				break;
			case UDP:
				info = _CB_fdInfo[i].info;
				if (orgrfds && FD_ISSET(i, orgrfds)) {
                    if (info->rbuf_head) {
						FD_SET(i, frdfds);
						ready++;
					}
				}
				break;
			default:
				if (orgrfds && FD_ISSET(i, orgrfds) && FD_ISSET(i, rfds)) {
					FD_SET(i, frdfds);
					ready++;
				}
				if (orgwfds && FD_ISSET(i, orgwfds) && FD_ISSET(i, wfds)) {
					FD_SET(i, fwrfds);
					ready++;
				}
				break;
		}
	}
	return ready;
}

// GCB_select() and _CB_select() are virtually identical. GCB_select()
// is only invoked by the application. All internal GCB functions call
// _CB_select(). We should consider combining the two functions.
// Here's how they differ:
// * GCB_select() has LOG() at beginning
// * cast to (long long) when first setting remained
//   (should be same in both)
// * GCB_select() triggers _CB_recoverGCBsockets()
//   (should _CB_select() also trigger it?)
// * GCB_select() sets _GCB_errno when returning failure 
// * in GCB_select(), if select() fails, retries select with caller's
//   args and logs result before returning failure (this was added to
//   aid in debugging bad file descriptor problems seen by Igor. This
//   may come out once we've determined the cause of the error)

int
GCB_select(int mxfd,
		fd_set *rdfds,
		fd_set *wrfds,
		fd_set *exfds,
		struct timeval *timeout)
{
	int retval, first, maxfd;
	fd_set *readfds = NULL;
	fd_set *writefds = NULL;
	fd_set *expfds = NULL;
	struct timeval tv, before, after;
	long long remained;
	time_t current;
	int bytes;
	int fready;
	fd_set *frdfds = NULL;
	fd_set *fwrfds = NULL;
	char str1[50], str2[50], str3[50];

	// TODO: The application can validly pass in a mxfd larger than
	//   _GCB_max_no_fds. In that case, we shouldn't ignore the higher
	//   fds that may be set in the bit arrays. But we would need to call
	//   _CB_setMaxNumFds(mxfd) so that our fd_sets are large enough.
	maxfd = (mxfd < 0 || mxfd > _GCB_max_no_fds) ? _GCB_max_no_fds : mxfd;

	snprintf(str1, sizeof(str1), "%s", _prtfds(maxfd,rdfds));
	snprintf(str2, sizeof(str2), "%s", _prtfds(maxfd,wrfds));
	snprintf(str3, sizeof(str3), "%s", _prtfds(maxfd,exfds));
	if (timeout) {
		LOG(D_FULL, "GCB_select(mxfd=%d, rd=%s, wr=%s, ex=%s <%d, %d>)\n",
				mxfd, str1, str2, str3, timeout->tv_sec, timeout->tv_usec);
	} else {
		LOG(D_FULL, "GCB_select(mxfd=%d, rd=%s, wr=%s, ex=%s <infinite>)\n",
				mxfd, str1, str2, str3);
	}

	if (!_CB_initialized) { _CB_initialize(); }

	first = 1;
	if (timeout) {
		remained = (long long)timeout->tv_sec * 1000000 + (long long)timeout->tv_usec;
	} else {
		// need to be equal to or greater than GCB_KEEP_EYE_RATE
		remained = GCB_KEEP_EYE_RATE * 1000000;
	}

	readfds = _CB_mallocFDSet();
	writefds = _CB_mallocFDSet();
	expfds = _CB_mallocFDSet();
	frdfds = _CB_mallocFDSet();
	fwrfds = _CB_mallocFDSet();

	// Some applications are smart enough to know that some kernels do not check
	// fds greater than mxfd and pass fd_set that is just as big as mx * 8. Hence
	// we must not access fds bigger than mxfd so as not to segfault
	bytes = (maxfd % 8 == 0) ? maxfd/8: maxfd/8 + 1;
	do {
		// trigger heartbeat
		current = time(NULL);
		if (current - _CB_lastHeartBeat > GCB_HEARTBEAT_RATE) {
			_CB_triggerHeartBeat(current);
			_CB_lastHeartBeat = current;
		}
        if (_Broker_down_since && current - _Broker_down_since > GCB_KEEP_EYE_RATE) {
            if (!_CB_lastRecovery || current - _CB_lastRecovery > GCB_KEEP_EYE_RATE) {
                _CB_recoverGCBsockets();
            }
        }
	
		// copy fds
		_CB_FD_ZERO(readfds);
		if (rdfds) {
			memcpy(readfds, rdfds, bytes);
		}
		_CB_FD_ZERO(writefds);
		if (wrfds) {
			memcpy(writefds, wrfds, bytes);
		}
		_CB_FD_ZERO(expfds);
		if (exfds) {
			memcpy(expfds, exfds, bytes);
		}
        maxfd = (mxfd < 0 || mxfd > _GCB_max_no_fds) ? _GCB_max_no_fds : mxfd;

		// Add fds for manager, blocked connections, listeners, and UDP sockets
		// Also remove fds for passive connects
		_CB_addFDs(&maxfd, readfds, writefds, expfds);

		// For the case that some CB fds are already ready, i.e., for the following
		// cases:
		//		- connection queues for listeners are not empty
		//		- blocked passive connection done or failed
		//		- a UDP msg has been queued at the receiving buffer
		// we do nonblocking select for the first time.
		// We also chop timeout value into multiple of GCB_HEARTBEAT_RATE so that
		// we can send HEARTBEAT
		if (first) {
			tv.tv_sec = 0;
			tv.tv_usec = 0;
		} else if (remained >= GCB_HEARTBEAT_RATE * 1000000) {
			tv.tv_sec = GCB_HEARTBEAT_RATE;
			tv.tv_usec = 0;
		} else {
			tv.tv_sec = remained / 1000000;
			tv.tv_usec = remained % 1000000;
		}
		if (gettimeofday(&before, NULL) < 0) {
			ERROR("GCB_select: gettimeofday() failed"); 
            _GCB_errno = GCB_ESYSCALLFAILED;
			fready = -1;
			goto done;
		}
		snprintf(str1, sizeof(str1), "%s", _prtfds(maxfd,readfds));
		snprintf(str2, sizeof(str2), "%s", _prtfds(maxfd,writefds));
		snprintf(str3, sizeof(str3), "%s", _prtfds(maxfd,expfds));
		LOG(D_FULL, "GCB_select: calling select(rd=%s, wr=%s, ex=%s)\n",
			str1, str2, str3);
		retval = select(maxfd, readfds, writefds, expfds, &tv);
		first = 0;
		if (gettimeofday(&after, NULL) < 0) {
			ERROR("GCB_select: gettimeofday() failed"); 
            _GCB_errno = GCB_ESYSCALLFAILED;
			fready = -1;
			goto done;
		}
		if (timeout) {
			remained = remained -
				(after.tv_sec - before.tv_sec) * 1000000 - (after.tv_usec - before.tv_usec);
		}
		if (retval < 0 ) {
			if (errno != EINTR) {
				LOG(D_BASIC, "GCB_select: select failed - %s\n", strerror(errno));
			}
            _GCB_errno = GCB_ESYSCALLFAILED;
			fready = -1;
			goto done;
		} else if (retval > 0) {
			/* Even if select returns zero, there are cases that either:
			   - UDP msgs have been queued,
			   - connections have been queued,
			   - blocked connections have been finished
			 */
			snprintf(str1, sizeof(str1), "%s", _prtfds(maxfd,readfds));
			snprintf(str2, sizeof(str2), "%s", _prtfds(maxfd,writefds));
			snprintf(str3, sizeof(str3), "%s", _prtfds(maxfd,expfds));
			LOG(D_FULL, "GCB_select: select() returned: "
				"ready=%d, rd=%s, wr=%s, ex=%s\n", retval, str1, str2, str3);

			// handle async events that happen during selecting
			_CB_handleAsyncEvent(maxfd, readfds, writefds);
		}
		// Adjust read and write ready set
		// TODO There's a potential bug here. A large maxfd here may cause
		//   adjustFDs() to read well past the end of rdfds and wrfds,
		//   which are passed in by our caller and are only guaranteed to
		//   be as large as mxfd. We only read from them, so the only
		//   danger is if we read into unmapped memory, which seems
		//   unlikely (famous last words).
		_CB_FD_ZERO(frdfds);
		_CB_FD_ZERO(fwrfds);
		fready = adjustFDs(maxfd, rdfds, wrfds, readfds, writefds, frdfds, fwrfds);
		if (fready < 0) {
			ERROR("GCB_select: adjustFDs failed");
            _GCB_errno = GCB_EMISC;
			fready = -1;
			goto done;
		} else if (fready > 0) {
			break;
		}
	} while ( remained > 0);

	if (rdfds) {
		memcpy(rdfds, frdfds, bytes);
	}
	if (wrfds) {
		memcpy(wrfds, fwrfds, bytes);
	}
	if (exfds) {
		memcpy(exfds, expfds, bytes);
	}
	snprintf(str1, sizeof(str1), "%s", _prtfds(maxfd,frdfds));
	snprintf(str2, sizeof(str2), "%s", _prtfds(maxfd,fwrfds));
	snprintf(str3, sizeof(str3), "%s", _prtfds(maxfd,expfds));
	LOG(D_FULL, "GCB_select: returning rd=%s, wr=%s, ex=%s\n",
		str1, str2, str3);
	_CB_lastCheck = time(NULL);
    _GCB_errno = 0;
 done:
	_CB_freeFDSet( readfds );
	_CB_freeFDSet( writefds );
	_CB_freeFDSet( expfds );
	_CB_freeFDSet( frdfds );
	_CB_freeFDSet( fwrfds );

	return fready;
}

int
_CB_select(int mxfd,
		fd_set *rdfds,
		fd_set *wrfds,
		fd_set *exfds,
		struct timeval *timeout)
{
	int retval, first, maxfd;
	fd_set *readfds = NULL;
	fd_set *writefds = NULL;
	fd_set *expfds = NULL;
	struct timeval tv, before, after;
	long long remained;
	time_t current;
	int bytes;
	int fready;
	fd_set *frdfds = NULL;
	fd_set *fwrfds = NULL;
	char str1[50], str2[50], str3[50];

    maxfd = (mxfd < 0 || mxfd > _GCB_max_no_fds) ? _GCB_max_no_fds : mxfd;

	snprintf(str1, sizeof(str1), "%s", _prtfds(maxfd,rdfds));
	snprintf(str2, sizeof(str2), "%s", _prtfds(maxfd,wrfds));
	snprintf(str3, sizeof(str3), "%s", _prtfds(maxfd,exfds));

	if (!_CB_initialized) { _CB_initialize(); }

	first = 1;
	if (timeout) {
		remained = timeout->tv_sec * 1000000 + timeout->tv_usec;
	} else {
		// need to be equal to or greater than GCB_KEEP_EYE_RATE
		remained = GCB_KEEP_EYE_RATE * 1000000;
	}

	readfds = _CB_mallocFDSet();
	writefds = _CB_mallocFDSet();
	expfds = _CB_mallocFDSet();
	frdfds = _CB_mallocFDSet();
	fwrfds = _CB_mallocFDSet();

	// Some applications are smart enough to know that some kernels do not check
	// fds greater than mxfd and pass fd_set that is just as big as mx * 8. Hence
	// we must not access fds bigger than mxfd so as not to segfault
	bytes = (maxfd % 8 == 0) ? maxfd/8: maxfd/8 + 1;
	do {
		// trigger heartbeat
		current = time(NULL);
		if (current - _CB_lastHeartBeat > GCB_HEARTBEAT_RATE) {
			_CB_triggerHeartBeat(current);
			_CB_lastHeartBeat = current;
		}
	
		// copy fds
		_CB_FD_ZERO(readfds);
		if (rdfds) {
			memcpy(readfds, rdfds, bytes);
		}
		_CB_FD_ZERO(writefds);
		if (wrfds) {
			memcpy(writefds, wrfds, bytes);
		}
		_CB_FD_ZERO(expfds);
		if (exfds) {
			memcpy(expfds, exfds, bytes);
		}
        maxfd = (mxfd < 0 || mxfd > _GCB_max_no_fds) ? _GCB_max_no_fds : mxfd;

		// Add fds for manager, blocked connections, listeners, and UDP sockets
		// Also remove fds for passive connects
		_CB_addFDs(&maxfd, readfds, writefds, expfds);

		// For the case that some CB fds are already ready, i.e., for the following
		// cases:
		//		- connection queues for listeners are not empty
		//		- blocked passive connection done or failed
		//		- a UDP msg has been queued at the receiving buffer
		// we do nonblocking select for the first time.
		// We also chop timeout value into multiple of GCB_HEARTBEAT_RATE so that
		// we can send HEARTBEAT
		if (first) {
			tv.tv_sec = 0;
			tv.tv_usec = 0;
		} else if (remained >= GCB_HEARTBEAT_RATE * 1000000) {
			tv.tv_sec = GCB_HEARTBEAT_RATE;
			tv.tv_usec = 0;
		} else {
			tv.tv_sec = remained / 1000000;
			tv.tv_usec = remained % 1000000;
		}
		if (gettimeofday(&before, NULL) < 0) {
			ERROR("_CB_select: gettimeofday() failed"); 
			fready = -1;
			goto done;
		}
		snprintf(str1, sizeof(str1), "%s", _prtfds(maxfd,readfds));
		snprintf(str2, sizeof(str2), "%s", _prtfds(maxfd,writefds));
		snprintf(str3, sizeof(str3), "%s", _prtfds(maxfd,expfds));
		LOG(D_FULL, "_CB_select: calling select(rd=%s, wr=%s, ex=%s)\n",
			str1, str2, str3);
		retval = select(maxfd, readfds, writefds, expfds, &tv);
		first = 0;
		if (gettimeofday(&after, NULL) < 0) {
			ERROR("_CB_select: gettimeofday() failed"); 
			fready = -1;
			goto done;
		}
		if (timeout) {
			remained = remained -
				(after.tv_sec - before.tv_sec) * 1000000 - (after.tv_usec - before.tv_usec);
		}
		if (retval < 0 ) {
			if (errno != EINTR) {
				ERROR("_CB_select: select() failed");
			}
			fready = -1;
			goto done;
		} else if (retval > 0) {
			/* Even if select returns zero, there are cases that either:
			   - UDP msgs have been queued,
			   - connections have been queued,
			   - blocked connections have been finished
			 */
            snprintf(str1, sizeof(str1), "%s", _prtfds(maxfd,readfds));
            snprintf(str2, sizeof(str2), "%s", _prtfds(maxfd,writefds));
            snprintf(str3, sizeof(str3), "%s", _prtfds(maxfd,expfds));
			LOG(D_FULL, "_CB_select: select() returned: "
				"ready=%d, rd=%s, wr=%s, ex=%s\n", retval, str1, str2, str3);

			// handle async events that happen during selecting
			_CB_handleAsyncEvent(maxfd, readfds, writefds);
		}
		// Adjust read and write ready set
		// TODO There's a potential bug here. A large maxfd here may cause
		//   adjustFDs() to read well past the end of rdfds and wrfds,
		//   which are passed in by our caller and are only guaranteed to
		//   be as large as mxfd. We only read from them, so the only
		//   danger is if we read into unmapped memory, which seems
		//   unlikely (famous last words).
		_CB_FD_ZERO(frdfds);
		_CB_FD_ZERO(fwrfds);
		fready = adjustFDs(maxfd, rdfds, wrfds, readfds, writefds, frdfds, fwrfds);
		if (fready < 0) {
			ERROR("_CB_select: adjustFDs failed");
			fready = -1;
			goto done;
		} else if (fready > 0) {
			break;
		}
	} while ( remained > 0);

	if (rdfds) {
		memcpy(rdfds, frdfds, bytes);
	}
	if (wrfds) {
		memcpy(wrfds, fwrfds, bytes);
	}
	if (exfds) {
		memcpy(exfds, expfds, bytes);
	}
	snprintf(str1, sizeof(str1), "%s", _prtfds(maxfd,frdfds));
	snprintf(str2, sizeof(str1), "%s", _prtfds(maxfd,fwrfds));
	snprintf(str3, sizeof(str1), "%s", _prtfds(maxfd,expfds));
	LOG(D_FULL, "_CB_select: returning rd=%s, wr=%s, ex=%s\n",
		str1, str2, str3);
	_CB_lastCheck = time(NULL);
 done:
	_CB_freeFDSet( readfds );
	_CB_freeFDSet( writefds );
	_CB_freeFDSet( expfds );
	_CB_freeFDSet( frdfds );
	_CB_freeFDSet( fwrfds );

	return fready;
}
