/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"
#include "GCB_errno.h"

// Global data structure
extern int _CB_initialized;
extern struct FDinfo *_CB_fdInfo;
extern time_t _CB_lastCheck;
extern time_t _CB_lastHeartBeat;
extern int _GCB_max_no_fds;

int
GCB_sendto(int fd,
		const void *msg, size_t len, int flags,
		const struct sockaddr *to, socklen_t tolen)
{
	struct SockInfo *info;
	struct sockaddr_in sinAddr, *sto;
	struct rcver *rcver;
	int rst;
	time_t current;

	/* Since UDP send will not block, we don't need to worry about
	   blocking in here */

	if (!_CB_initialized) { _CB_initialize(); }

    if (fd < 0 || fd >= _GCB_max_no_fds || _CB_fdInfo[fd].type == NOT_DEFINED) {
		rst = sendto(fd, msg, len, flags, to, tolen);
		_GCB_errno = (rst < 0) ? GCB_ESYSCALLFAILED : 0;
		return rst;
    }

	current = time(NULL);
	if (current - _CB_lastCheck >= GCB_KEEP_EYE_RATE) {
		if (_CB_checkAsyncEvent() < 0) {
			ERROR("_CB_checkAsyncEvent() failed");
		}
	}

	// Check my type of socket
	info = _CB_fdInfo[fd].info;
	if (_CB_fdInfo[fd].type != UDP) {
        ERROR("GCB_sendto() called for non-UDP");
		errno = EBADF;
        _GCB_errno = GCB_ESOCKTNOSUPPORT;
		return -1;
	}
	LOG(D_FULL, "GCB_sendto(%d[UDP, GCB], to = %s, size=%d)\n",
			fd, _sin2str((struct sockaddr_in *) to), len);
	if (info->status == CONNECTED) {
        LOG(D_FULL, "GCB_sendto() called for CONNECTED socket\n");
		rst = sendto(fd, msg, len, flags, to, tolen);
        LOG(D_FULL, "GCB_sendto() returning %d\n", rst);
        _GCB_errno = (rst < 0) ? GCB_ESYSCALLFAILED : 0;
		return rst;
	}

	/* Bind the socket to an ephemeral address, if application has not
	 * bound it
	 */
	if (info->status == CREATED) {
		bzero(&sinAddr, sizeof(sinAddr));
		sinAddr.sin_family = AF_INET;
		if (_CB_bind(fd, (struct sockaddr *) &sinAddr, sizeof(sinAddr)) < 0) {
			return -1;
		}
	}

	/* See if the peer is registered as a receiver in our data structure */
	sto = (struct sockaddr_in *)to;
	rcver = _CB_getRcvr(info, sto->sin_addr.s_addr, sto->sin_port);
	if (rcver) {
		bzero(&sinAddr, sizeof(sinAddr));
		sinAddr.sin_family = AF_INET;
		sinAddr.sin_addr.s_addr = rcver->sinIP;
		sinAddr.sin_port = rcver->sinPort;
		rst = sendto(fd, msg, len, flags, (struct sockaddr *) &sinAddr, tolen);
        _GCB_errno = (rst < 0) ? GCB_ESYSCALLFAILED : 0;
        LOG(D_FULL, "GCB_sendto() returning %d\n", rst);
		return rst;
	}

	// GCB routing table lookup
	rst = _CB_directConnect(sto->sin_addr.s_addr);
    if (rst > 0) { // direct
        goto direct_sendto;
    } else if (rst < 0) { // error in reading the routing table
        LOG(D_BASIC, "Routing table must be in a bad shape."
                " Trying direct connection anyway...\n");
        goto direct_sendto;
    }

    // Try GCB procedure
    rst = _CB_openchannel(info, sto->sin_addr.s_addr, sto->sin_port, &sinAddr);
    if (rst < 0) {
		LOG(D_FULL, "GCB_sendto(): <fd = %d, to = %s>\n", fd, _sin2str(sto));
		LOG(D_FULL, "GCB_sendto(): _CB_openchannel() failed\n");
        return -1;
    } else if (rst == 0) {
        goto direct_sendto;
    }
	rst = sendto(fd, msg, len, flags, (struct sockaddr *)&sinAddr, sizeof(sinAddr));
    _GCB_errno = (rst < 0) ? GCB_ESYSCALLFAILED : 0;
    LOG(D_FULL, "GCB_sendto() returning %d\n", rst);
	return rst;

direct_sendto:
    rst = sendto(fd, msg, len, flags, to, tolen);
    LOG(D_FULL, "GCB_sendto() returning %d\n", rst);
    _GCB_errno = (rst < 0) ? GCB_ESYSCALLFAILED : 0;
    return rst;
}
