/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"
#include "client.h"

extern int _CB_initialized;
extern int _CB_maxfd;
extern struct FDinfo *_CB_fdInfo;
extern time_t _CB_lastCheck;
extern time_t _CB_lastHeartBeat;
extern time_t _CB_lastFlush;

// I will leave unpleasant literals in this file at the moment because this is one of
// the places that I have introduced many errors
static int
unserializeConns(struct SockInfo *sock, char *str)
{
	char *token;
	struct conn *con = NULL;
	struct sockaddr_in sinAddr;
	int working = 0;

	while ((token = strtok(NULL, "|*"))) {
		if (*token != '<') {
			con = (struct conn *)malloc(sizeof(struct conn));
			if (!con) { ABORT("malloc failed"); }
			con->fd = atoi(token);
			_CB_fdInfo[con->fd].type = HIDDEN;
			working = 1;
		} else {
			if (!working) {
				ERROR("unserializeConns: <ip:port> at unexpected time");
				return -1;
			}
			if (_str2sin(token, &sinAddr)) {
				ERROR("unserializeConns: failed to parse <ip:port> of conn, "
					  "token: '%s'", token);
				return -1;
			}
			con->ip = sinAddr.sin_addr.s_addr;
			con->port = sinAddr.sin_port;
			con->next = NULL;
			if (!sock->head) {
				sock->head = sock->tail = con;
			} else {
				sock->tail->next = con;
				sock->tail = con;
			}
			working = 0;
		}
	}
	if (working) {
		ERROR("unserializeConns: incomplete conns");
		return -1;
	}
	return 0;
}

/* Serialization of GCB socket
   Depending on whether 'fd' is owner or not, serialization takes one of two forms:
   + "owner#manager#type#<cip:cport>#PEER_ADDR#active#status#<ip:port>#CONNECTION_QUEUE"
   	 - PEER_ADDR: empty or "<ip:port>"
     - CONNECTION_QUEUE: empty or "fd|<ip:port>*fd|<ip:port>*...*fd|<ip:port>"
   + "fd#!owner"
 */
 /* TODO: Potential buffer overflow. This function should take the length of str and limit its output.  (Even better, possibly return a new buffer of the appropriate length.) */
int
_CB_serializeSock(int fd, struct SockInfo *sock, GCBString str)
{
	struct conn *cons;
	char buf[128];

	if (sock->owner != fd) {
		snprintf(buf, sizeof(buf), "%d#!%d", fd, sock->owner);
		GCB_string_appendGP(str, buf);
		return 0;
	}

	// owner & manager
	snprintf(buf, sizeof(buf), "%d#%d#", sock->owner, sock->manager);
	GCB_string_appendGP(str, buf);


	// type
	if (_CB_fdInfo[fd].type == TCP) {
		GCB_string_appendGP(str, "T#");
	} else if (_CB_fdInfo[fd].type == UDP) {
		GCB_string_appendGP(str, "U#");
	} else {
		return -1;
	}

	// <cip:cport>
	snprintf(buf, sizeof(buf), "%s#", _ipport2str(sock->cip, sock->cport));
	GCB_string_appendGP(str, buf);

	// if connected "<peer_ip:port>#", otherwise "#"
	if (sock->peer) {
		GCB_string_appendGP(str, _sin2str(sock->peer));
	}
	GCB_string_appendGP(str, "#");

	// active
	if (sock->active == ACTIVE) {
		GCB_string_appendGP(str, "a#");
	} else if (sock->active == PASSIVE) {
		GCB_string_appendGP(str, "p#");
	} else {
		GCB_string_appendGP(str, "n#");
	}

	// status
	switch(sock->status) {
		case NONE:
			// We do not support serialization of non-CBsocket
			return -1;
		case CREATED:
			GCB_string_appendGP(str, "N#");
			break;
		case BOUND:
			GCB_string_appendGP(str, "B#");
			break;
		case LISTENING:
			GCB_string_appendGP(str, "L#");
			break;
		case CONNECTED:
			GCB_string_appendGP(str, "C#");
			break;
		case PASSIVE_CONNECTING:
			GCB_string_appendGP(str, "P#");
			break;
		case CONN_FAILED:
			GCB_string_appendGP(str, "F#");
			break;
		default:
			return -1;
	}

	// <ip:port>
	GCB_string_appendGP(str, _ipport2str(sock->ip, sock->port));
	GCB_string_appendGP(str, "#");

	// connection queue: empty or "fd|<ip:port>*fd|<ip:port>*...*fd|<ip:port>"
	if (sock->head) {
		for (cons = sock->head; cons; cons = cons->next) {
			snprintf(buf, sizeof(buf), "%d|", cons->fd);
			GCB_string_appendGP(str, buf);
			GCB_string_appendGP(str, _ipport2str(cons->ip, cons->port));
			GCB_string_appendGP(str, "*");
		}
		GCB_string_trim_last(str); /* delete the last "*" */
	}
	return 0;
}

/* Serialization of GCB socket
   Depending on whether 'fd' is owner or not, serialization takes one of two forms:
   + "owner#manager#type#<cip:cport>#PEER_ADDR#active#status#<ip:port>#CONNECTION_QUEUE"
   	 - PEER_ADDR: empty or "<ip:port>"
     - CONNECTION_QUEUE: empty or "fd|<ip:port>*fd|<ip:port>*...*fd|<ip:port>"
   + "fd#!owner"
 */
struct SockInfo *
_CB_unserializeSock(char *str, int *fd)
{
	int mfd;
	struct SockInfo *sock;
	char *cur_str, *next_str;
	struct sockaddr_in sinAddr;
	int owner;

	// fd
	cur_str = str;
	next_str = strchr(cur_str, '#');
	*next_str = '\0';
	*fd = atoi(cur_str);

	cur_str = next_str + 1;
	if (*cur_str == '!') {
		// Serialization for non-owner fd
		owner = atoi(cur_str+1);
		_CB_fdInfo[*fd].type = _CB_fdInfo[owner].type;
		sock = _CB_fdInfo[*fd].info = _CB_fdInfo[owner].info;
		sock->refs++;
		if (*fd >= _CB_maxfd) {
			_CB_maxfd = *fd + 1;
		}
		return sock;
	} else {
		next_str = strchr(cur_str, '#');
		*next_str = '\0';
		mfd = atoi(cur_str);
	}

	sock = (struct SockInfo *)calloc(sizeof(struct SockInfo), 1);
	if (!sock) { ABORT("malloc failed"); }

	// owner, manager, refs
	sock->owner = *fd;
	sock->manager = mfd;
	sock->refs = 1;

	// type. T = TCP, U = UDP
	cur_str = next_str + 1;
	next_str = strchr(cur_str, '#');
	*next_str = '\0';
	if (*cur_str == 'T') {
		_CB_fdInfo[*fd].type = TCP;
	} else {
		_CB_fdInfo[*fd].type = UDP;
	}
	_CB_fdInfo[*fd].info = sock;

    if (mfd >= 0) {
        _CB_fdInfo[mfd].type = MANAGER;
        _CB_fdInfo[mfd].info = sock;
    }

	// <cip:cport>
	cur_str = next_str + 1;
	next_str = strchr(cur_str, '#');
	*next_str = '\0';
	if (_str2sin(cur_str, &sinAddr)) {
		ERROR("_CB_unserializeSock: failed to parse <cip:cport>: "
			  "cur_str is '%s'", cur_str);
		_CB_freeSockInfo(sock);
        return NULL;
	}
	sock->cip = sinAddr.sin_addr.s_addr;
	sock->cport = sinAddr.sin_port;

	// <peer>
	cur_str = next_str + 1;
	if (*cur_str == '#') {
		next_str = cur_str;
	} else {
		next_str = strchr(cur_str, '#');
		*next_str = '\0';
		if (_str2sin(cur_str, &sinAddr)) {
			ERROR("_CB_unserializeSock: failed to parse <peer>: "
				  "cur_str is '%s'", cur_str);
		    _CB_freeSockInfo(sock);
            return NULL;
		}
		sock->peer = malloc(sizeof(struct sockaddr_in));
        if (!(sock->peer)) { ABORT("malloc failed"); }
		sock->peer->sin_addr.s_addr = sinAddr.sin_addr.s_addr;
		sock->peer->sin_port= sinAddr.sin_port;
	}

	// active
	cur_str = next_str + 1;
	next_str = strchr(cur_str, '#');
	*next_str = '\0';
	if (*cur_str == 'a') {
		sock->active = ACTIVE;
	} else {
		sock->active = PASSIVE;
	}
	
	// status
	cur_str = next_str + 1;
	next_str = strchr(cur_str, '#');
	*next_str = '\0';
	switch(*cur_str) {
		case 'N':
			sock->status = CREATED;
			break;
		case 'B':
			sock->status = BOUND;
			break;
		case 'L':
			sock->status = LISTENING;
			break;
		case 'C':
			sock->status = CONNECTED;
			break;
		case 'P':
			sock->status = PASSIVE_CONNECTING;
			break;
		case 'F':
			sock->status = CONN_FAILED;
			break;
		default:
			ERROR("_CB_unserializeSock: invalid string passed to unserialize");
		    _CB_freeSockInfo(sock);
            return NULL;
	}

	// <ip:port>
	cur_str = next_str + 1;
	next_str = strchr(cur_str, '#');
	*next_str = '\0';
	if (_str2sin(cur_str, &sinAddr)) {
		ERROR("_CB_unserializeSock: failed to parse <cip:cport>: "
			  "cur_str is '%s'", cur_str);
		_CB_freeSockInfo(sock);
        return NULL;
	}
	sock->ip = sinAddr.sin_addr.s_addr;
	sock->port = sinAddr.sin_port;

	// connection queue
	cur_str = next_str + 1;
	if (*cur_str != '\0') {
		// nonempty connection list
		if (unserializeConns(sock, cur_str) < 0) {
			ERROR("unserializeConns failed");
			_CB_freeSockInfo(sock);
			return NULL;
		}
	}

	if (*fd > mfd) {
		if (*fd >= _CB_maxfd) {
			_CB_maxfd = *fd + 1;
		}
	} else if (mfd >= _CB_maxfd) {
		_CB_maxfd = mfd + 1;
	}

	return sock;
}
