/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include <GCB.h>
#include "client.h"
#include "GCB_errno.h"

// Global data structure
extern int _CB_initialized;
extern int _CB_maxfd;
extern struct FDinfo *_CB_fdInfo;
extern time_t _CB_lastHeartBeat;

/* NOT reentrant! */
static char * 
itoa(int n)
{
	static char intbuf[22]; /* Big enough for 64-bit int */
	snprintf(intbuf, sizeof(intbuf), "%d", n);
	intbuf[sizeof(intbuf)-1] = 0; 
	return intbuf;
}

/* NOT reentrant! (as it uses the above itoa) */
static char * 
socket_domain_as_string(int domain)
{
	switch(domain) {
		case AF_INET: return "AF_INET (IPv4)";
#if defined(AF_LOCAL)
		case AF_LOCAL: return "AF_LOCAL/UNIX";
#endif
#if defined(AF_UNIX) && !defined(AF_LOCAL)
		case AF_UNIX: return "AF_UNIX";  /* == AF_LOCAL */
#endif
#if defined(AF_INET6)
		case AF_INET6: return "AF_INET6 (IPv6)";
#endif
	}
	return itoa(domain);
}

/* NOT reentrant! (as it uses the above itoa) */
static char * 
socket_type_as_string(int type) 
{
	static char intbuf[8];
	switch(type) {
		case SOCK_STREAM: return "SOCK_STREAM";
		case SOCK_DGRAM: return "SOCK_DGRAM";
	}
	return itoa(type);
}


int
GCB_socket(int domain, int type, int protocol)
{
	int fd;
	struct SockInfo *info;

	// Initialize global data structure if necessary
	if (!_CB_initialized) { _CB_initialize(); }

	// Check if the socket is in the family and type that GCB supports
	if (domain != AF_INET || (type != SOCK_STREAM && type != SOCK_DGRAM)) {
        fd = socket(domain, type, protocol);
        LOG(D_FULL, "GCB_socket(%s, %s, %d): Argument combination GCB cannot handle.  Passed directly to socket(), which returned fd %d\n", socket_domain_as_string(domain), socket_type_as_string(type), protocol, fd);
        _GCB_errno = (fd < 0) ? GCB_ESYSCALLFAILED : 0;
        return fd;
	}

	if ((fd = socket(domain, type, protocol)) < 0) {
		ERROR("GCB_socket: socket(%s, %s, %d) failed", socket_domain_as_string(domain), socket_type_as_string(type), protocol);
        _GCB_errno = GCB_ESYSCALLFAILED;
		return -1;
	}

	LOG(D_FULL, "GCB_socket(%s, %s, %d): socket returned fd %d\n", socket_domain_as_string(domain), socket_type_as_string(type), protocol, fd);

	// Create a SockInfo for this socket
	info = (struct SockInfo *) calloc(sizeof(struct SockInfo), 1);
	if (info == NULL) {
		ABORT("GCB_socket: can't create SockInfo. Out of memory (calloc failed).");
		close(fd);
		return -1;
	}
	info->owner = fd;
	info->manager = NO_MNG;
	info->refs = 1;
	info->active = (getBroker() != NULL) ? ACTIVE : PASSIVE;
	info->status = CREATED;

	// Register the sock
	_CB_fdInfo[fd].type = (type == SOCK_STREAM) ? TCP : UDP;
	_CB_fdInfo[fd].info = info;
	if (fd >= _CB_maxfd) {
		_CB_maxfd = fd + 1;
	}

    _GCB_errno = 0;
	return fd;
}
