/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

/*
This file is a simple wrapping layer that will transparently
attach the GCB library to a program that invokes the standard
libc network functions, when statically linked.

Note that this module has no effect and will not be included
in a program unless special flags are given to the linker.
Thus, you may safely ignore this if you are not interested.

To force transparent linking, give these flags to the compiler:

-Wl,--wrap,socket -Wl,--wrap,bind -Wl,--wrap,listen -Wl,--wrap,connect -Wl,--wrap,accept -Wl,--wrap,select -Wl,--wrap,recv -Wl,--wrap,read -Wl,--wrap,recvfrom -Wl,--wrap,send -Wl,--wrap,sendto -Wl,--wrap,write -Wl,--wrap,close -Wl,--wrap,dup -Wl,--wrap,dup2 -Wl,--wrap,getsockname -Wl,--wrap,getsockaddr -Wl,--wrap,getsockopt -Wl,--wrap,execv -Wl,--wrap,execve -Wl,--wrap,execl

This will only work with the GNU linker.
*/

#include "gcb_common.h"
#include "generic_socket.h"

extern int __inside_gcb_wrapper;

#define WRAPPER( name, params, args ) \
int __real_##name params ;\
int __wrap_##name params \
{\
	if(__inside_gcb_wrapper) {\
		return __real_##name args ;\
	} else {\
		int result;\
		__inside_gcb_wrapper = 1;\
		result = Generic_##name args ;\
		__inside_gcb_wrapper = 0;\
		return result;\
	}\
}

/**
 * Special case wrapper for exec*() functions.
 *
 * We can't just use the regular wrapper, since if a calling process
 * uses clone() to fork, we end up hitting the wrapper and sharing the
 * address space of the parent. We'd set the __inside_gcb_wrapper flag
 * to 1, but then never clear it, since we never return from exec*.
 * So, we have to do a special trick to clear the flag before we
 * invoke the real function, just to be safe.
 */
#define EXEC_WRAPPER( name, params, args ) \
int __real_##name params ;\
int __wrap_##name params \
{\
	if(__inside_gcb_wrapper) {\
		__inside_gcb_wrapper = 0;\
		return __real_##name args ;\
	} else {\
		int result;\
		__inside_gcb_wrapper = 1;\
		result = Generic_##name args ;\
		__inside_gcb_wrapper = 0;\
		return result;\
	}\
}


WRAPPER(socket,(int domain, int type, int protocol),(domain,type,protocol))
WRAPPER(bind,(int fd, struct sockaddr *my_addr, socklen_t addrlen),(fd,my_addr,addrlen))
WRAPPER(listen,(int fd, int backlog),(fd,backlog))
WRAPPER(connect,(int fd, const struct sockaddr *serv_addr, socklen_t addrlen),(fd,serv_addr,addrlen))
WRAPPER(accept,(int fd, struct sockaddr * peer, socklen_t *addrlen),(fd,peer,addrlen))
WRAPPER(select,(int maxfd,fd_set *readfds,fd_set *writefds,fd_set *exceptfds,struct timeval *timeout),(maxfd,readfds,writefds,exceptfds,timeout))
WRAPPER(recv,(int fd, void *buf, size_t len, int flags),(fd,buf,len,flags))
WRAPPER(read,(int fd, void *buf, size_t count),(fd,buf,count))
WRAPPER(recvfrom,(int fd,  void *buf, size_t len, int flags,struct sockaddr *from, socklen_t *fromlen),(fd,buf,len,flags,from,fromlen))
WRAPPER(send,(int fd, const void *msg, size_t len, int flags),(fd,msg,len,flags))
WRAPPER(sendto,(int fd,const void *msg, size_t len, int flags,const struct sockaddr *to, socklen_t tolen),(fd,msg,len,flags,to,tolen))
WRAPPER(write,(int fd, const void *buf, size_t count),(fd,buf,count))
WRAPPER(close,(int fd),(fd))
WRAPPER(dup,(int fd),(fd))
WRAPPER(dup2,(int oldfd, int newfd),(oldfd,newfd))
WRAPPER(getsockname,(int fd, struct sockaddr * name, socklen_t *namelen),(fd,name,namelen))
WRAPPER(getsockopt,(int fd,int level,int optname,void *opt_val,socklen_t *optlen),(fd,level,optname,opt_val,optlen))

EXEC_WRAPPER(execv,(const char *path, char *const argv[]),(path,argv))
EXEC_WRAPPER(execve,(const char *path, char *const argv[], char *const envp[]),(path,argv,envp))

/* Since execl() uses a variable number of arguments, we have to wrap this
 * function in a less clean way. However, since we end up hitting the
 * __wrap_execv() defined above, we're safe from the clone() problem.
 */
void inheritCBsockets();
int __execl(const char *path, const char *arg, ...);
int __wrap_execl(const char *path, const char *arg, ...)
{
    va_list ap;
    char *next_arg;
    char const *argv[20];
    int noargs = 0;

    argv[noargs++] = arg;

    va_start(ap, arg);
    while(1) {
        next_arg = va_arg(ap, char *);
        argv[noargs++] = next_arg;
        if (!next_arg) {
            break;
        }
    }
    va_end(ap);

    if (!__inside_gcb_wrapper) {
        inheritCBsockets();
    }
    return execv(path, (char **const)argv);
}
