/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/


/*
  This file defines getExecPath(), which tries to return the full path
  to the file that the current process is executing.  On platforms
  where this information cannot be reliably determined, it returns
  NULL.  On platforms where it is possible to collect this
  information, the full path is stored in a string allocated with
  malloc() that must be free()'ed when the caller is done with it.

  Written by Derek Wright <wright@cs.wisc.edu> 2004-05-13

  WARNING: This code was lifted directly from the Condor source tree,
  slightly modified to use the autoconf-way instead of including
  condor_common.h, converted to use GCB's error logging interface
  (ERROR()) instead of Condor's logging interface (dprintf()), and
  checked into GCB's source tree.  Yes, this isn't ideal, but the real
  solution for sharing an exact copy of the source between the two
  projects would be a totally seperate utility library that both GCB
  and Condor used as an external.  That's overkill for this 1 file,
  and I don't have time to do that right now.  So, for now, just know
  that if you modify the copy here, you should fix Condor's copy in
  src/condor_util_lib/get_exec_path.c.
*/


#include "gcb_common.h"

/*
  Each platform that supports this functionality uses a completely
  different method for finding it out, so they are implemented in
  seperate helper functions, depending on the method.  configure
  figures out what method, if any, is available, and tells us via
  config.h and a bunch of HAVE_* #define's...
*/

#if HAVE_PROC_SELF_EXE
char*
getExecPath_proc_self_exe()
{
	int rval;
	char* full_path;
	char path_buf[PATH_MAX];
	rval = readlink( "/proc/self/exe", path_buf, PATH_MAX );
	if( rval < 0 ) {

		/*
		   This error will occur when running a program from the valgrind
		   memory debugger.  For this case, compiling with valgrind.h and
		   testing RUNNING_ON_VALGRIND can prevent this error message.  But
		   this remedy is too much overhead for now.
		*/
		ERROR("getExecPath: readlink(\"/proc/self/exe\") failed");
		return NULL;
	}
	if( rval == PATH_MAX ) {
		ERROR("getExecPath: unable to find full path from /proc/self/exe");
		return NULL;
	}
		/* Oddly, readlink doesn't terminate the string for you, so
		   we've got to handle that ourselves... */
	path_buf[rval] = '\0';
	full_path = strdup( path_buf );
	return full_path;
}
#endif /* HAVE_PROC_SELF_EXE */


#if HAVE_GETEXECNAME
char*
getExecPath_getexecname()
{
	const char* name;
	char* cwd = NULL;
	char* full_path = NULL;
	int size;

	name = getexecname();
	if( name[0] == '/' ) {
		return strdup(name);
	}
		/*
		  if we're still here, we didn't get the full path, and the
		  man pages say we can prepend the output of getcwd() to find
		  the real value...
		*/ 
	cwd = getcwd( NULL, PATH_MAX );
	size = strlen(cwd) + strlen(name) + 2;
	full_path = malloc(size);
	snprintf( full_path, size, "%s/%s", cwd, name );
	free( cwd );
	cwd = NULL;
	return full_path;
}
#endif /* HAVE_GETEXECNAME */


#if defined( WIN32 )
#include <psapi.h>
char*
getExecPath_win32()
{
	char buf[PATH_MAX];
	HANDLE hProc;
	DWORD result;
	HMODULE hMod;
	DWORD cbNeeded;

	hProc = GetCurrentProcess();

	if ( ! EnumProcessModules( hProc, &hMod, sizeof(hMod), 
		&cbNeeded) )
	{
		LOG(D_BASIC, "getExecPath: EnumProcessModules() failed " "(err=%li)\n",
                GetLastError() );
		CloseHandle(hProc);
		return NULL;
	}
		
	result = GetModuleFileNameEx(
		hProc,     /* process handle */
		hMod,      /* handle to module we want to know about */
		buf,       /* buffer to receive name */
		MAX_PATH   /* size of buffer needed */
		);
	
	CloseHandle(hProc);
	
	if ( result == 0 ) {
		LOG(D_BASIC, "getExecPath: unable to find full path from "
			   "GetModuleFileNameEx() (err=%li)\n", GetLastError() );
		return NULL;
		
	} else {
		return strdup(buf);
	}
}
#endif /* defined(WIN32) */


#if HAVE_MACH_O_DYLD_H

#include <mach-o/dyld.h>
#include <crt_externs.h>

typedef int (*NSGetExecutablePathProcPtr)(char *buf, size_t *bufsize);

int
getExecPath_mach_o_NSGetExecutablePath( char* buf, size_t* bufsize )
{
	return ((NSGetExecutablePathProcPtr)NSAddressOfSymbol(NSLookupAndBindSymbol("__NSGetExecutablePath")))(buf, bufsize);
}

int
getExecPath_mach_o_NSGetArgv( char* buf, size_t* bufsize )
{
	int len = PATH_MAX;
	int argc = 0;
    char *path;
    char **ptr;

		/* _NSGetArgv() returns a pointer to the argv array.  After
		   all the args, there should be a NULL in the table.  Next
		   comes the environment, and another NULL.  After that, the
		   next entry will be the value of the path given to exec().
		*/
	ptr = (char**)(*(_NSGetArgv()));
		/* Get the value of our original argc. */
	argc = *(_NSGetArgc());
		/* skip over all the args: */
	ptr += argc;
		/* ptr should be pointing at the NULL terminating the args. */
	if( *ptr != NULL ) {
		return -1;
	} else {
			/* skip the NULL */
		ptr += 1;
	}
		/* now, skip the environment, searching for the next NULL. */
	while( *ptr != NULL ) {
		ptr += 1;
	}
		/* ptr should be pointing at the NULL terminating the env. */
	if( *ptr != NULL ) {
		return -1;
	} else { 
			/* skip the NULL */
		ptr += 1;
	}
		/* Now we're finally pointing at the thing we want */
	path = *ptr;
	len = strlen( path ) + 1;
	if( len > *bufsize ) {
		return -1;
	}
	strncpy( buf, path, len );
	*bufsize = len;
	return len;
}


char*
getExecPath_mach_o()
{
	char buf[PATH_MAX];
	char full_buf[PATH_MAX];
	char* path = NULL;
	size_t size = PATH_MAX;
	int rval = 0;

    memset( buf, '\0', PATH_MAX );
    memset( full_buf, '\0', PATH_MAX );

		// Figure out dynamically if we can use the special method or
		// if we have to do it ourselves...
	if( NSIsSymbolNameDefined("__NSGetExecutablePath") ) {
			// We've got the handy function that does all this work
			// for us, so use it.
		rval = getExecPath_mach_o_NSGetExecutablePath( buf, &size );
	    if( rval < 0 ) {
			LOG(D_BASIC, "getExecName(): NSGetExecutablePath() "
				   "returned failure (%d): %s (errno %d)\n", 
				   rval, strerror(errno), errno );
			return NULL;
	    }
	} else {
			// No handy function that does it for us, so instead we
			// use a slightly more convoluted method...
		rval = getExecPath_mach_o_NSGetArgv( buf, &size );
		if( rval < 0 ) {		
			LOG(D_BASIC, "getExecName(): NSGetArgv() "
				   "returned failure (%d): %s (errno %d)\n", 
				   rval, strerror(errno), errno );
			return NULL;
	    }
	}
		/* 
		   Either way, the resulting path might still have some
		   relative paths in it.  So, convert to a canonicalized path,
		   resolving '/./' and '/../'
		*/
	path = realpath( buf, full_buf );
	if( path ) {
		path = strdup( full_buf );
		return path;
	}
	LOG(D_BASIC, "getExecName(): realpath() returned failure on \"%s\": "
		   "%s (errno %d)\n", full_buf, strerror(errno), errno );
	return NULL;
}
#endif /* HAVE_MACH_O_DYLD_H */


/*
  Now, the public method that just invokes the right platform-specific
  helper (if there is one)
*/
char*
getExecPath( void )
{
	char* rval = NULL;
#if HAVE_PROC_SELF_EXE
	rval = getExecPath_proc_self_exe();
#elif HAVE_GETEXECNAME
	rval= getExecPath_getexecname();
#elif HAVE_MACH_O_DYLD_H
	rval = getExecPath_mach_o();
#elif defined( WIN32 )
	rval = getExecPath_win32();
#endif
	return rval;
}

