/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/


#include "gcb_common.h"

#define FD_SET_CACHE_SIZE	9

static int _CB_maxNumFds = 0;
static int _CB_fdSetSize = 0;
static fd_set *_CB_fdSetCache[FD_SET_CACHE_SIZE];
/* This points to the largest index of _CB_fdSetCache that holds a valid
 * pointer to a cached fd_set array. -1 means there are no fd_set arrays
 * cached.
 */
static int _CB_fdSetCacheIdx = -1;

void
_CB_setMaxNumFds( int num_fds )
{
	const int nfdbits = 8 * sizeof(fd_set);
	int new_fd_set_size;
	int i;

	if ( num_fds > _CB_maxNumFds ) {
		_CB_maxNumFds = num_fds;
	}

	new_fd_set_size = ( _CB_maxNumFds + (nfdbits - 1) ) / nfdbits;

	if ( _CB_fdSetSize == 0 ) {
		for ( i = 0; i < FD_SET_CACHE_SIZE; i++ ) {
			_CB_fdSetCache[i] = NULL;
		}
	}

	if ( new_fd_set_size > _CB_fdSetSize ) {
		_CB_fdSetSize = new_fd_set_size;
		for ( i = 0; i <= _CB_fdSetCacheIdx; i++ ) {
			free( _CB_fdSetCache[i] );
			_CB_fdSetCache[i] = NULL;
		}
	}
}

fd_set *
_CB_mallocFDSet()
{
	assert( _CB_fdSetCacheIdx >= -1 &&
			_CB_fdSetCacheIdx < FD_SET_CACHE_SIZE );

	if ( _CB_fdSetSize == 0 ) {
		_CB_setMaxNumFds( getdtablesize() );
	}

	if ( _CB_fdSetCacheIdx < 0 ) {
		return (fd_set *)malloc( _CB_fdSetSize * sizeof(fd_set) );
	} else {
		fd_set *ret_val = _CB_fdSetCache[_CB_fdSetCacheIdx];
		_CB_fdSetCache[_CB_fdSetCacheIdx] = NULL;
		_CB_fdSetCacheIdx--;
		return ret_val;
	}

}

void
_CB_freeFDSet( fd_set *ptr )
{
	assert( _CB_fdSetCacheIdx >= -1 &&
			_CB_fdSetCacheIdx < FD_SET_CACHE_SIZE );

	if( ptr == NULL) { return; }

	if ( _CB_fdSetCacheIdx >= FD_SET_CACHE_SIZE - 1 ) {
		free( ptr );
	} else {
		_CB_fdSetCacheIdx++;
		_CB_fdSetCache[_CB_fdSetCacheIdx] = ptr;
	}
}

void
_CB_FD_ZERO( fd_set *fds )
{
	memset( fds, 0, _CB_fdSetSize * sizeof(fd_set) );
}

char *
_prtfds(int maxfd, fd_set *fds)
{
	static char buf[50];
	int i, size;

	if (fds == NULL) {
		sprintf(buf, "NULL");
		return buf;
	}
	sprintf(buf, "<");
	for(i=0; i<maxfd; i++) {
		if (fds && FD_ISSET(i, fds)) {
			if ((size = strlen(buf)) > 40) {
				strcat(buf, "...>");
				return buf;
			}
			sprintf(&buf[strlen(buf)], "%d ", i);
		}
	}
	strcat(buf, ">");
	return buf;
}

/* Convert a string of the form "<xx.xx.xx.xx:pppp>" to a sockaddr_in  TCP */
/* (Also allow strings of the form "<hostname>:pppp>")  */
int
_str2sin( const char *addr, struct sockaddr_in *sin )
{
	char    *cur_byte;
	char    *end_string;
	int 	temp=0;
	char*	addrCpy;
	char*	string;
	char*   colon = 0;
	struct  hostent *hostptr;

	if( ! addr ) {
		return -1;
	}
	addrCpy = strdup(addr);
	string = addrCpy;
	string++;					/* skip the leading '<' */

	/* allow strings of the form "<hostname:pppp>" */
	if( !(colon = strchr(string, ':')) ) {
			// Not a valid sinful string, return failure
		free(addrCpy);
		ERROR("could not find ':' in the string passed");
		return -1;
	}
	*colon = '\0';
	if ( !_valid_ip(string, NULL) &&	// only call gethostbyname if not numbers
		((hostptr=gethostbyname(string)) != NULL && hostptr->h_addrtype==AF_INET) )
	{
			sin->sin_addr = *(struct in_addr *)(hostptr->h_addr_list[0]);
			string = colon + 1;
	}
	else
	{	
		/* parse the string in the traditional <xxx.yyy.zzz.aaa> form ... */	
		*colon = ':';
		cur_byte = (char *) &(sin->sin_addr);
		for(end_string = string; end_string != 0; ) {
			end_string = strchr(string, '.');
			if (end_string == 0) {
				end_string = strchr(string, ':');
				if (end_string) colon = end_string;
			}
			if (end_string) {
				*end_string = '\0';
				*cur_byte = atoi(string);
				cur_byte++;
				string = end_string + 1;
				*end_string = '.';
			}
		}
	}
	
	string[strlen(string) - 1] = '\0'; /* Chop off the trailing '>' */
	sin->sin_port = htons((short)atoi(string));
	sin->sin_family = AF_INET;
	string[temp-1] = '>';
	string[temp] = '\0';
	*colon = ':';
	free(addrCpy);
	return 0;
}

void
_sin2strbuf(const struct sockaddr_in *sin, char * buf)
{
	if( ! sin ) {
		buf[0] = 0;
		return;
	}
	_ipport2strbuf(sin->sin_addr.s_addr, sin->sin_port, buf);
}

char *
_sin2str(const struct sockaddr_in *sin)
{
	static  char    buf[24];
	if( ! sin ) {
		buf[0] = 0;
		return buf;
	}
	buf[0] = '<';
	buf[1] = '\0';
	_sin2strbuf(sin, &buf[1]);
	sprintf(&buf[strlen(buf)], ">");
	return buf;
}

char *
_fd2str(int sockd)
{
	struct sockaddr_in	addr;
	socklen_t addr_len;
	static char *mynull = "\0";

	addr_len = sizeof(addr);

	if (getsockname(sockd, (struct sockaddr *)&addr, &addr_len) < 0) 
		return mynull;

	return ( _sin2str( &addr ) );
}

char *
_sin2host(const struct sockaddr_in *from, char ***aliases)
{
    struct hostent  *hp;
#ifndef WIN32
	struct hostent  *gethostbyaddr();
#endif

	if( !from ) {
		// make certain from is not NULL before derefencing it
		return NULL;
	}

    if( (hp=gethostbyaddr((char *)&from->sin_addr,
                sizeof(struct in_addr), AF_INET)) == NULL ) {
		// could not find a name for this address
        return NULL;
    } else {
		// CAREFULL: we are returning a staic buffer from gethostbyaddr.
		// The caller had better use the result immediately or copy it.
		// Also note this is not thread safe.  (as are lots of things in internet.c).
		if( aliases ) {
			*aliases = hp->h_aliases;
		}
		return hp->h_name;
    }
}

void
_display_from(struct sockaddr_in  *from)
{
    struct hostent  *hp;
#ifndef WIN32
	struct hostent  *gethostbyaddr();
#endif

	if( !from ) {
		ERROR("from NULL source" );
		return;
	}

    if( (hp=gethostbyaddr((char *)&from->sin_addr,
                sizeof(struct in_addr), AF_INET)) == NULL ) {
        LOG(D_BASIC, "from (%s), port %d\n",
				inet_ntoa(from->sin_addr), ntohs(from->sin_port));
    } else {
        LOG(D_BASIC, "from %s, port %d\n", hp->h_name, ntohs(from->sin_port));
    }
}

char *
_subnet(char* host)
{
	char			subnetname[50];
	char			*subnet_ptr;
	char			*host_addr_string;
	int				subnet_length;
	struct			in_addr	in;
	uint32_t		ipaddr;

	if (_myIP(&ipaddr) < 0) {
		return strdup("");
	}

	memcpy((char *) &in,(char *)&ipaddr, sizeof(uint32_t));
	host_addr_string = inet_ntoa( in );
	if( host_addr_string ) {
		subnet_ptr = (char *) strrchr(host_addr_string, '.');
		if(subnet_ptr == NULL) {
			return strdup("");
		}
		subnet_length = subnet_ptr - host_addr_string;
		strncpy(subnetname, host_addr_string, subnet_length);
		subnetname[subnet_length] = '\0';
		return (strdup(subnetname));
	}
	return strdup("");
}

int
_same_host(const char *h1, const char *h2)
{
	struct hostent *he1, *he2;
	char cn1[50];

	if (strcmp(h1, h2) == 0) {
		return 1;
	}
	
	if ((he1 = gethostbyname(h1)) == NULL) {
		return -1;
	}

	// stash h_name before our next call to gethostbyname
	strncpy(cn1, he1->h_name, 50);

	if ((he2 = gethostbyname(h2)) == NULL) {
		return -1;
	}

	return (strcmp(cn1, he2->h_name) == 0);
}

/*
  Return TRUE if the given domain contains the given hostname, FALSE
  if not.  Origionally taken from condor_starter.V5/starter_common.C. 
  Moved here on 1/18/02 by Derek Wright.
*/
int
_in_domain(const char *host, const char *domain)
{
	const char	*ptr;

	for( ptr=host; *ptr; ptr++ ) {
		if( strcmp(ptr,domain) == 0 ) {
			return 1;
		}
	}
	return 0;
}

/*
  is_ipaddr() returns TRUE if buf is an ascii IP address (like
  "144.11.11.11") and false if not (like "cs.wisc.edu").  Allow
  wildcard "*".  If we return TRUE, and we were passed in a non-NULL 
  sin_addr, it's filled in with the integer version of the ip address. 
*/
int
_valid_ip(const char *inbuf, struct in_addr *sin_addr)
{
	int len;
	char buf[17];
	int part = 0;
	int i,j,x;
	char save_char;
	unsigned char *cur_byte = NULL;
	if( sin_addr ) {
		cur_byte = (unsigned char *) sin_addr;
	}

	len = strlen(inbuf);
	if ( len < 3 || len > 16 ) 
		return 0;	// shortest possible IP addr is "1.*" - 3 chars
	
	// copy to our local buf
	strncpy( buf, inbuf, 17 );

	// on IP addresses, wildcards only permitted at the end, 
	// i.e. 144.92.* , _not_ *.92.11
	if ( buf[0] == '*' ) 
		return 0;

	// strip off any trailing wild card or '.'
	if ( buf[len-1] == '*' || buf[len-1] == '.' ) {
		if ( buf[len-2] == '.' )
			buf[len-2] = '\0';
		else
			buf[len-1] = '\0';
	}

	// Make certain we have a valid IP address, and count the parts,
	// and fill in sin_addr
	i = 0;
	for(;;) {
		
		j = i;
		while (buf[i] >= '0' && buf[i] <= '9') i++;
		// make certain a number was here
		if ( i == j )
			return 0;	
		// now that we know there was a number, check it is between 0 & 255
		save_char = buf[i];
		buf[i] = '\0';
		x = atoi( &(buf[j]) );
		if( x < 0 || x > 255 ) {
			return 0;
		}
		if( cur_byte ) {
			*cur_byte = x;	/* save into sin_addr */
			cur_byte++;
		}
		buf[i] = save_char;

		part++;
		
		if ( buf[i] == '\0' ) 
			break;
		
		if ( buf[i] == '.' )
			i++;
		else
			return 0;

		if ( part >= 4 )
			return 0;
	}
	
	if( cur_byte ) {
		for (i=0; i < 4 - part; i++) {
			*cur_byte = (unsigned char) 255;
			cur_byte++;
		}
	}
	return 1;
}

int
_valid_sin(const char *sinful)
{
	char* tmp;
	if( !sinful ) return 0;
	if( !(sinful[0] == '<') ) return 0;
	if( !(tmp = strchr(sinful, ':')) ) return 0;
	if( !(tmp = strrchr(sinful, '>')) ) return 0;
	return 1;
}

int
_str2port(const char* addr)
{
	char *sinful, *tmp;
	int port = 0;

	if( ! (addr && _valid_sin(addr)) ) {
		return -1;
	}

	sinful = strdup( addr );
	if( (tmp = strrchr(sinful, '>')) ) {
		*tmp = '\0';
		tmp = strchr( sinful, ':' );
		if( tmp && tmp[1] ) {
			port = atoi( &tmp[1] );
		} 
	}
	free( sinful );
	return port;
}

uint32_t
_str2ip(const char* addr)
{
	char *sinful, *tmp;
	unsigned int ip = 0;
	struct in_addr sin_addr;

	if( ! (addr && _valid_sin(addr)) ) {
		return -1;
	}

	sinful = strdup( addr );
	if( (tmp = strchr(sinful, ':')) ) {
		*tmp = '\0';
		if( _valid_ip(&sinful[1], &sin_addr) ) {
			ip = sin_addr.s_addr;
		}
	} else {
		ERROR("_valid_ip(\"%s\") is true, but can't find ':'");
	}
	free( sinful );
	return ip;
}

char*
_str2ipstr(const char* addr) 
{
	char *tmp;
	static char result[50];
	char sinful[50];

	if( ! (addr && _valid_sin(addr)) ) {
		return NULL;
	}

	strncpy( sinful, addr, 50 );
	tmp = strchr( sinful, ':' );
	if( tmp ) {
		*tmp = '\0';
	} else {
		return NULL;
	}
	if( _valid_ip(&sinful[1], NULL) ) {
		strncpy( result, &sinful[1], 50 );
		return result;
	}
	return NULL;
}

char*
_str2host(const char* addr) 
{
	char *tmp;
	static char result[50];
    struct sockaddr_in sin;

	if( ! (addr && _valid_sin(addr)) ) {
		return NULL;
	}

    _str2sin( addr, &sin );
	if( (tmp = _sin2host(&sin, NULL)) ) {
		strncpy( result, tmp, 50 );
	} else {
		return NULL;
	}
	return result;
}

char *
_ip2host(unsigned int ip)
{
	struct hostent *ent;
	ent = gethostbyaddr((char *) &ip, sizeof(int), AF_INET);
	if (!ent) {
		return NULL;
	}
	return ent->h_name;
}

int
_is_loopback(uint32_t ip)
{
    return (ip == inet_addr("127.0.0.1"));
}

/* Check if the ip is in private ip address space */
/* ip: in host byte order */
int
_priv_net(uint32_t ip)
{
	return ((ip & 0xFF000000) == 0x0A000000 ||      // 10/8
			(ip & 0xFFF00000) == 0xAC100000 ||      // 172.16/12
			(ip & 0xFFFF0000) == 0xC0A80000);       // 192.168/16
}

/* Check if two ip addresses, given in network byte, are in the same network */
int
_same_net(uint32_t ipA, uint32_t ipB)
{
	unsigned char *byteA, *fA, *byteB;
	int i, index;

	fA = byteA = (char *)&ipA;
	byteB = (char *)&ipB;

	if (*fA < 128) { // A class
		index = 1;
	} else if(*fA < 192) { // B class
		index = 2;
	} else {	// C class
		index = 3;
	}

	for (i = 0; i < index; i++) {
		if (*byteA != *byteB) {
			return 0;
		}
		byteA++;
		byteB++;
	}

	return 1;
}

// ip: network-byte order
// port: network-byte order
// buf: At least 22 chars long or overrun!
void
_ipport2strbuf(const unsigned int ip, const unsigned short port, char * buf)
{
	_ip2strbuf(ip, buf);
	sprintf(&buf[strlen(buf)], ":%d", ntohs(port));
}

// ip: network-byte order
// port: network-byte order
char *
_ipport2str(const unsigned int ip, const unsigned short port)
{
	static  char    buf[24];
	buf[0] = '<';
	buf[1] = '\0';
	_ipport2strbuf(ip, port, &buf[1]);
	sprintf(&buf[strlen(buf)], ">");
	return buf;
}




// ip: IP address to convert.  Must be in network order
// buf: buffer to fill with human readable string. 
//      MUST be at least 16 chars long, otherwise
//      This function will happily walk off the end.
void 
_ip2strbuf(const uint32_t ip, char * buf)
{
	int             i;
	char    *cur_byte;
	unsigned char   this_byte;
	unsigned char * aschars = (unsigned char *)&ip;


	sprintf(buf, "%u.%u.%u.%u", aschars[0], aschars[1], aschars[2], aschars[3]);
	return;
	cur_byte = (char *) &ip;
	buf[0] = 0;
	for (i = 0; i < sizeof(ip); i++) {
		this_byte = (unsigned char) *cur_byte;
		sprintf(&buf[strlen(buf)], "%u.", this_byte);
		cur_byte++;
	}
}

// ip: in network order
char *
_ip2str(const uint32_t ip)
{
	static  char    buf[24];
	buf[0] = '<';
	buf[1] = '\0';
	_ip2strbuf(ip, &buf[1]);
	sprintf(&buf[strlen(buf)], ">");
	return buf;
}


/* Get all ip addresses of this host */
// @args:
//   - ip_arr: value-result array that caller allocates and this routine fills
//               with ip addresses of this machine
//   - size: value-result arg that caller sets with the size of 'ip_arr' and this
//           routine sets with the number of ips being returned in 'ip_arr'
// @return:	-1, if failed to find a valid ip address
//			0, if succeed
int
_all_myIP(uint32_t *ip_arr, int *size)
{
	int sockfd, len, lastlen;
	char *ptr, *buf, lastname[IFNAMSIZ], *cptr;
	struct ifconf ifc;
	struct ifreq *ifr, ifrcopy;
	struct sockaddr_in *sinptr;
    int no_ips = 0;

	// Create a socket to do ioctl
	sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (sockfd < 0) {
		ERROR("_all_myIP: socket(AF_INET, SOCK_DGRAM,0) failed");
		return -1;
	}

	// Get the list of interfaces in the system
	lastlen = 0;
	len = 100 * sizeof(struct ifreq); // initial buffer size guess
	while(1) {
		if ((buf = (char *)malloc(len)) == NULL) {
			ERROR("_all_myIP: malloc failed. Out of memory.");
			close(sockfd); return -1;
		}
		ifc.ifc_len = len;
		ifc.ifc_buf = buf;
		// The problem of ioctl is that some implementations of OS does not
		// return an error when buf is not large enough to hold the result.
		// Instead the result is truncated and success is returned. Hence
		// we try multiple ioctl with increasing buf size, until we get the
		// same length returned from two consecutive ioctl calls
		if (ioctl(sockfd, SIOCGIFCONF, &ifc) < 0) {
			if (errno != EINVAL || lastlen != 0) {
				ERROR("_all_myIP: ioctl(SIOCGIFCONF) failed");
                free(buf); close(sockfd); return -1;
			}
		} else {
			if (ifc.ifc_len == lastlen) { // success
				break;
			}
			lastlen = ifc.ifc_len;
		}
		len += 10 * sizeof(struct ifreq);
		free(buf);
	}

	// Skip interfaces not up. Return only one entry for aliases.
	lastname[0] = 0;
	for(ptr = buf; ptr < buf + ifc.ifc_len; ) {
		ifr = (struct ifreq *) ptr;
		// Figure out the size of ifreq structure
		// Difficulty is that some systems provide length field but others not
#ifdef HAVE_SOCKADDR_SA_LEN
		len = max(sizeof(struct sockaddr), ifr->ifr_addr.sa_len);
		len += sizeof(ifr->ifr_name);
#else
		len = sizeof(struct ifreq);
#endif
		// Move ptr to the next interface
		ptr += len;

		// For now, ignore non ipv4 addresses
		if (ifr->ifr_addr.sa_family != AF_INET) {
			continue;
		}

		// In solaris, alias ip has interface name associated with it
		// in the form of "primary_interface_name:number", while in BSD
		// alias ip's share the same interface name. Hence we must check
		// only the primary interface name part
		if ((cptr = strchr(ifr->ifr_name, ':')) != NULL) {
			*cptr = 0;
		}
		// Check if the interface is the same as the previous one
		if (!strncmp(ifr->ifr_name, lastname, IFNAMSIZ)) {
			continue;
		}
		memcpy(lastname, ifr->ifr_name, IFNAMSIZ);

		// Now, get the attributes of the current interface
		ifrcopy = *ifr;
		if (ioctl(sockfd, SIOCGIFFLAGS, &ifrcopy) < 0) {
			ERROR("_all_myIP: ioctl(SIOCGIFFLAGS) failed");
            free(buf); close(sockfd); return -1;
		}

		// If not up, ignore
		if ( !(ifrcopy.ifr_flags & IFF_UP) ) {
			continue;
		}

        // Now, copy the ip
        if (no_ips == *size) {
            ERROR("_all_myIP: caller did not pass enough space");
            free(buf); close(sockfd); return -1;
        }
        sinptr = (struct sockaddr_in *) &ifr->ifr_addr;
        ip_arr[no_ips++] = sinptr->sin_addr.s_addr;
    }
    *size = no_ips;
    free(buf);
	close(sockfd);
    return 0;
}

/* Get an ip address of this host, preferably public one */
// @args:	ipaddr - ip address found will be returned via this arg
//					 in network byte order
// @return:	-1, if failed to find a valid ip address
//			0, if a public ip address found
//			1, if could not find a public ip but a private ip found
int
_myIP(uint32_t *ipaddr)
{
    uint32_t ipp_arr[10];
    int size = 10;
	int result = -1;
    int i;

    if (_all_myIP(ipp_arr, &size) < 0) {
        ERROR("_myIP: _all_myIP failed");
        return -1;
    }

    // Return an IP in favor of public
    for (i=0; i<size; i++) {
        *ipaddr = ipp_arr[i];
        if (ipp_arr[i] == inet_addr("127.0.0.1")) {
            continue;
        } else if (_priv_net(ntohl(ipp_arr[i]))) {
			result = 1;
			continue;
		}
		return 0;
	}
	return result;
}


static char* _gcb_version_string = "$GCBVersion: " VERSION " $";
const char*
_gcb_version( void )
{
	return _gcb_version_string;
}


static char* _gcb_build_date_string = "$GCBBuildDate: " __DATE__ " $";
const char*
_gcb_build_date( void )
{
	return _gcb_build_date_string;
}



/*
Given a timeval, canonize it; this, ensure that tv_usec is less
than 1,000,000.  If tv_usec is greater that 1,000,000, put the
excess into tv_sec.
*/
static void canonize_timeval(struct timeval * tv) 
{
	if( ! tv ) {
		ERROR("Internal error: canonize_timeval called with null pointer.");
		return;
	}
	if( (tv->tv_usec / 1000000) > 0 ) {
		tv->tv_sec += tv->tv_usec / 1000000;
		tv->tv_usec %= 1000000;
	}
}

/*
As strcmp, but for timeval; comparing times.

if *lhs is later than *rhs, return greater than 0.
if *lhs is earlier than *rhs, return less than 0.
if *lhs is identical to *rhs, return 0.
*/
static int timevalcmp(struct timeval lhs, struct timeval rhs)
{
	canonize_timeval(&lhs);
	canonize_timeval(&rhs);

	if(lhs.tv_sec > rhs.tv_sec) { return 1; }
	if(lhs.tv_sec < rhs.tv_sec) { return -1; }
	if(lhs.tv_usec > rhs.tv_usec) { return 1; }
	if(lhs.tv_usec < rhs.tv_usec) { return -1; }
	return 0;
}


/* Subtract logical time *rhs from *lhs, putting result into *result.  */
static void timeval_subtract (struct timeval * result, struct timeval lhs, struct timeval rhs)
{
	if(timevalcmp(lhs,rhs) < 0) {
		/* timeval cannot represent negative times. */
		ERROR("Internal error: timeval_subtract LHS < RHS");
		return;
	}
	if( ! result ) {
		ERROR("Internal error: timeval_subtract called with null pointer.");
		return;
	}
	canonize_timeval(&lhs);
	canonize_timeval(&rhs);
	if( lhs.tv_usec < rhs.tv_usec) {
		lhs.tv_sec--;
		lhs.tv_usec += 1000000;
	}
	result->tv_sec = lhs.tv_sec - rhs.tv_sec;
	result->tv_usec = lhs.tv_usec - rhs.tv_usec;
}

/* Documentation in util.h */
ssize_t read_with_timeout(int fd, void * buf, size_t count, struct timeval timelimit) {
	fd_set *readfd = NULL;
	ssize_t received = 0;
	struct timeval end_time;
	gettimeofday(&end_time, 0);
	end_time.tv_sec += timelimit.tv_sec;
	end_time.tv_usec += timelimit.tv_usec;
	canonize_timeval(&end_time);

	readfd = _CB_mallocFDSet();

	while(received < count) {
		int active;
		int received_this;
		struct timeval now;
		struct timeval timeout;

		gettimeofday(&now, 0);
		if( timevalcmp(now, end_time) > 0) {
			LOG(D_BASIC, "Connection didn't send expected %d bytes within time limit of %d.%06d seconds.  Closing connection.\n", count, timelimit.tv_sec, timelimit.tv_usec);
			goto done;
		}

		timeval_subtract (&timeout, end_time, now);
		_CB_FD_ZERO(readfd);
		FD_SET(fd, readfd);

		active = select(fd+1, readfd, NULL, NULL, &timeout);
		if( active == 0 ) { /* Timed out */
			LOG(D_BASIC, "Connection didn't send expected %d bytes within time limit of %d.%06d seconds.  Closing connection.\n", count, timelimit.tv_sec, timelimit.tv_usec);
			goto done;
		} 
		if( active < 0 ) {
			ERROR("Failed to check if a socket can be read (select)");
			received = -1;
			goto done;
		}

		received_this = read(fd, buf+received, count-received);
		if( received_this < 0 ) {
			if( errno == EINTR || errno == EAGAIN ) {
				continue;
			} else {
				ERROR("Failed to read from a socket.");
				received = -1;
				goto done;
			}
		}
		if( received_this == 0 ) {
			close(fd);
			goto done;
		}
		received += received_this;
	}

 done:
	_CB_freeFDSet(readfd);

	return received;
}


int write_with_timeout(int fd, const char * buf, size_t count, struct timeval timelimit) {
	fd_set * writefd = NULL;
	ssize_t sent = 0;
	struct timeval end_time;
	gettimeofday(&end_time, 0);
	end_time.tv_sec += timelimit.tv_sec;
	end_time.tv_usec += timelimit.tv_usec;
	canonize_timeval(&end_time);

	writefd = _CB_mallocFDSet();

	while(sent < count) {
		int active;
		int sent_this;
		struct timeval now;
		struct timeval timeout;

		gettimeofday(&now, 0);
		if( timevalcmp(now, end_time) > 0) {
			LOG(D_BASIC, "Connection didn't send expected %d bytes within time limit of %d.%06d seconds.  Closing connection.\n", count, timelimit.tv_sec, timelimit.tv_usec);
			goto done;
		}


		timeval_subtract (&timeout, end_time, now);
		_CB_FD_ZERO(writefd);
		FD_SET(fd, writefd);

		active = select(fd+1, NULL, writefd, NULL, &timeout);
		if( active == 0 ) { /* Timed out */
			LOG(D_BASIC, "Connection didn't read expected %d bytes within time limit of %d.%06d seconds.  Closing connection.\n", count, timelimit.tv_sec, timelimit.tv_usec);
			goto done;
		} 
		if( active < 0 ) {
			ERROR("Failed to check if a socket can be written to (select)");
			goto done;
		}

		sent_this = write(fd, buf+sent, count-sent);
		if(sent_this < 0) {
			if( errno == EINTR || errno == EAGAIN ) {
				continue;
			} else {
				ERROR("Failed to write to a socket.");
				goto done;
			}
		}

		sent += sent_this;
	}

done:
	_CB_freeFDSet(writefd);
	if(sent == count) {
		return sent;
	}
	return -1;
}
