/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#ifndef _USE_GCB_H
#define _USE_GCB_H

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "GCB_errno.h"
#include "GCB_query.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Create a GCB socket. To be supported with connection brokering, the socket
   should be IPv4 TCP or UDP socket. Otherwise this functions just calls
   "socket" with the same arguments and passes the result.
 */
// @args:
//   domain: must be AF_INET
//   type: SOCK_STREAM for TCP or SOCK_DGRAM for UDP
//   protocol: usually 0
// @return: same as "socket"
// ERRORS:
//   GCB initialization failure: ABORT
//   socket, getsockname, calloc failure: errno set by OS
//   rudp_socket failure: errno returned by rudp_socket
//   GCB slot for the socket is not clean: ABORT
int GCB_socket(int domain, int type, int protocol);
int GCB_active(int fd, char act);
int GCB_bind(int fd, struct sockaddr *my_addr, socklen_t addrlen);
int GCB_listen(int fd, int backlog);
int GCB_connect(int fd, const struct sockaddr *serv_addr, socklen_t addrlen);
int GCB_accept(int fd, struct sockaddr * peer, socklen_t *addrlen);
int GCB_select(
		int maxfd,
		fd_set *readfds,
		fd_set *writefds,
		fd_set *exceptfds,
		struct timeval *timeout);
int GCB_recv(int fd, void *buf, size_t len, int flags);
int GCB_read(int fd, void *buf, size_t count);
int GCB_recvfrom(
		int fd,  void *buf, size_t len, int flags,
		struct sockaddr *from, socklen_t *fromlen);
int GCB_send(int fd, const void *msg, size_t len, int flags);
int GCB_sendto(
		int fd,
		const void *msg, size_t len, int flags,
		const struct sockaddr *to, socklen_t tolen);
int GCB_write(int fd, const void *buf, size_t count);
int GCB_close(int fd);
int GCB_dup(int fd);
int GCB_dup2(int oldfd, int newfd);
int GCB_getsockname(int fd, struct sockaddr * name, socklen_t *namelen);
/** 
GCB_getsockname may return the address and port of the GCB
Broker/Inagent.  Sometimes you need the real, local address and
port.  You can call getsockname, but that doesn't work if you're
using linkerwrapping to automatically translate getsockname into
GCB_getsockname.  Thus, GCB_real_getsockname.  It will only ever
return the local address and port.  Note that it may do further
checks, for example it will return failure if you hand it a
GCB internal socket.

Note that using this information defeats much of the purpose of
GCB, so do so sparingly.  You almost certainly never want to hand
this information to a different machine.

If you're looking to allow local hosts to optimize their
connections by not having to contact the GCB broker/inagent,
instead look to GCB_getdirectcontact() on the server side and
GCB_smartconnect() on the client side. (As of Nov 9, 2006, these
are plans only and do not exist.)
*/
int GCB_real_getsockname(int fd, struct sockaddr * name, socklen_t *namelen);
int GCB_getsockopt(
		int fd,
		int level,
		int optname,
		void *opt_val,
		socklen_t *optlen);
int GCB_execv(const char *path, char *const argv[]);
int GCB_execl(const char *path, const char *arg, ...);
int GCB_execve(const char *path, char *const argv [], char *const envp[]);
void GCB_Broker_down_callback_set(void (*app_callback)(void), int timeout);
void GCB_Recovery_failed_callback_set(void (*app_callback)(void));
char *GCB_strerror(int GCB_err);

/**
 * Sometimes, a GCB-enabled client might want to just do a local
 * bind() and *not* register the socket at a GCB broker. In
 * particular, outbound connections don't need to be registered at a
 * GCB broker. Unfortunately, there's no good way for GCB_bind() to
 * know this, so we have to expose another function if the caller
 * really wants to bind() an outbound connection. However, the GCB
 * client code needs to know about this, so that when we finally get
 * to GCB_connect(), we realize the socket is already bound(), etc.
 * Note, you don't *have* to use this function for bind()'ing outbound
 * sockets, but it's vastly more efficient this way.
 */
int GCB_local_bind(int fd, struct sockaddr *my_addr, socklen_t addrlen);

int GCB_broker_query(const char *broker_ip, int query, int *result);

#ifdef __cplusplus
}
#endif

#endif // #ifndef _USE_GCB_H
