/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#ifndef _GCB_COMMON_H
#define _GCB_COMMON_H

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#if STDC_HEADERS
# include <stdarg.h>
# include <stdlib.h>
# include <stddef.h>
#else
# if HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif
#if HAVE_STRING_H
# if !STDC_HEADERS && HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif
#if HAVE_STRINGS_H
# include <strings.h>
#endif
#if HAVE_INTTYPES_H
# include <inttypes.h>
#else
# if HAVE_STDINT_H
#  include <stdint.h>
# endif
#endif
#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif
#if HAVE_LIMITS_H
# include <limits.h>
#endif
#if HAVE_ERRNO_H
# include <errno.h>
#endif
#if HAVE_SYS_IOCTL_H
# include <sys/ioctl.h>
#endif

#if HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif
#if HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif
#if ! HAVE_MSG_DONTWAIT && HAVE_MSG_NONBLOCK
# define MSG_DONTWAIT MSG_NONBLOCK
# define HAVE_MSG_DONTWAIT 1
#endif

#if HAVE_SYS_SOCKIO_H
# include <sys/sockio.h>
#endif

#if HAVE_NETDB_H
# include <netdb.h>
#endif
#if HAVE_NET_IF_H
# include <net/if.h>
#endif
#if HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif

/* do we need a configure test for this? */
#include <assert.h>

#include "log.h"
#include "util.h"

#ifdef __cplusplus
extern "C" {
#endif

#define BROKER_LOCAL_PORT 65430
#define BROKER_PUBLIC_PORT 65432

// active
#define ACTIVE 'A'
#define PASSIVE 'B'
/* opcode, errcode, etc interchanged between client and CB */
#define TCP_BIND 'C'
#define UDP_BIND 'D'
#define TCP_CLIENT 'E'
#define UDP_CLIENT 'F'
#define BIND_OK 'G'
#define RELAY_SERVER 'H'
#define LISTEN 'I'
#define CONNECT 'J'
#define ACTIVE_CONNECT 'K'
#define PASSIVE_CONNECT 'L'
	/* REMOTE_REPORT is for reporting log information to another daemon.
	   It should probably not be used for anything GCB needs to function.
	   The intended use is for clients to send notes to the broker for
	   the broker to note in its logs.

	   The specific reports you can issue are listed below. The report
	   goes in the "active" field.
	*/
#define REMOTE_REPORT 'l'
#define ACK 'M'
#define CONTACT_FWD 'N'
#define NEW_SOCK 'n'
#define CONTACT 'O'
#define INIT_COMM 'P'
#define PASSIVE_INIT 'Q'
	/* Request information for the daemon. Supported queries are
	   found in GCB_query.h.  The query is put in cip in the command.
	   The reply should be of type DATA_RESPONSE with the cip
	   the same as the query and the result is stuffed into ip.
	   cip and ip are 32-bit unsigned integers (although a given
	   query could be creative and cram something else in there).  
	   Conceivably other fields could be used for more complex queries
	   or responses, but none do as of August 2007.
	*/
#define DATA_QUERY 'q'
#define ACTIVE_INIT 'R'
#define DATA_RESPONSE 'r'
#define TRIGGER_FWD 'S'
#define TRIGGER 'T'
#define DUMMY 'U'
#define DUMMY_ACK 'V'
#define TCP_DEL 'W'
#define UDP_DEL 'X'
#define HEARTBEAT 'Y'
#define NAK 'Z'

#define PEER_NOT_REGISTERED '9'


/* Things that can be set in the "active" field in REMOTE_REPORT
   
   active is a char, so we've safely got 0-127 as options. (We should probably
   make it explicitly signed or unsigned so we can safely use the whole range,
   but that's for another day.)
   
   I'm avoiding 0 and 1 as they tend to be magical and likely to be passed
   in for the wrong reasons..
*/
	/* 
		REMOTE_REPORT_PASSIVE:

		client -> broker: I was asked to do a passive connection, here's the
		result.
	*/
#define REMOTE_REPORT_PASSIVE 2

// Network functions

// Stripe a message with given arguments
// #include <CBcommon.h>
// @args:   *buf - buffer which will contain the striped msg
//          opcode, cip, cport, regAt, active, ... - values to be striped into the msg
//			all ips and ports are in network byte order
//			Note: regAt should be in host byte order
// This function should always succeed. If fail, the process will be aborted
// @return: size, in host byte order, of the striped msg, if succeed
unsigned short _CB_stripeMsg(
		char *buf,
	   	char opcode,
	   	uint32_t cip,
	   	uint16_t cport,
	   	char active,
	   	uint32_t ip,
	   	uint16_t port,
	   	uint32_t dip,
	   	uint16_t dport,
	   	char errcode);

/* Parse _CB msg which is "size" bytes long and stored in "buf" */
// #include <CBcommon.h>
// @arg:
//      opcode, cip, cport, regAt, active, dip, dport, sips, sport, errcode are set with
//      the values taken from the message.
//      Note: "buf" must point msg[2].
//		Note: regAt is passed in host byte order
// @return: 0, if succeed
//          -1, otherwise
int _CB_unstripeMsg(
		int nobytes,
		char *buf,
		char *opcode,
		uint32_t *cip,
		uint16_t *cport,
		char *active,
		uint32_t *ip,
		uint16_t *port,
		uint32_t *dip,
		uint16_t *dport,
		char *errcode);

/* Send GCB message through the management socket fd */
// @ret: size of the msg sent, if success
//       -1, if communication error occurred
int _CBsend(
		int fd,
		char opcode,
		uint32_t cip,
		uint16_t cport,
		char active,
		uint32_t ip,
		uint16_t port,
		uint32_t dip,
		uint16_t dport,
		char errcode);

/* Receive GCB message through the management socket fd */
// @ret: size of the msg received, if success
//       0, if the connection was closed
//       -1, if communication error occurred or _CB_unstripeMsg failed
int _CBrecv(
		int fd,
		char *opcode,
		uint32_t *cip,
		uint16_t *cport,
		char *active,
		uint32_t *ip,
		uint16_t *port,
		uint32_t *dip,
		uint16_t *dport,
		char *errcode);

/* Set the maximum file descriptor that _CB_mallocFDSet()'s fd_sets
 * should be able to accommodate. The default is getdtablesize(), but
 * we can inherit higher-numbered file descriptors from our parent
 * process.
 */
void _CB_setMaxNumFds( int num_fds );

/* malloc()s and returns a chunk of memory large enough to hold
 * the max file descriptor we can get from the OS for use with select().
 * The caller is responsible for deallocating the memory with
 * _CB_freeFDSet().
 */
fd_set *_CB_mallocFDSet();

/* Free a chunk of memory given out by _CB_mallocFDSet().
 */
void _CB_freeFDSet( fd_set *ptr );

/* Zero out the entire bit-array returned by a call to _CB_mxallocFDSet().
 */
void _CB_FD_ZERO( fd_set *fds );

#ifdef __cplusplus
}
#endif

#endif /* _GCB_COMMON_H */
