/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#ifndef _Generic_LOG_H
#define _Generic_LOG_H

#ifdef __cplusplus
extern "C" {
#endif

extern char *Generic_log_srcfile;
extern int Generic_log_lineno;

#define D_BASIC (1 << 0)
#define D_FULL (1 << 13)

#define LOG \
	Generic_log_srcfile = __FILE__;\
	Generic_log_lineno = __LINE__;\
	Generic_log
#define ERROR \
	Generic_log_srcfile = __FILE__;\
	Generic_log_lineno = __LINE__;\
	Generic_error
#define ABORT \
    Generic_log_srcfile = __FILE__;\
    Generic_log_lineno = __LINE__;\
    Generic_abort

#define JOB_LOG \
    Generic_log_srcfile = __FILE__;\
    Generic_log_lineno = __LINE__;\
	Generic_job_log

// Set log file and log level to 'path' and 'level', respectively
// Before this routine is called, every log will be printed to stdout
void Generic_logInit(char *path, int level, int maxsize);

// With this function, application let Generic use its own function to log.
// The only application that is expected to use is Condor at this moment.
// Since Generic_log, Generic_error, and Generic_abort will eventuall call the function
// passed by the application by calling this, it should still call LOG, ERROR,
// and ABORT
void Generic_set_log_va(void(*app_log_va)(int level, char *fmt, va_list args));

void Generic_stop_logging();
void Generic_resume_logging();

// Print debug message of debug level 'level'. If 'level' is higher
// than 'Generic_debug_level', the msg will not be logged
void Generic_log(int level, char *fmt, ...);

// Print error message to the log file and exit
// The error message passed will be prefixed with pid, file, and line where error occured
// and also postfixed with strerror(errno)
void Generic_abort(char *fmt, ...);

// Same as Generic_abort without exiting
void Generic_error(char *fmt, ...);

void Generic_job_log(char *msg);


/***************************************************************************
 *
 *  The new logging system
 *
 ***************************************************************************/

/* Values for severity argument to gcb_log */

/* Debugging chatter.  Probably only useful for desperate Condor
 * developers grasping at straws to debug a problem.*/
#define GLS_SUPERVERBOSE 32

/* Normal functionality, reports of all-okay. Examples would
 * include GCB_connect() succeeding.
 */
#define GLS_OK 64

/* Normal functionality, but noteworthy. Examples would include
 * one-off pieces of information, like that you've selected a
 * broker. */
#define GLS_NOTEABLE 96

/* Something has gone wrong, but in a potentially expected and
 * recoverable way. This would include dropped connections,
 * gibberish commands, and down servers.
 */
#define GLS_RECOVERABLE_ERROR 128

/* Something has gone wrong,
but in a potentially expected and
 * recoverable way. */
#define GLS_SERIOUS_ERROR 160

/* Serious, typically unrecoverable and unlikely errors.
 * Includes malloc() and socket() failing (likely because of
 * resource exhaustion).  Should be rare.  Should probably be
 * investigated, but might just be bad luck (like another user
 * forkbombed the machine).
 */
#define GLS_FATAL 192

/* Things have gone very badly.  Implausible errors, "can't
 * happen" code, and the like.  Should be very rare and every
 * single instance should be investigated.
 */
#define GLS_CATASTROPHE 224

/* Use this for the port for any of the port arguments to
 * gcb_log if a port is not applicable at this point.
 * This way we can distinguish between "0" and "N/A".  You should
 * NEVER use this anywhere besides gcb_log functions.
 */
#define GL_PORT_NA 131070

/* Use this for an IP address for any of the IP address arguments
 * to gcb_log if an IP address is not applicable at this
 * point. This way we can distinguish between 0.0.0.0 and "N/A".
 *
 * The given IP (10.31.141.156) is a private reversed address in
 * RFC1918 space.  It was randomly chosen to minimize the chance
 * of actual use.  You should NEVER use this anywhere besides
 * gcb_log functions.
 */
uint32_t GL_IP_NA();


/*

- unsigned int msgid - Unique identifier for a log message.  This
  should be absolutely unique. When old messages are deleted, the
  macro and first argument should be left somewhere in the source
  to permanently "retire" the code.  Please update this following
  ranges of available IDs as you use them.  Currently available
  IDs: 1-MAXINT.

- const char * filename - Automatically provided by the clog and
  friends macros.  The filename will be truncated to the basename
  before posting.

- int lineno - Automatically provided by the clog and friends
  macros. 

- const char * subsystem - Automatically provided by the wrapper
  macros.  Should be "GCB broker", "GCB relay server", or "GCB
  client library".

- unsigned char severity - How important is it that someone
  notice this?  Use the macros abover (GLS_*) instead of picking
  your own number.

- const char * functionname -  What function is calling
  gcb_log?  NULL is allowable but discouraged.
 
- const char * gcb_task,
	- What logical task is GCB engaging in?  Typically this would
	  be the function that GCB is replacing, like socket, listen,
	  bind, etc.  We can't always be sure what task we're in
	  (e.g. _CB_openchannel is called by both connect and
	  sendto).  Ideally we would be able to provide the
	  information anyway.
	- Perhaps in the long run all of the GCB_* functions should
	  set a global variable with their own name, ensuring this
	  information is always available (and possibly automatically
	  passed in by the gcb_log macro)
	- NULL is allowable, but discouraged.
 
- const char * msg - Human readable message indicating the
  problem.  Doesn't need to include information passed directly
  as arguments below.  NULL is allowable, but strongtly
  discouraged.
 
- uint32_t my_ip_addr, uint32_t my_port - Who do I think I am
  (for real, no GCB trickery)?  For the broker/relay server, the
  real ip/port of a "server" (someone using 'socket/bind/listen')
  using the broker/relay.  Use GL_IP_NA or GL_PORT_NA as
  appropriate if a given argument is not applicable here.  port
  is 32-bits so that GL_PORT_NA can be an invalid port number.

- uint32_t broker_ip_addr, uint32_t relay_server_port - What broker
  and relay server are we using?  (In the case of the broker/rs,
  what is my own IP/port?) Use GL_IP_NA or GL_PORT_NA as
  appropriate if a given argument is not applicable here.  port
  is 32-bits so that GL_PORT_NA can be an invalid port number.
 
- uint32 remote_ip_addr, uint32 remote_port - Who am I trying to
  talk to?  This would serv_addr in connect(), or the remote side
  you've accept()ed a connection from.  For the broker/relay
  server, the ip of the client trying to connect to a proxied
  server.  Use GL_IP_NA or GL_PORT_NA as appropriate if a given
  argument is not applicable here.  port is 32-bits so that
  GL_PORT_NA can be an invalid port number.
 
For maximum compatbility with web servers, the total message
length should stay below 2,048 bytes.  This includes the
relative path as well as pretty much all of the arguments to the
log call.  So the msg should probably be "short", less than,
say, 1,000 bytes.


clog: client logging
blog: broker logging
rlog: relay_server logging
*/

/* Will be strdup()ed. Can be NULL to clear. */
void gcb_log_set_task(const char * task);
const char * gcb_log_get_task();

/* To clear, set to GL_IP_NA() */
void gcb_log_set_my_ip_addr(uint32_t addr);
uint32_t gcb_log_get_my_ip_addr();

/* To clear, set to GL_IP_NA() */
void gcb_log_set_broker_ip_addr(uint32_t addr);
uint32_t gcb_log_get_broker_ip_addr();

#define clog(msgid, severity, functionname, msg, my_port, relay_server_port, remote_ip_addr, remote_port) \
	gcb_log_impl(msgid, __FILE__, __LINE__, "GCB client library", severity, functionname, gcb_log_get_task(), msg, gcb_log_get_my_ip_addr(), my_port, gcb_log-get_broker_ip_addr(), relay_server_port, remote_ip_addr, remote_port)

#define rlog(msgid, severity, functionname, gcb_task, msg, my_ip_addr, my_port, broker_ip_addr, relay_server_port, remote_ip_addr, remote_port) \
	gcb_log_impl(msgid, __FILE__, __LINE__, "GCB relay server", severity, functionname, gcb_task, msg, my_ip_addr, my_port, broker_ip_addr, relay_server_port, remote_ip_addr, remote_port)

#define blog(msgid, severity, functionname, gcb_task, msg, my_ip_addr, my_port, broker_ip_addr, relay_server_port, remote_ip_addr, remote_port) \
	gcb_log_impl(msgid, __FILE__, __LINE__, "GCB broker", severity, functionname, gcb_task, msg, my_ip_addr, my_port, broker_ip_addr, relay_server_port, remote_ip_addr, remote_port)

void
gcb_log_impl(
	unsigned int msgid,

	const char * filename,
	int lineno,

		const char * subsystem,

	unsigned char severity,

	const char * functionname,
	const char * gcb_task,

	const char * msg,

	uint32_t my_ip_addr,
	uint32_t my_port, /* local or GCB */

	uint32_t broker_ip_addr, /* Port is fixed BROKER_PUBLIC_PORT=65432 */
	uint32_t relay_server_port, /* IP addr is broker_ip_addr */

	uint32_t remote_ip_addr,
	uint32_t remote_port
	);

#ifdef __cplusplus
}
#endif

#endif
