/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#ifndef __SS_UTIL_H
#define __SS_UTIL_H

#include "gcb_common.h"

#ifdef __cplusplus
extern "C" {
#endif

// prt_fds
char *_prtfds(int maxfd, fd_set *fds);
// string_to_sin
int _str2sin(const char *addr, struct sockaddr_in *sin);
/** \brief Fill buffer with human readable string representation of a
           sockaddr_in  (the IP address and port) "192.168.0.1:8080"

	\param sin 
	\param buf At least 22 chars long or overrun!
*/
void _sin2strbuf(const struct sockaddr_in *sin, char * buf);
/** sin_to_string. Returns pointer to internal static buffer,
thus not reentrant.  
*/
char *_sin2str(const struct sockaddr_in *sin);
// sock_to_string
char *_fd2str(int sockd);
// sin_to_hostname
char *_sin2host(const struct sockaddr_in *from, char ***aliases);
// display_from
void _display_from(struct sockaddr_in *from);
// calc_subnet_name
char *_subnet(char* host);
// same_host
int _same_host(const char *h1, const char *h2);
// host_in_domain
int _in_domain( const char *host, const char *domain);
// is_ipaddr
int _valid_ip(const char *inbuf, struct in_addr *sin_addr);
// is_valid_sinful
int _valid_sin(const char *sinful);
// string_to_port
int _str2port(const char* addr);
// string_to_ip
uint32_t _str2ip(const char* addr);
// string_to_ipstr
char *_str2ipstr(const char* addr);
// string_to_hostname
char *_str2host(const char* addr);
// ip_to_name
char *_ip2host(unsigned int ip);
// is_priv_net
int _priv_net(uint32_t ip);
// in_same_net
int _same_net(uint32_t ipA, uint32_t ipB);

/** \brief Fill buffer with human readable string representation of an
           IP address and port. "192.168.0.1:8080"

	\param ip network-byte order
	\param port network-byte order
	\param buf At least 22 chars long or overrun!
*/
void _ipport2strbuf(const unsigned int ip, const unsigned short port, char * buf);
/** ipport_to_string.  Returns a static internal buffer, so not reentrant.
wraps output in angle brackets: "<192.168.0.1:8080>"
*/
char *_ipport2str(const unsigned int ip, const unsigned short port);
/** \brief Fill buffer with human readable string representation 
           of an IP address. "192.168.0.1"

    \param ip  expected to be in network order.
    \param buf Filled with resulting string.  Must be at least
	           16 chars long or overrun may occur.
*/
void _ip2strbuf(const uint32_t ip, char * buf);
/** ip_to_string.  Returns a static internal buffer, so not reentrant.
wraps output in angle brackets: "<192.168.0.1>"
*/
char *_ip2str(const uint32_t ip);
/** Get all ip addresses of this host 

 @args:
   - ip_arr: value-result array that caller allocates and this routine fills
               with ip addresses of this machine
   - size: value-result arg that caller sets with the size of 'ip_arr' and this
           routine sets with the number of ips being returned in 'ip_arr'
 @return: -1, if failed to find a valid ip address
          0, if succeed
*/
int _all_myIP(uint32_t *ip_arr, int *size);
// getIPaddr
int _myIP(uint32_t *ipaddr);

const char* _gcb_version( void );
const char* _gcb_build_date( void );



/*
Designed to behave similarly to read(), with several changes:

- It automatically retries when read() fails with EINTR or
  EAGAIN.   Callers should _not_ implement their own retry for
  this case.  So long as there isn't an error or timeout, "count"
  bytes _will_ be read and returned.

- If "timelimit" passes without receiving all "count" bytes,
  further reading is aborted and the number of bytes read so far
  are returned.

Return codes:

- == count : if the return code is the same as "count", then all
  "count" bytes were succesfully read.  The caller can safely
  continue using the "fd".

- >= 0 && < count : Either "timelimit" passed without receiving
  all "count" bytes or the connection was closed.  The returned
  value is how many bytes in "buf" were successfully read.  The
  caller should not continue using "fd", instead just close()ing
  it.

- < 0 : An error occured in select() or read().  errno should be
  appropriately set.  "buf" may or may not contain some
  partial data read off of "fd", but that information is not
  available to the caller.  The handle should be close()d.

This isn't as good as proper non-blocking IO, but it's
easier to add to existing code.
*/
ssize_t read_with_timeout(int fd, void * buf, size_t count, struct timeval timelimit);

int write_with_timeout(int fd, const char * buf, size_t bytes, struct timeval timelimit);

#ifdef __cplusplus
}
#endif

#endif
