/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB_query.h"
#include "get_exec_path.h"
#include <iostream>
using namespace std;
#include "broker.h"
#include <sys/wait.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/stat.h>

#define FLUSH_INTERVAL 90

// max length of string representations
#define MAX_RELAYER_STR 100
#define MAX_CLIENT_STR 300 + MAX_RELAYER_STR

int _GCB_max_no_fds;

static int
shouldRelay()
{
    char *answer;
    answer = getenv("GCB_OPERATION_MODE");
    return (!answer || strcasecmp(answer, "relay")) ? 0 : 1;
}

Client::Client(int type, uint32_t cip, uint16_t cport, char active,
			uint32_t ip, uint16_t port, bool listen, Relayer *relayer)
{
	this->type = type;
	this->cip = cip;
	this->cport = cport;
	this->active = active;
	this->ip = ip;
	this->port = port;
	listener = listen;
	this->relayer = relayer;
	pre = next = NULL;
}

char *
Client::tostr()
{
	static char str[MAX_CLIENT_STR];

	if (type == TCP) {
		sprintf(str, "TCP:");
	} else {
		sprintf(str, "UDP:");
	}
	if (listener) {
		strcat(str, "Listen:");
	} else {
		strcat(str, "Bind:");
	}
	if (active == ACTIVE) {
		strcat(str, "Active:");
	} else {
		strcat(str, "Passive:");
	}
	sprintf(&str[strlen(str)], "canon=%s:", _ipport2str(cip, cport));
	sprintf(&str[strlen(str)], "real=%s:", _ipport2str(ip, port));
	if (relayer) {
		sprintf(&str[strlen(str)], "relayer=<%s>", relayer->tostr()); 
	} else {
		strcat(str, "no relayer");
	}
	return str;
}

void
Client::ip_broker(char * result)
{
	strncpy(result, _ipport2str(cip, cport), 24);
	result[23] = 0;
}

void
Client::ip_client(char * result)
{
	strncpy(result, _ipport2str(ip, port), 24);
	result[23] = 0;
}

Relayer::Relayer(char *ipaddr, char *path)
{
	// 1. create pipes
	int fd1[2];
	if ( pipe(fd1) < 0) {
		ABORT("Failed to create RelayServer: pipe() failed, errno %d (%s)",
			  errno, strerror(errno));
	}
	int fd2[2];
	if ( pipe(fd2) < 0) {
		ABORT("Failed to create RelayServer: pipe() failed, errno %d (%s)",
			  errno, strerror(errno));
	}

	// 2. create a listening TCP socket for GCB command exchange with clients
	int sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		ABORT("Failed to create RelayServer: socket() failed, errno %d (%s)",
			  errno, strerror(errno));
	}
	struct sockaddr_in sinaddr;
	bzero(&sinaddr, sizeof(struct sockaddr_in));
	sinaddr.sin_family = AF_INET;
	if (!inet_aton(ipaddr, &sinaddr.sin_addr)) {
		ABORT("Failed to create RelayServer: inet_aton() failed");
	}
	if (bind(sock, (struct sockaddr *)&sinaddr, sizeof(sinaddr)) < 0) {
		ABORT("Failed to create RelayServer: bind() failed, errno %d (%s)",
			  errno, strerror(errno));
	}
	if (listen(sock, 100) < 0) {
		ABORT("Failed to create RelayServer: listen() failed, errno %d (%s)",
			  errno, strerror(errno));
	}

	// 3. get ip:port
	socklen_t addrlen = sizeof(struct sockaddr_in);
	if (getsockname(sock, (struct sockaddr *)&sinaddr, &addrlen) < 0) {
		ABORT("Failed to create RelayServer: getsockname() failed, errno %d (%s)",
			  errno, strerror(errno));
	}

	// 4. fork
	pid_t pno = fork();
	if (pno < 0) {
		ABORT("Failed to create RelayServer: fork() failed, errno %d (%s)",
			  errno, strerror(errno));
	} else if (pno == 0) { // child
		close(fd1[0]);
		close(fd2[1]);
		dup2(fd2[0], 0);
		dup2(fd1[1], 1);
		dup2(sock, 2);
		for (int i=3; i < _GCB_max_no_fds; i++) {
			close(i);
		}

		// 5. exec RelayServer
		char *argv[6];
		char arg1[5], arg3[5], arg4[10];
		sprintf(arg1, "-i");
		argv[0] = path;
		argv[1] = arg1;
		argv[2] = ipaddr;
		sprintf(arg3, "-p");
		sprintf(arg4, "%d", ntohs(sinaddr.sin_port));
		argv[3] = arg3;
		argv[4] = arg4;
		argv[5] = NULL;
		execv(path, argv);
		ABORT("Failed to create RelayServer: execv() failed, errno %d (%s)",
			  errno, strerror(errno));
	}

	// parent
	close(fd1[1]);
	close(fd2[0]);
	rdfd = fd1[0];
	wrfd = fd2[1];
	mip = sinaddr.sin_addr.s_addr;
	mport = sinaddr.sin_port;
	pid = pno;
	noClients = 0;

	char* max_client_str = getenv("GCB_MAX_CLIENTS_PER_RELAY_SERVER");
	if (max_client_str) {
		maxClients = atol(max_client_str);
	} else { 
		maxClients = GCB_DEFAULT_MAX_CLIENTS;
	}

	pre = next = NULL;
	close(sock);
}

Relayer::~Relayer()
{
	close(rdfd);
	close(wrfd);
	if (kill(pid, SIGKILL) < 0) {
		if (errno != ESRCH) {
			ABORT("Failed to kill RelayServer (pid: %ld), errno %d (%s)",
				  (long)pid, errno, strerror(errno));
		}
	}
}

char *
Relayer::tostr()
{
	static char str[MAX_RELAYER_STR];

	sprintf(str, "rdfd:%d, wrfd:%d, pid:%ld, mngAddr: %s, noClients:%d",
			rdfd, wrfd, (long)pid, _ipport2str(mip, mport), noClients);
	return str;
}

bool
Relayer::moreClient()
{
	return (noClients < maxClients);
}

void
Relayer::addClient()
{
	noClients++;
}

void
Relayer::deleteClient()
{
	noClients--;
}

Broker::Broker(char *ipaddr, char *relay_path)
{
	LOG(D_BASIC, "++++++++++++++++++++++++++++++\n");
	LOG(D_BASIC, "+ STARTING Broker (pid: %ld)\n", (long)getpid());
	LOG(D_BASIC, "+ %s\n", _gcb_version());
	LOG(D_BASIC, "+ %s\n", _gcb_build_date());

    // ipaddr must be in xxx.xxx.xxx.xxx Can't be bigger than 20 (to be safe)
	snprintf(_canonAddr, 20, "%s", ipaddr);;
	struct in_addr inaddr;
	if (!inet_aton(ipaddr, &inaddr)) {
		ABORT("inet_aton with Broker's ip failed");
	}
	_mngIP = inaddr.s_addr;

	struct sockaddr_in sinaddr;
	socklen_t namelen = sizeof(struct sockaddr_in); 

	_local = socket(AF_INET, SOCK_STREAM, 0);
	if (_local < 0) {
		ABORT("socket() failed on local command socket, errno %d (%s)",
			  errno, strerror(errno));
	}
	int on = 1;
	if (setsockopt(_local, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) {
		ABORT("setsockopt() failed on local command socket: errno %d (%s)",
			  errno, strerror(errno));
	}
    bzero(&sinaddr, sizeof(struct sockaddr_in));
	sinaddr.sin_family = AF_INET;
	sinaddr.sin_addr.s_addr = _mngIP;
	sinaddr.sin_port = htons(BROKER_LOCAL_PORT);
	if (bind(_local, (struct sockaddr *)&sinaddr, namelen) < 0) {
		ABORT("bind(%s:%d) failed on local command socket: errno %d (%s)",
			  inet_ntoa(sinaddr.sin_addr), BROKER_LOCAL_PORT, errno,
			  strerror(errno));
	}
    // Two years ago, backlog 5 means a few hundreds. Now it seems that the number is
    // correct.
	if (listen(_local, 200) < 0) {
		ABORT("listen() failed on local command socket: errno %d (%s)",
			  errno, strerror(errno));
	}
	LOG(D_BASIC, "+ Local command socket listening at %s:%d\n",
		inet_ntoa(sinaddr.sin_addr), BROKER_LOCAL_PORT);

	_public = socket(AF_INET, SOCK_STREAM, 0);
	if (_public < 0) {
		ABORT("socket() failed on public command socket: errno %d (%s)",
			  errno, strerror(errno));
	}
	if (setsockopt(_public, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) {
		ABORT("setsockopt() failed on public command socket: errno %d (%s)",
			  errno, strerror(errno));
	}
    bzero(&sinaddr, sizeof(struct sockaddr_in));
	sinaddr.sin_family = AF_INET;
	sinaddr.sin_addr.s_addr = _mngIP;
	sinaddr.sin_port = htons(BROKER_PUBLIC_PORT);
	if (bind(_public, (struct sockaddr *)&sinaddr, namelen) < 0) {
		ABORT("bind(%s:%d) failed on public command socket: errno %d (%s)",
			  inet_ntoa(sinaddr.sin_addr), BROKER_PUBLIC_PORT, errno,
			  strerror(errno));
	}
	if (listen(_public, 200) < 0) {
		ABORT("listen() failed on public command socket: errno %d (%s)",
			  errno, strerror(errno));
	}
	LOG(D_BASIC, "+ Public command socket listening at %s:%d\n",
		inet_ntoa(sinaddr.sin_addr), BROKER_PUBLIC_PORT);

	for (int i=0; i<BRK_CLNT_HASH_SIZE; i++) {
		_clients[i] = NULL;
	}

    // get the max # of fds per process that OS supports
    _GCB_max_no_fds = getdtablesize();
    _relayerFDs = (Relayer **) malloc(sizeof(Relayer *) * _GCB_max_no_fds);
    if (!_relayerFDs) { ABORT("malloc failed"); }
	for (int i = 0; i < _GCB_max_no_fds; i++) {
		_relayerFDs[i] = NULL;
	}

    if (strlen(relay_path) >= PATH_MAX) {
        ABORT("The path for RelayServer is too long");
    }
    _relayer_path = strdup(relay_path);

	LOG(D_BASIC, "+ Using relay_server: %s\n", _relayer_path);
	LOG(D_BASIC, "++++++++++++++++++++++++++++++\n");

	_noRelayers = 0;

	char* max_relay_str = getenv("GCB_MAX_RELAY_SERVERS");
	if (max_relay_str) {
		_maxRelayers = atol(max_relay_str);
	} else { 
		_maxRelayers = GCB_DEFAULT_MAX_RELAY_SERVERS;
	}

	// Create the initial RelayServer
	// we touch _relayers in createRelayer(), so we must initialize it first
	_relayers = NULL;  
	_relayers = createRelayer();
}

Relayer *
Broker::createRelayer()
{
	// 1. new Relayer
	Relayer *relayer = new Relayer(_canonAddr, _relayer_path);
	LOG( D_FULL, " New RelayServer[%s] created\n", relayer->tostr() );

	// 2. register the Relayer at _relayerFDs[pipe[0]]
	_relayerFDs[relayer->rdfd] = relayer;

	// 3. register the Relayer at _relayers
	relayer->next = _relayers;
	if (relayer->next) {
		relayer->next->pre = relayer;
	}
	_relayers = relayer;

	// 4. increment the # of RelayServer
	_noRelayers++;

	return relayer;
}

void
Broker::deleteRelayer(int fd)
{
	// find the relayer
	Relayer *relayer = _relayerFDs[fd];
	LOG(D_FULL, "Deleting RelayServer[%s]...\n", relayer->tostr());

	// Delete clients the RelayServer is servicing
	for (int i = 0; i < BRK_CLNT_HASH_SIZE; i++) {
		Client *client = _clients[i];
		while (client) {
			if (client->relayer == relayer) {
				Client *ptr = client;
				client = client->next;
				LOG(D_FULL, "\tdeleting client<%s>...\n", ptr->tostr());
				if (ptr->pre) {
					ptr->pre->next = client;
				} else {
					_clients[i] = client;
				}
				if (client) {
					client->pre = ptr->pre;
				}
				delete ptr;
			} else {
				client = client->next;
			}
		}
	}

	// delete the relayer from _relayers
	if (relayer->pre) {
		relayer->pre->next = relayer->next;
	} else {
		_relayers = relayer->next;
	}
	if (relayer->next) {
		relayer->next->pre = relayer->pre;
	}

	// unregister the relayer from _relayerFDs
	_relayerFDs[fd] = NULL;

	// decrease the number of RelayServer
	_noRelayers--;

	delete relayer;
}

void
Broker::deleteClient(Client *client)
{
	Relayer *relayer = client->relayer;
	if (relayer) {
		// Decrease the number of clients being served by the RelayServer
		relayer->deleteClient();
	}

	// unregister from _clients[]
	int index = (client->cip + client->cport) % BRK_CLNT_HASH_SIZE;
	if (client->pre) {
		client->pre->next = client->next;
	} else {
		_clients[index] = client->next;
	}
	if (client->next) {
		client->next->pre = client->pre;
	}

	delete client;
}

void
Broker::handleClientReq(void)
{
	// Accept the connection
	char clientstr[25];
	struct sockaddr_in client;
	socklen_t addrlen = sizeof(struct sockaddr_in);
	int newfd = accept(_local, (struct sockaddr *)&client, &addrlen);
	if (newfd < 0) {
		LOG(D_BASIC, "Failed to accept() new connection on local command "
			"socket: errno %d (%s)\n", errno, strerror(errno));
		return;
	}
	_sin2strbuf( &client, clientstr );

	char opcode;
	uint32_t cip, dip, ip;
	uint16_t cport, dport, port;
	char active, errcode;
	int received;
	// Receive NEW_SOCK
	received = _CBrecv(newfd, &opcode, &cip, &cport, &active, &ip, &port, &dip, &dport, &errcode);
	if (received <= 0) {
		LOG(D_BASIC, "Failed to read request on local command socket from %s\n",
			clientstr);
		close(newfd);
		return;
	}
	if (opcode != NEW_SOCK && opcode != DATA_QUERY) {
		// TODO print the name of the opcode rather than the value
		LOG(D_BASIC, "Invalid command '%c' received on local command socket "
			"from %s\n", opcode, clientstr);
		close(newfd);
		return;
	}

	if (opcode == DATA_QUERY) {
		if(cip == GCB_DATA_QUERY_FREE_SOCKS) {
			// cip identifies the data that the client is querying
			// for now, all we support is the number of additional NEW_SOCKs
			// we can honor
			// ip will hold our answer
			int num_free = 0;
			Relayer *relayer;
			for (relayer = _relayers; relayer; relayer = relayer->next) {
				num_free += relayer->maxClients - relayer->noClients;
			}
			num_free += (_maxRelayers - _noRelayers) * _relayers->maxClients;

			(void) _CBsend(newfd, DATA_RESPONSE, cip, 0, 0, num_free, 0, 0, 0, 0);

			close(newfd);
			return;
		}
		LOG(D_BASIC, "Invalid DATA_QUERY type '%d' received on local command "
			"socket from %s\n", cip, clientstr);
		(void) _CBsend(newfd, NAK, cip, 0, 0, 0, 0, 0, 0, EINVAL);
		return;
	}

	// 1. find a RelayServer that can serve more clients
	Relayer *relayer;
	for (relayer = _relayers; relayer; relayer = relayer->next) {
		if (relayer->moreClient()) { break; }
	}
	// 2. If not found, create another RelayServer
	if (relayer == NULL) {
		if (_noRelayers > _maxRelayers) {
			LOG(D_BASIC, "No RelayServer can afford more clients, and we reached the maximum # of RelayServers (%d)\n", _maxRelayers);
			close(newfd);
			return;
		}
		relayer = createRelayer();
	}

	// Reply RELAY_SERVER
	(void) _CBsend(newfd, RELAY_SERVER, 0, 0, 0, relayer->mip, relayer->mport, 0, 0, 0);

	close(newfd);
	return;
}

void
Broker::handleConnectionReq()
{
	char opcode; uint32_t cip, dip, ip; uint16_t cport, dport, port; char active, errcode;
	int received;

	// Accept the connection
	char clientstr[25];
	struct sockaddr_in client;
	socklen_t addrlen = sizeof(struct sockaddr_in);
	int newfd = accept(_public, (struct sockaddr *)&client, &addrlen);
	if (newfd < 0) {
		LOG(D_BASIC, "Failed to accept() new connection on public command "
			"socket: errno %d (%s)\n", errno, strerror(errno));
		return;
	}
	_sin2strbuf( &client, clientstr );

	// Receive a request
	received = _CBrecv(newfd, &opcode, &cip, &cport, &active, &ip, &port, &dip, &dport, &errcode);
	if (received <= 0) {
		LOG(D_BASIC, "Failed to read request on public command socket from %s. %s\n",
			clientstr, (received == 0)?"The connection closed early.":"There was an error.");
		close(newfd);
		return;
	}
	if (opcode != CONNECT && opcode != INIT_COMM) {
		LOG(D_BASIC, "Invalid command '%c' received on local command socket "
			"from %s\n", opcode, clientstr);
		close(newfd);
		return;
	}

	// Find the listening client which is registered with dip:dport
	int index = (dip + dport) % BRK_CLNT_HASH_SIZE;
	Client *cptr = _clients[index];
	Client *rcver = NULL;
	int proto = (opcode == CONNECT) ? TCP : UDP;
	while(cptr) {
		if (cptr->cport == dport && cptr->cip == dip &&
			cptr->type == proto && cptr->listener) {
			rcver = cptr;
			break;
		}
		cptr = cptr->next;
	}

	// If no client found, reply PEER_NOT_REGISTERED error
	if (rcver == NULL) {
		_CBsend(newfd, NAK, 0, 0, 0, 0, 0, 0, 0, PEER_NOT_REGISTERED);
		close(newfd);
		return;
	}

    // If this broker is configured to do relay all the time
    if (shouldRelay()) {
        // Let the connector use RelayServer
        char rop = (opcode == CONNECT) ? ACTIVE_CONNECT : ACTIVE_INIT;
        _CBsend(newfd, rop, 0, 0, 0, rcver->cip, rcver->cport, 0, 0, 0);
        close(newfd);
        return;
    }

	// If the client can accept connection, or
	// connector is also a client of this Broker and in the same network as the client,
	// let the connector make direct connection to the client
	if (rcver->active == PASSIVE || cip == _mngIP && _same_net(ip, rcver->ip)) {
		char rop = (opcode == CONNECT) ? ACTIVE_CONNECT : ACTIVE_INIT;
		_CBsend(newfd, rop, 0, 0, 0, rcver->ip, rcver->port, 0, 0, 0);
		close(newfd);
		return;

	// If the listener is ACTIVE and the connector is PASSIVE,
	// we let the listener connect to the connector. In this case
	// we need to notify both parties
	} else if (active == PASSIVE) {
		if (opcode == CONNECT) {
			// For TCP, we need to get ACK from the connector so as not to make the listener
			// try to connect to the connector before it is ready to accept a connection.
			// 1. send PASSIVE_CONNECT to the connector
			if (_CBsend(newfd, PASSIVE_CONNECT, 0, 0, 0, 0, 0, 0, 0, 0) < 0) {
				LOG(D_BASIC, "handleConnectionReq: CONNECT received\n");
				LOG(D_BASIC, "\t\tcan't send PASSIVE_CONNECT to the connector\n");
				close(newfd);
				return;
			}
			// 2. receive ACK from the connector
			char rop;
			received = _CBrecv(newfd, &rop, &cip, &cport, &active, &ip, &port, &dip, &dport, &errcode);
			if (received <= 0) {
				LOG(D_BASIC, "handleConnectionReq: CONNECT received\n");
				LOG(D_BASIC, "\t\tsent PASSIVE_CONNECT to the connector\n");
				LOG(D_BASIC, "\t\t_CBrecv failed\n");
				close(newfd);
				return;
			}
			if (rop != ACK) {
				LOG(D_BASIC, "handleConnectionReq: CONNECT received\n");
				LOG(D_BASIC, "\t\tsent PASSIVE_CONNECT to the connector\n");
				LOG(D_BASIC, "\t\tOther than ACK received\n");
				close(newfd);
				return;
			}
			// Now we can close the connection to the remote client
			close(newfd);
			// Send CONTACT_FWD to the RelayServer
			if (_CBsend(rcver->relayer->wrfd, CONTACT_FWD, rcver->cip,
						rcver->cport, 0, ip, port, 0, 0, 0) < 0) {
				LOG(D_BASIC, "handleConnectionReq: CONNECT received\n");
				LOG(D_BASIC, "\t\tPASSIVE_CONNECT sent to the connector\n");
				LOG(D_BASIC, "\t\tcan't send CONTACT_FWD to the RelayServer\n");
				LOG(D_BASIC, "\t\tdeleting the RelayServer and all clients being served by it...\n");
				deleteRelayer(rcver->relayer->rdfd);
			}
			return;
		} else {	// INIT_COMM
			// 1. Send PASSIVE_INIT to the connector
			if (_CBsend(newfd, PASSIVE_INIT, 0, 0, 0, 0, 0, 0, 0, 0) < 0) {
				LOG(D_BASIC, "handleConnectionReq: INIT_COMM received\n");
				LOG(D_BASIC, "\t\tcan't send PASSIVE_INIT to connector\n");
				close(newfd);
				return;
			}
			// Now we can close the connection to the remote client
			close(newfd);
			// Send TRIGGER_FWD to the RelayServer
			if (_CBsend(rcver->relayer->wrfd, TRIGGER_FWD, cip, cport,
						0, ip, port, rcver->cip, rcver->cport, 0) < 0) {
				LOG(D_BASIC, "handleConnectionReq: INIT_COMM received\n");
				LOG(D_BASIC, "\t\tPASSIVE_INIT sent to the connector\n");
				LOG(D_BASIC, "\t\tcan't send TRIGGER_FWD to the RelayServer\n");
				LOG(D_BASIC, "\t\tdeleting the RelayServer and all clients being served by it...\n");
				deleteRelayer(rcver->relayer->rdfd);
			}
			return;
		}

	// IF both parties want to ACTIVE and not in the same network
	} else {
		// Let the connector use RelayServer
		char rop = (opcode == CONNECT) ? ACTIVE_CONNECT : ACTIVE_INIT;
		_CBsend(newfd, rop, 0, 0, 0, rcver->cip, rcver->cport, 0, 0, 0);
		close(newfd);
		return;
	}
}

void
Broker::handleRelayServerEvent(int fd)
{
	LOG(D_FULL, "handleRelayServerEvent for Relayer<%s>\n", _relayerFDs[fd]->tostr());

	// Receive command from the RelayServer
	char opcode, active; uint32_t cip; uint16_t cport; uint32_t dip; uint16_t dport;
	uint32_t ip; uint16_t port; char errcode;
	int rst = _CBrecv(fd, &opcode, &cip, &cport, &active, &ip, &port, &dip, &dport, &errcode);
	if ( rst <= 0) {
		LOG(D_BASIC, "handleRelayServerEvent: _CBrecv failed for Relayer<%s>\n",
			_relayerFDs[fd]->tostr());
		LOG(D_BASIC, "\t\tdeleting the RelayServer and all clients being served by it\n");
		deleteRelayer(fd);
		return;
	}

	// Serve the command
	Relayer *relayer = _relayerFDs[fd];
	int index, proto;
	Client *cptr, *client;
	switch (opcode) {
		case TCP_CLIENT:
		case UDP_CLIENT:
			// 1. create the client
			if (opcode == TCP_CLIENT) {
				client = new Client(TCP, cip, cport, ACTIVE, ip, port, false, relayer);
			} else {
				client = new Client(UDP, cip, cport, ACTIVE, ip, port, true, relayer);
			}
			if (!client) {
				LOG(D_BASIC, "handleRelayServerEvent: (TCP or UDP_CLIENT)\n");
				LOG(D_BASIC, "\tnew Client failed - errno %d (%s)\n", errno,
					strerror(errno));
				ABORT("Dying");
			}

			// 2. register the client to the relayer and _clients bucket
			relayer->addClient();
			index = (cip + cport) % BRK_CLNT_HASH_SIZE;
			cptr = _clients[index];
			if (cptr) {
				cptr->pre = client;
				client->next = cptr;
			}
			_clients[index] = client;
			LOG(D_FULL, "[Client ADDed] _clients[%d]\n", index);
			LOG(D_FULL, "\t%s\n", client->tostr());

			{
				char ip_broker[24];
				client->ip_broker(ip_broker);
				char ip_client[24];
				client->ip_client(ip_client);
				LOG(D_BASIC, "Listening at %s (%s) on behalf of %s\n",
					ip_broker, (opcode == TCP_CLIENT ? "TCP" : "UDP"),
					ip_client);
			}

			break;
		case LISTEN:
			// Find the client and tag it as a listener
			index = (cip + cport) % BRK_CLNT_HASH_SIZE;
			client = _clients[index];
			while (client) {
				if (client->cip == cip && client->cport == cport && client->type == TCP) {
					client->listener = true;
					return;
				}
				client = client->next;
			}
			LOG(D_BASIC, "handleRelayServerEvent: LISTEN(%s) rcved\n", _ipport2str(cip, cport));
			LOG(D_BASIC, "\tno TCP client with cip:cport found\n");
			ABORT("Dying");
			break;
		case TCP_DEL:
		case UDP_DEL:
			// Find the client and delete it
			proto = (opcode == TCP_DEL) ? TCP : UDP;
			index = (cip + cport) % BRK_CLNT_HASH_SIZE;
			client = _clients[index];
			while (client) {
				if (client->cip == cip && client->cport == cport && client->type == proto) {

					char ip_broker[24];
					client->ip_broker(ip_broker);
					char ip_client[24];
					client->ip_client(ip_client);
					LOG(D_BASIC, "Closing socket at %s (%s) on behalf of %s\n",
						ip_broker, (opcode == TCP_DEL ? "TCP" : "UDP"),
						ip_client);
					deleteClient(client);
					return;
				}
				client = client->next;
			}
			LOG(D_BASIC, "handleRelayServerEvent: (TCP|UDP)_DEL(%s) rcved\n", _ipport2str(cip, cport));
			LOG(D_BASIC, "\tno client with cip:cport found\n");
			break;
		default:
			LOG(D_BASIC, "handleRelayServerEvent: unexpected msg received\n");
	}
}

void
Broker::run()
{
	fd_set rdfds;

	while(1) {
		FD_ZERO(&rdfds);
		FD_SET(_local, &rdfds);
		FD_SET(_public, &rdfds);
		int maxfd = (_local > _public) ? _local + 1 : _public + 1;

		// Add pipe fds connected to RelayServers
		for (int i = 0; i < _GCB_max_no_fds; i++) {
			if (_relayerFDs[i] != NULL) {
				FD_SET(i, &rdfds);
				if (i >= maxfd) {
					maxfd = i + 1;
				}
			}
		}

		int ready = select(maxfd, &rdfds, NULL, NULL, NULL);
		if (ready < 0) {
			if (errno == EINTR) continue;
			LOG(D_BASIC, "run - select: errno %d (%s)\n", errno,
				strerror(errno));
			return;
		}

		if (FD_ISSET(_local, &rdfds)) {
			handleClientReq();
		}

		if (FD_ISSET(_public, &rdfds)) {
			handleConnectionReq();
		}

		Relayer *relayer = _relayers;
		while (relayer) {
			if (FD_ISSET(relayer->rdfd, &rdfds)) {
				handleRelayServerEvent(relayer->rdfd);
			}
			relayer = relayer->next;
		}
	} // end of while(1)
}

void inline
usage (char *msg)
{
	cerr << msg << endl << endl;
	cerr << "Usage: Broker [-l logfile] [-d debug_level] [-r relay_server] -i ipaddr\n";
	cerr << "\tlogfile: log file name\n";
	cerr << "\tdebug_level: 'always', 'network', or 'fulldebug'\n";
    cerr << "\trelay_server: full pathname of GCB relay server\n";
	cerr << "\tipaddr: ip address to which Broker should bind\n";
	return;
}

void
sigchldhandler(int signo)
{
	int status;
	pid_t pid = waitpid(-1, &status, WNOHANG);
	if (pid < 0) {
		ABORT("sigchldhandler: waitpid() failed: errno %d (%s)", errno,
			  strerror(errno));
	}
	if (pid == 0) {
			// waitpid() is just telling us nothing exited, return silently
		return;
	}
	if (WIFSIGNALED(status)) {
		LOG(D_BASIC, "< RelayServer(pid=%ld) died with signal %d >\n",
			(long)pid, WTERMSIG(status));
	} else {
		LOG(D_BASIC, "< RelayServer(pid=%ld) exited with status %d >\n",
			(long)pid, WEXITSTATUS(status));
	}
	return;
}

int
main(int argc, char *argv[])
{
    char ipaddr[30];
	char logfile[PATH_MAX], *namestr, *levelstr, *sizestr;
    char relayer_path[PATH_MAX], *relayerstr;
	int debug_level = D_BASIC;
	long maxsize;
	bool file_given = false;
	bool level_given = false;
	uint32_t mngIP = 0;
    struct in_addr inaddr;
    struct stat statbuf;
	int c;

	/* get command line arguments */
    relayer_path[0] = '\0';
	while ( (c = getopt(argc, argv, "i:l:d:r:")) != -1 ) {
		switch (c) {
			case 'i':
				if (mngIP != 0) {
					usage("ip address defined multiple times");
					return -1;
				}
                snprintf(ipaddr, sizeof(ipaddr), "%s", optarg);
                if (!inet_aton(ipaddr, &inaddr)) {
                    usage("invalid IP address given");
                    return -1;
                }
                mngIP = inaddr.s_addr;
				break;
			case 'l':
				snprintf(logfile, sizeof(logfile), "%s", optarg);
				file_given = true;
				break;
			case 'd':
				if (!strcasecmp(optarg, "fulldebug")) {
					debug_level = D_FULL;
				}
				level_given = true;
				break;
            case 'r':
                if (relayer_path[0] != '\0') {
                    usage("relay_server defined multiple times");
                    return -1;
                }
                snprintf(relayer_path, sizeof(relayer_path), "%s", optarg);
                break;
			default:
				usage("Invalid argument");
				return -1;
		}
	}

		// setup logging ASAP, so we can log errors when running as a
		// daemon (i.e. stdout/err might be visible to anyone)
	if (!file_given) {
		namestr = getenv("GCB_BROKER_LOG");
		if (namestr) {
			snprintf(logfile, PATH_MAX, "%s", namestr );
		} else {
			namestr = getenv("GCB_LOG_DIR");
			if (namestr) {
				snprintf(logfile, PATH_MAX, "%s/GCB_BrokerLog", namestr);
			} else {
				snprintf(logfile, PATH_MAX, "GCB_BrokerLog" );
			}
		}
	}
	if (!level_given) {
		levelstr = getenv("GCB_BROKER_DEBUG");
		if (!levelstr) {
			levelstr = getenv("GCB_DEBUG_LEVEL"); 
		}
		if (levelstr) {
			if (!strcasecmp(levelstr, "fulldebug")) {
				debug_level = D_FULL;
			}
		}
	}

	sizestr = getenv("GCB_BROKER_MAX_LOG");
	if( ! sizestr ) {
		sizestr = getenv("GCB_MAX_LOG");
	}
	if (sizestr) {
		maxsize = atol(sizestr);
	} else { 
		maxsize = 640000;
	}

	Generic_logInit(logfile, debug_level, maxsize);

		// now that logging is working, we can check sanity on a few
		// other thigns, and if there's an error, we'll print to both
		// the log and stderr, just in case...

	if (mngIP == 0) {
		ABORT("ip address must be specified with -i argument");
	}

    if (relayer_path[0] == '\0') {
        relayerstr = getenv("GCB_RELAY_SERVER");
        if (relayerstr) {
            snprintf(relayer_path, PATH_MAX, "%s", relayerstr);
        } else {
            char *mypath = getExecPath();
            if (!mypath) {
				ABORT("Can't figure out the path to RelayServer");
            }
            char *pos = strrchr (mypath, '/');
            if (!pos) {
				ABORT("Strange!! The fullpath does not contain \'/\'");
            }
            *(++pos) = '\0';
            strcpy(relayer_path, mypath);
            free(mypath);
            if (!strcasecmp(&argv[0][strlen(argv[0])-strlen("static")], "static")) {
                strcat(relayer_path, "gcb_relay_server_static");
            } else {
                strcat(relayer_path, "gcb_relay_server");
            }
        }
    }
    if (stat(relayer_path, &statbuf) < 0) {
		ABORT("stat() on relay_server %s failed: errno %d (%s)", relayer_path,
			  errno, strerror(errno));
    }

	// Ignore SIGPIPE
	struct sigaction act;
	act.sa_handler = SIG_IGN;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	if (sigaction(SIGPIPE, &act, NULL) < 0) {
		ABORT("sigaction(SIGPIPE) failed: errno %d (%s)", errno,
			  strerror(errno));
	}

		/*
		  we should also ignore SIGHUP, since there's no config file
		  to re-read, but if we're running under a condor_master, it
		  will send us a SIGHUP if it's told to reconfig its children.
		  admins might try to send a SIGHUP, too.  so, for now, we
		  just ignore it so we don't die in either case.
		*/
	if (sigaction(SIGHUP, &act, NULL) < 0) {
		ABORT("sigaction(SIGHUP) failed: errno %d (%s)", errno, strerror(errno));
	}
	
	// Register SIGCHLD handler so that system may not clog with zombies
	act.sa_handler = sigchldhandler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	if (sigaction(SIGCHLD, &act, NULL) < 0) {
		ABORT("sigaction(SIGCHLD) failed: errno %d (%s)", errno, strerror(errno));
	}

	// Create a Broker and let it run
	Broker *broker = new Broker(ipaddr, relayer_path);
	if (broker == NULL) {
		ABORT("Can't create Broker object!");
	}
	
	broker->run();

	ABORT("Error: Broker exited");
	return -1;
}
