/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#ifndef __BROKER
#define __BROKER

// socket type
#define TCP 203
#define UDP 204

// defaults used if these aren't set in the environment
const int GCB_DEFAULT_MAX_RELAY_SERVERS = 50;
const int GCB_DEFAULT_MAX_CLIENTS = 10;

// This structure represents a relay server that this broker has spawned
class Relayer {
	public:
		Relayer(char *ipaddr, char *path);
		~Relayer();
		char *tostr();
		bool moreClient();
		void addClient();
		void deleteClient();
	private:
        // <ip:port> at which clients can send requests to the relay server
		uint32_t mip;
		uint16_t mport;
        // # of clients that the relay server is currently serving
		int noClients;
        // maximum # of clients that the relay server will serve
		int maxClients;
        // the pipe that the broker can recv reports from the relay server
		int rdfd;
        // the pipe that the broker can send commands to the relay server
		int wrfd;
		pid_t pid;
        // Relayers are linked together to find a relay server that can serve more
        // clients
		Relayer *pre;
		Relayer *next;
	friend class Broker;
};

// This structure represents a socket inside a private network being served by this
// broker
class Client {
	public:
		Client(int type, uint32_t cip, uint16_t cport, char active,
			uint32_t ip, uint16_t port, bool listen, Relayer *relayer);
		char *tostr();
		void ip_broker(char * result); // ASSERT(sizeof(result) >= 24);
		void ip_client(char * result); // ASSERT(sizeof(result) >= 24);
	private:
        // leased (i.e. official) ip:port
		uint32_t cip;
		uint16_t cport;
        // local (private) ip:port
		uint32_t ip;
		uint16_t port;
        // TCP or UDP
		int type;
        // can accept connections: PASSIVE or ACTIVE?
		char active;
        // is it listening?
		bool listener;
        // what relay server is serving this client?
		Relayer *relayer;
        // clients are chained together in a hashed bucket. hash key is cip;cport.
		Client *pre;
		Client *next;
	friend class Broker;
};

#define BRK_CLNT_HASH_SIZE 173 // some prime number

class Broker {
	public:
		Broker(char *ipaddr, char *relay_path);
		void run();

	private:
        // public ip address of the broker. this ip and port will be leased to
        // clients
		char _canonAddr[30];
        // same as _canonAddr but in different format
		uint32_t _mngIP;
        // fd to talk with clients. only the first request will be accepted.
        // subsequent requests must be sent to the relay server that this broker
        // introduced to the client as a response of the first request.
		int _local;
        // fd to talk with clients that want to talk to clients being served by this
        // broker
		int _public;
        // # of relay servers that this broker has spawned
		int _noRelayers;
		// maximum # of relay servers we're willing to spawn
		int _maxRelayers;
        // chain of relay servers
		Relayer *_relayers;
        // relay servers are also indexed by their fds
		Relayer **_relayerFDs;
        // hashed bucket of clients being served
		Client *_clients[BRK_CLNT_HASH_SIZE];
        // full path of the relay server exec
        char *_relayer_path;

		Relayer * createRelayer();
		void deleteRelayer(int fd);
		void deleteClient(Client *client);
		void handleClientReq(void);
		void handleConnectionReq();
		void handleRelayServerEvent(int fd);
};

#endif
