/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include <sys/stat.h>
#include <sys/wait.h>
#include "gcb_common.h"
#include "get_exec_path.h"

static int
daemon_init(void)
{
	pid_t pid;
	int i;

	if ((pid = fork()) < 0) {
		return -1;
	} else if (pid != 0) {
		// parent goes bye-bye
		_exit(0);
	}
	// become session leader
	setsid();
	// let daemons not inherit umask
	umask(0);
	// close all the file descriptors
	// JEF TODO remove bad assumption that fd table size is 1024
	for (i=0; i<1024; i++) {
		close(i);
	}
	return 0;
}

void inline
usage (char *msg)
{
	printf("%s\n\n", msg);
	printf("Usage: Broker_master -i ipaddr\n");
	printf("       [-f master_logfile] [-l broker_logfile] [-d debug_level]\n");
	printf("       [-b broker] [-r relay_server]\n");
	printf("\tipaddr: ip address to which Broker should bind\n");
	printf("\tmaster_logfile: BrokerMaster's log file name\n");
	printf("\tbroker_logfile: Broker's log file name\n");
	printf("\tdebug_level: 'always', 'network', or 'fulldebug'\n");
	printf("\t             Applies to both BrokerMaster's and Broker's log\n");
	printf("\tbroker: full pathname of GCB broker\n");
	printf("\trelay_server: full pathname of GCB relay server\n");
	return;
}

int
main(int argc, char *argv[])
{
	char ipaddr[30], dlevel[30], *namestr, *levelstr, *sizestr, *brokerstr;
	long maxsize = 6400;
	char master_logfile[50], broker_logfile[50];
    char broker_path[100], relayer_path[100];
	int debug_level = D_BASIC;
	int master_file_given = 0;
	int broker_file_given = 0;
	int ip_given = 0;
	int level_given = 0;
	int Broker_argc;
	int exit_status;
	char *Broker_argv[20];
	char arg1[5], arg2[30], arg3[5], arg4[100], arg5[5], arg6[20], arg7[5];
	pid_t pid;
    struct stat statbuf;
	int c;
	char* pos;

	/* get command line arguments */
    broker_path[0] = '\0';
    relayer_path[0] = '\0';
	while ( (c = getopt(argc, argv, "i:l:d:f:b:r:")) != -1 ) {
		switch (c) {
			case 'i':
				if (ip_given) {
					usage("ip address defined multiple times");
					return -1;
				}
				strcpy(ipaddr, optarg);
				ip_given = 1;
				break;
			case 'f':
				if (master_file_given) {
					usage("master_log_file defined multiple times");
					return -1;
				}
				strcpy(master_logfile, optarg);
				master_file_given = 1;
				break;
			case 'l':
				if (broker_file_given) {
					usage("log_file defined multiple times");
					return -1;
				}
				strcpy(broker_logfile, optarg);
				broker_file_given = 1;
				break;
			case 'd':
				if (level_given) {
					usage("debug_level defined multiple times");
					return -1;
				}
				strcpy(dlevel, optarg);
				if (!strcasecmp(dlevel, "fulldebug")) {
					debug_level = D_FULL;
				}
				level_given = 1;
				break;
            case 'b':
                if (broker_path[0] != '\0') {
                    usage("broker defined multiple times");
                    return -1;
                }
                strcpy(broker_path, optarg);
                break;
            case 'r':
                if (relayer_path[0] != '\0') {
                    usage("relay_server defined multiple times");
                    return -1;
                }
                strcpy(relayer_path, optarg);
                break;
			default:
				usage("Invalid argument");
				return -1;
		}
	}

	if (!ip_given) {
		usage ("ip address should be specified");
		return -1;
	}
    if (broker_path[0] == '\0') {
        brokerstr = getenv("Broker");
        if (brokerstr) {
            strcpy(broker_path, brokerstr);
        } else {
            char *mypath = getExecPath();
            if (!mypath) {
                printf("Can't figure out the path to Broker\n");
                return -1;
            }
            pos = strrchr (mypath, '/');
            if (!pos) {
                printf("Strange!! The fullpath does not contain \'/\'\n");
                return -1;
            }
            *(++pos) = '\0';
            strcpy(broker_path, mypath);
            free(mypath);
            if (!strcasecmp(&argv[0][strlen(argv[0])-strlen("static")], "static")) {
                strcat(broker_path, "gcb_broker_static");
            } else {
                strcat(broker_path, "gcb_broker");
            }
        }
    }
    if (stat(broker_path, &statbuf) < 0) {
        printf("stat on %s failed: %s\n", broker_path, strerror(errno));
        return -1;
    }

	// Make this process a daemon
	if (daemon_init() < 0) {
		printf("Could not daemonize the process\n");
		return -1;
	}

	if (!master_file_given) {
		namestr = getenv("BrokerMasterLog");
		if (namestr) {
			strcpy(master_logfile, namestr);
		} else {
			strcpy(master_logfile, "BrokerMasterLog");
		}
	}

	if (!level_given) {
		levelstr = getenv("BrokerMasterDebug");
		if (levelstr) {
			if (!strcasecmp(levelstr, "fulldebug")) {
				debug_level = D_FULL;
			}
		}
	}

	sizestr = getenv("BrokerMasterMaxLog");
	if (sizestr) {
		maxsize = atol(sizestr);
	}

	Generic_logInit(master_logfile, debug_level, maxsize);

	// Now bring the Broker up and do a naive form of what condor_master does
	Broker_argv[0] = broker_path;
	sprintf(arg1, "-i");
	Broker_argv[1] = arg1;
	sprintf(arg2, "%s", ipaddr);
	Broker_argv[2] = arg2;
	Broker_argc = 3;
	if (broker_file_given) {
		sprintf(arg3, "-l");
		Broker_argv[3] = arg3;
		sprintf(arg4, "%s", broker_logfile);
		Broker_argv[4] = arg4;
		Broker_argc = 5;
	}
	if (level_given) {
		sprintf(arg5, "-d");
		Broker_argv[Broker_argc++] = arg5;
		sprintf(arg6, "%s", dlevel);
		Broker_argv[Broker_argc++] = arg6;
	}
    if (relayer_path[0] != '\0') {
        sprintf(arg7, "-r");
        Broker_argv[Broker_argc++] = arg7;
        Broker_argv[Broker_argc++] = relayer_path;
    }
	Broker_argv[Broker_argc] = NULL;
	while(1) {
		pid = fork();
		if (pid < 0) {
			ERROR("forking Broker failed\n");
			return -1;
		} else if (pid == 0) {
			LOG(D_BASIC, "***** Bring a new broker up... \n");
			// Chile process
			execv(broker_path, Broker_argv);
			return -1;
		} else {
			(void) waitpid(pid, &exit_status, 0);
			if ( WIFEXITED( exit_status ) ) {
				LOG(D_BASIC, "***** Broker [pid=%ld] exited with status %d\n",
					(long)pid, WEXITSTATUS( exit_status ) );
			} else {
				LOG(D_BASIC, "***** Broker [pid=%ld] exited with signal %d\n",
					(long)pid, WTERMSIG( exit_status ) );
			}
		}
	}
}
