/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "relay_server.h"
#include <iostream>
using namespace std;
#include <signal.h>

#define FLUSH_INTERVAL 90

int _GCB_max_no_fds;

static int
direct2clients()
{
	char *answer;
	answer = getenv("GCB_ACTIVE_TO_CLIENT");
	return (!answer || strcasecmp(answer, "yes")) ? 0 : 1;
}

Client::Client(int type, int owner, int manager, uint32_t cip,
			   uint16_t cport, uint32_t ip, uint16_t port)
{
	this->type = type;
	this->owner = owner;
	this->manager = manager;
	this->cip = cip;
	this->cport = cport;
	this->ip = ip;
	this->port = port;
	maps = NULL;
	pre = next = NULL;
}

void
Client::rmMaps()
{
	struct Mapping *ptr;
	while(maps) {
		ptr = maps;
		maps = maps->next;
		free(ptr);
	}
}

void
Client::rmMap(int fd)
{
	struct Mapping *ptr = maps, *pre = NULL;
	while(ptr) {
		if (ptr->to == fd) {
			if (pre) {
				pre->next = ptr->next;
			} else {
				maps = ptr->next;
			}
			free(ptr);
			return;
		}
		pre = ptr;
		ptr = ptr->next;
	}
}

Client::~Client()
{
	rmMaps();
}

// @ret: 1, if the mapping found
//       0, if no mapping found
int
Client::getMapping(struct sockaddr_in *from, int *to)
{
	struct Mapping *ptr;
	for (ptr = maps; ptr; ptr = ptr->next) {
		if (!memcmp(&(ptr->from), from, sizeof(struct sockaddr_in))) {
			*to = ptr->to;
			return 1;
		}
	}
	return 0;
} 
int
Client::createMapping(struct sockaddr_in *from, int to)
{
	// add the mapping to the client's information
	Mapping *ptr = (struct Mapping *) malloc(sizeof(struct Mapping));
	if (!ptr) {
		LOG(D_BASIC, "Client::createMapping - malloc: %s\n", strerror(errno));
		return -1;
	}
	memcpy(&(ptr->from), from, sizeof(struct sockaddr_in));
	ptr->to = to;
	ptr->next = maps;
	maps = ptr;

	return 0;
}

char *
Client::tostr()
{
	static char str[300];

	if (type == TCP) {
		sprintf(str, "TCP:");
	} else {
		sprintf(str, "UDP:");
	}
	sprintf(&str[strlen(str)], "owner=%2d:", owner);
	sprintf(&str[strlen(str)], "mnger=%2d:", manager);
	sprintf(&str[strlen(str)], "canon=%s:", _ipport2str(cip, cport));
	sprintf(&str[strlen(str)], "real=%s:", _ipport2str(ip, port));

	struct Mapping *map;
	sprintf(&str[strlen(str)], "Maps:");
	for (map = maps; map; map = map->next) {
		if (strlen(str) > 280) {
			sprintf(&str[strlen(str)], " ...");
			break;
		}
		sprintf(&str[strlen(str)], "%s<->fd[%4d];", _sin2str(&(map->from)), map->to);
	}
	return str;
}

RelayServer::RelayServer(char *ip, char *port)
{
	LOG(D_BASIC, "++++++++++++++++++++++++++++++\n");
	LOG(D_BASIC, "+ RelayServer created (pid: %ld)\n", (long)getpid());
	LOG(D_BASIC, "+ %s\n", _gcb_version());
	LOG(D_BASIC, "+ %s\n", _gcb_build_date());
	LOG(D_BASIC, "+ Listening at %s:%s\n", ip, port);

	_rdfd = 0;
	_wrfd = 1;
	_mfd = 2;
	struct in_addr inaddr;
	if (!inet_aton(ip, &inaddr)) {
		ABORT("inet_aton failed");
	}
	strcpy(_canonAddr, ip);
	_cip = _mngIP = inaddr.s_addr;
	_mngPort = htons(atoi(port));

	_lastFlush = time(NULL);

    _GCB_max_no_fds = getdtablesize();
    _FDs = (FDinfo *) malloc(sizeof(FDinfo) * _GCB_max_no_fds);
    if (!_FDs) { ABORT("malloc failed"); }
	for (int i = 3; i < _GCB_max_no_fds; i++) {
		_FDs[i].type = NOT_DEFINED;
		_FDs[i].fdTo = -1;
		bzero(&(_FDs[i].sinTo), sizeof(struct sockaddr_in));
		_FDs[i].status = NONE;
		_FDs[i].noMsgs = 0;
		_FDs[i].head = _FDs[i].tail = NULL;
		_FDs[i].tcp_bytes = 0;
		_FDs[i].client = NULL;
		_FDs[i].skipped = -1;
        _FDs[i].tcp_buf = NULL;
	}

	for (int i = 0; i < RLY_CLNT_HASH_SIZE; i++) {
		_clients[i] = NULL;
	}
	LOG(D_BASIC, "++++++++++++++++++++++++++++++\n");
}

void
RelayServer::startAddTCPrelay(int sock)
{
	// 1. accept a connection
	struct sockaddr_in from;
	socklen_t addrlen = sizeof(from);
	int fd_a = accept(sock, (struct sockaddr *)&from, &addrlen);
	if (fd_a < 0) {
		LOG(D_BASIC, "startAddTCPrelay - fd_a = accept(sock, ...): %s\n", strerror(errno));
		return;
	}
    // 1.1 make fd_a nonblocking
    int val = fcntl(fd_a, F_GETFL, 0);
    if (val < 0) {
        LOG(D_BASIC, "startAddTCPrelay\n");
        LOG(D_BASIC, "\tcan't fcntl(fd_a,...) failed: %s\n", strerror(errno));
        close(fd_a);
        return;
    }
    if (fcntl(fd_a, F_SETFL, val | O_NONBLOCK) < 0) {
        LOG(D_BASIC, "startAddTCPrelay\n");
        LOG(D_BASIC, "\tsetting fd_a nonblocking...\n");
        LOG(D_BASIC, "\tcan't fcntl(fd_a,...) failed: %s\n", strerror(errno));
        close(fd_a);
        return;
    }

	// 2. make connection to the client
	Client *client = _FDs[sock].client;
	int fd_b = socket(AF_INET, SOCK_STREAM, 0);
	if (fd_b < 0) {
		LOG(D_BASIC, "startAddTCPrelay\n");
		LOG(D_BASIC, "\tTCP connection from %s\n", _sin2str(&from));
		LOG(D_BASIC, "\tcan't create a socket: %s\n", strerror(errno));
		close(fd_a);
		return;
	}
	// 2.1 if ACTIVE_TO_CLIENT is defined, make active connection to the client
	if (direct2clients()) {
		struct sockaddr_in cli_addr;
		// LOG(D_FULL, "\tMaking active connection\n");
		// 2.1.1 make fd_b nonblocking
		val = fcntl(fd_b, F_GETFL, 0);
		if (val < 0) {
			LOG(D_BASIC, "startAddTCPrelay\n");
			LOG(D_BASIC, "\tcan't fcntl(fd_b,...) failed: %s\n", strerror(errno));
			close(fd_a);
			close(fd_b);
			return;
		}
		if (fcntl(fd_b, F_SETFL, val | O_NONBLOCK) < 0) {
			LOG(D_BASIC, "startAddTCPrelay\n");
			LOG(D_BASIC, "\tsetting fd_b nonblocking...\n");
			LOG(D_BASIC, "\tcan't fcntl(fd_b,...) failed: %s\n", strerror(errno));
			close(fd_a);
			close(fd_b);
			return;
		}
		// 2.1.2 connect to client's ip:port
		bzero(&cli_addr, sizeof(cli_addr));
		cli_addr.sin_family = AF_INET;
		cli_addr.sin_addr.s_addr = client->ip;
		cli_addr.sin_port = client->port;
		if (connect(fd_b, (struct sockaddr *)&cli_addr, sizeof(cli_addr))) {
			if (errno == EINPROGRESS) {
				// 2.1.2.1 mark fd_a = HALF_CONNECTED and fd_b = BLOCKED
				_FDs[fd_a].type = TCP;
				_FDs[fd_a].fdTo = fd_b;
				_FDs[fd_a].status = HALF_CONNECTED;
				_FDs[fd_a].client = client;
				_FDs[fd_b].type = TCP;
				_FDs[fd_b].fdTo = fd_a;
				_FDs[fd_b].status = BLOCKED;
				_FDs[fd_b].client = client;
				LOG(D_FULL, "[ADD] relay rule: %d <-> %d(client side):\n", fd_a, fd_b);
				LOG(D_FULL, "\t%s[HALF_CONNECTED] <-> %s[BLOCKED]\n",
						_ipport2str(from.sin_addr.s_addr, from.sin_port), _sin2str(&cli_addr));
				return;
			} else {
				LOG(D_BASIC, "startAddTCPrelay\n");
				LOG(D_BASIC, "\tTCP connection from %s\n", _sin2str(&from));
				LOG(D_BASIC, "\tCan't actively connect to the client [%s]\n", client->tostr());
				close(fd_a);
				close(fd_b);
				return;
			}
		}
		// Here, nonblocking connect succeeds. This happens for local connections at some platforms
		// 2.1.2.1 register the mapping fd_a <-> fd_b
		_FDs[fd_a].type = TCP;
		_FDs[fd_a].fdTo = fd_b;
		_FDs[fd_a].status = CONNECTED;
		_FDs[fd_a].client = client;
		_FDs[fd_b].type = TCP;
		_FDs[fd_b].fdTo = fd_a;
		_FDs[fd_b].status = CONNECTED;
		_FDs[fd_b].client = client;
        // malloc tcp_buffer at fd_a and fd_b
        _FDs[fd_a].tcp_buf = (char *)malloc(TCP_RELAY_BUF_SIZE);
        _FDs[fd_b].tcp_buf = (char *)malloc(TCP_RELAY_BUF_SIZE);
        if (!_FDs[fd_a].tcp_buf || !_FDs[fd_b].tcp_buf) {
            ABORT("malloc failed");
        }
        _FDs[fd_a].tcp_bytes = _FDs[fd_b].tcp_bytes = 0;
		LOG(D_FULL, "[ADD] relay rule: %d <-> %d(client side):\n", fd_a, fd_b);
		LOG(D_FULL, "\t%s[CONNECTED] <-> %s[CONNECTED]\n",
				_ipport2str(from.sin_addr.s_addr, from.sin_port), _sin2str(&cli_addr));
		return;
	} else {
	// 2.2 otherwise, make passive connection to the client
		// 2.2.1 bind fd_b and get fd_b's address
		struct sockaddr_in lo_addr;
		bzero(&lo_addr, sizeof(lo_addr));
		lo_addr.sin_family = AF_INET;
		lo_addr.sin_addr.s_addr = _cip;
		addrlen = sizeof(lo_addr);
		if (bind(fd_b, (struct sockaddr *)&lo_addr, addrlen) < 0) {
			LOG(D_BASIC, "startAddTCPrelay\n");
			LOG(D_BASIC, "\tTCP connection from %s\n", _sin2str(&from));
			LOG(D_BASIC, "\tMaking passive connection...\n");
			LOG(D_BASIC, "\tbinding sock to _cip:ephemeral - %s\n", strerror(errno));
			close(fd_a);
			close(fd_b);
			return;
		}
		if (getsockname(fd_b , (struct sockaddr *)&lo_addr, &addrlen) < 0) {
			LOG(D_BASIC, "startAddTCPrelay\n");
			LOG(D_BASIC, "\tTCP connection from %s\n", _sin2str(&from));
			LOG(D_BASIC, "\tMaking passive connection...\n");
			LOG(D_BASIC, "\tgetsockname: %s\n", strerror(errno));
			close(fd_a);
			close(fd_b);
			return;
		}
		// 2.2.2 make fd_b passive
		listen(fd_b, 5);
		// 2.2.3 send CONTACT(to fd_b) msg to the client
		int rst = _CBsend(client->manager, CONTACT, 0, 0, 0, _cip, lo_addr.sin_port, 0, 0, 0);
		if (rst < 0) {
			LOG(D_BASIC, "startAddTCPrelay\n");
			LOG(D_BASIC, "\tTCP connection from %s\n", _sin2str(&from));
			LOG(D_BASIC, "\tMaking passive connection...\n");
			LOG(D_BASIC, "\tcan't send (CONTACT) to the client [%s]\n", client->tostr());
			LOG(D_BASIC, "\tdeleting the client...\n");
			deleteClient(client);
			return;
		}
		// 2.2.4 mark fd_a = HALF_CONNECTED and fd_b = LISTENING
		_FDs[fd_a].type = TCP;
		_FDs[fd_a].fdTo = fd_b;
		_FDs[fd_a].status = HALF_CONNECTED;
		_FDs[fd_a].client = client;
		_FDs[fd_b].type = TCP;
		_FDs[fd_b].fdTo = fd_a;
		_FDs[fd_b].status = LISTENING;
		_FDs[fd_b].blkedAt = time(NULL);
		_FDs[fd_b].client = client;
		LOG(D_FULL, "[ADD] relay rule: %d <-> %d(client side):\n", fd_a, fd_b);
		LOG(D_FULL, "\t%s[HALF_CONNECTED] <-> %s[LISTENING]\n", _sin2str(&from), _sin2str(&lo_addr));
		return;
	}
}

void
RelayServer::finishActiveAdd(int fd_b)
{
	int fd_a = _FDs[fd_b].fdTo;

	// check if the connection succeed
	int error;
	socklen_t len = sizeof(error);
	if (!getsockopt(fd_b, SOL_SOCKET, SO_ERROR, &error, &len)) {
		if (!error) { // connection succeed
			// 1.2 mark fd_a = CONNECTED and fd_b = CONNECTED
			_FDs[fd_a].status = CONNECTED;
			_FDs[fd_b].status = CONNECTED;
            // malloc tcp_buffer both at fd_a and fd_b
            _FDs[fd_a].tcp_buf = (char *)malloc(TCP_RELAY_BUF_SIZE);
            _FDs[fd_b].tcp_buf = (char *)malloc(TCP_RELAY_BUF_SIZE);
            if (!_FDs[fd_a].tcp_buf || !_FDs[fd_b].tcp_buf) {
                ABORT("malloc failed");
            }
            _FDs[fd_a].tcp_bytes = _FDs[fd_b].tcp_bytes = 0;
			LOG(D_FULL, "[ADD] relay rule: %d <-> %d(client side): becomes CONNECTED\n", fd_a, fd_b);
			return;
		} else {
			LOG(D_BASIC, "finishAddTCPrelay(%d)\n", fd_b);
			LOG(D_BASIC, "\tblocked connection returned failure: %s\n", strerror(error));
		}
	} else {
		LOG(D_BASIC, "finishAddTCPrelay(%d)\n", fd_b);
		LOG(D_BASIC, "\tgetsockopt failed:%s\n", strerror(errno));
	}
	// Here, either getsockopt failed or connection failed
	LOG(D_BASIC, "\t[Deleting] relay rule: %d <-> %d(client side)\n", fd_a, fd_b);
	_FDs[fd_a].type = NOT_DEFINED;
	_FDs[fd_b].type = NOT_DEFINED;
	close(fd_a);
	close(fd_b);
	return;
}

void
RelayServer::finishPassiveAdd(int fd_b)
{
	int fd_a = _FDs[fd_b].fdTo;
	Client *client = _FDs[fd_b].client;

	// Accept connection from client
	struct sockaddr_in cli_addr;
	socklen_t addrlen = sizeof(cli_addr);
	int fd_c = accept(fd_b, (struct sockaddr *)&cli_addr, &addrlen);
	if (fd_c < 0) {
		LOG(D_BASIC, "finishAddTCPrelay(%d)\n", fd_b);
		LOG(D_BASIC, "\taccept: %s\n", strerror(errno));
		LOG(D_BASIC, "\t[Deleting] relay rule: %d <-> %d(client side)\n", fd_a, fd_b);
		_FDs[fd_a].type = NOT_DEFINED;
		_FDs[fd_b].type = NOT_DEFINED;
		close(fd_a);
		close(fd_b);
		return;
	}
	// close the listening socket
	LOG(D_BASIC, "\t[Deleting] relay rule: %d <-> %d(client side)\n", fd_a, fd_b);
	_FDs[fd_b].type = NOT_DEFINED;
	close(fd_b);
    // make fd_c nonblocking
    int val = fcntl(fd_c, F_GETFL, 0);
    if (val < 0) {
        LOG(D_BASIC, "startAddTCPrelay\n");
        LOG(D_BASIC, "\tcan't fcntl(fd_c,...) failed: %s\n", strerror(errno));
		_FDs[fd_a].type = NOT_DEFINED;
		close(fd_a);
        close(fd_c);
        return;
    }
    if (fcntl(fd_c, F_SETFL, val | O_NONBLOCK) < 0) {
        LOG(D_BASIC, "startAddTCPrelay\n");
        LOG(D_BASIC, "\tsetting fd_c nonblocking...\n");
        LOG(D_BASIC, "\tcan't fcntl(fd_c,...) failed: %s\n", strerror(errno));
		_FDs[fd_a].type = NOT_DEFINED;
		close(fd_a);
        close(fd_c);
        return;
    }
	// register the mapping fd_a <-> fd_c
	_FDs[fd_a].fdTo = fd_c;
	_FDs[fd_a].status = CONNECTED;
	_FDs[fd_c].type = TCP;
	_FDs[fd_c].fdTo = fd_a;
	_FDs[fd_c].status = CONNECTED;
	_FDs[fd_c].client = client;
    // malloc tcp_buffer both at fd_a and fd_c
    _FDs[fd_a].tcp_buf = (char *)malloc(TCP_RELAY_BUF_SIZE);
    _FDs[fd_c].tcp_buf = (char *)malloc(TCP_RELAY_BUF_SIZE);
    if (!_FDs[fd_a].tcp_buf || !_FDs[fd_c].tcp_buf) {
        ABORT("malloc failed");
    }
    _FDs[fd_a].tcp_bytes = _FDs[fd_c].tcp_bytes = 0;
	LOG(D_FULL, "[ADD] relay rule: %d <-> %d(client side):\n", fd_a, fd_c);
	LOG(D_FULL, "\t[CONNECTED] <-> %s[CONNECTED]\n", _sin2str(&cli_addr));
	return;
}

// Relay data from TCP socket sock to _FDs[sock].fdTo
void
RelayServer::relayTCPsock(int sock)
{
	int relayfd = _FDs[sock].fdTo;

	int bytes = recv(sock, _FDs[relayfd].tcp_buf, TCP_RELAY_BUF_SIZE, 0);
	if (bytes < 0) {
		LOG(D_BASIC, "Relaying TCP packets %d -> %d...\n", sock, relayfd);
		LOG(D_BASIC, "\trecv(%d): %s\n", sock, strerror(errno));
		goto deleteRelay;
	} else if (bytes > 0) {
		_FDs[relayfd].tcp_bytes = bytes;
		LOG(D_FULL, "RECV from [fd=%d] %d bytes\n", sock, bytes);
		return pushTCPdata(relayfd);
	}
	LOG(D_FULL, "[TCP RELAYING fd[%d] -> fd[%d]]...\n", sock, relayfd);
	LOG(D_FULL, "\tfd[%d] closed. Deleting rule\n", sock);

deleteRelay:
	close(sock);
	close(relayfd);
	// unregister sock and relayfd
	_FDs[sock].type = NOT_DEFINED;
	_FDs[relayfd].type = NOT_DEFINED;
    // free tcp_buffer at sock and relayfd
    free(_FDs[sock].tcp_buf);
    free(_FDs[relayfd].tcp_buf);
	return;
}

// Push TCP data in the buffer
void
RelayServer::pushTCPdata(int fd)
{
	int sent = send(fd, _FDs[fd].tcp_buf, _FDs[fd].tcp_bytes, 0);
	if (sent < 0) {
	   if (errno == EAGAIN) {
		   return;
	   }
       LOG(D_BASIC, "pushTCPdata(%d) send failed: %s\n", fd, strerror(errno));
       int relayfd = _FDs[fd].fdTo;
       close(fd);
       close(relayfd);
       _FDs[fd].type = NOT_DEFINED;
       _FDs[relayfd].type = NOT_DEFINED;
       // free tcp_buffer at fd and relayfd
       free(_FDs[fd].tcp_buf);
       free(_FDs[relayfd].tcp_buf);
       return;
    }

	LOG(D_FULL, "SENT to [fd=%d] %d bytes\n", fd, sent);
	_FDs[fd].tcp_bytes -= sent;
	if (_FDs[fd].tcp_bytes > 0) {
        memmove(_FDs[fd].tcp_buf, &(_FDs[fd].tcp_buf[sent]), _FDs[fd].tcp_bytes);
	}
	return;
}

void
RelayServer::handlePktToClient(int sock)
{
	Client *client = _FDs[sock].client;

	// 1. receive data from the application that wants to talk to the client
	char buf[65536];
	struct sockaddr_in from;
	socklen_t addrlen = sizeof(from);
	int bytes = recvfrom(sock, buf, sizeof(buf), 0, (struct sockaddr *)&from, &addrlen);
	if (bytes < 0) {
		LOG(D_BASIC, "handlePktToClient(%s)\n", client->tostr());
		LOG(D_BASIC, "\trecvfrom(fd=%d,...): %s\n", sock, strerror(errno));
		return;
	}
	LOG(D_FULL, "RECV [from: %s] %d bytes\n", _sin2str(&from), bytes);

	// 2. get the mapping
	int fd, rst;
	struct sockaddr_in phyTo;
	rst = client->getMapping(&from, &fd);
	char str[30];
	if (rst == 0) { // no mapping with "from" found
		// 3. create a new mapping and add a UDP relay rule
		// 3.1 create the mapping
		fd = socket(AF_INET, SOCK_DGRAM, 0);
		if (fd < 0) {
			LOG(D_BASIC, "handlePktToClient(%s)\n", client->tostr());
			LOG(D_BASIC, "\tsocket: %s\n", strerror(errno));
			return;
		}
		if (client->createMapping(&from, fd)) {
			LOG(D_BASIC, "handlePktToClient(%s)\n", client->tostr());
			LOG(D_BASIC, "\tClient::createMapping failed\n");
			close(fd);
			return;
		}

		// 3.2 if ACTIVE_TO_CLIENT is not defined, we need to open a UDP channel
		if (!direct2clients()) {
			// 3.2.1 send TRIGGER to client
			struct sockaddr_in my_addr;
			addrlen = sizeof(my_addr);
			bzero(&my_addr, addrlen);
			my_addr.sin_family = AF_INET;
			my_addr.sin_addr.s_addr = _cip;
			if (bind(fd, (struct sockaddr *) &my_addr, addrlen) < 0) {
				LOG(D_BASIC, "handlePktToClient: bind failed - %s\n", strerror(errno));
				close(fd);
				return;
			}
			if (getsockname(fd , (struct sockaddr *)&my_addr, &addrlen) < 0) {
				LOG(D_BASIC, "Passing a UDP packet from %s\n", _sin2str(&from));
				LOG(D_BASIC, "\tAdding a mapping to %s...\n", _ipport2str(client->cip, client->cport));
				LOG(D_BASIC, "\t\tgetsockname: %s\n", strerror(errno));
				close(fd);
				return;
			}
			rst = _CBsend(client->manager, TRIGGER, my_addr.sin_addr.s_addr, my_addr.sin_port,
						0,  my_addr.sin_addr.s_addr, my_addr.sin_port, 0, 0, 0);
			if (rst < 0) {
				LOG(D_BASIC, "Passing a UDP packet from %s\n", _sin2str(&from));
				LOG(D_BASIC, "\tAdding a mapping to %s...\n", _ipport2str(client->cip, client->cport));
				LOG(D_BASIC, "\t\tsendCmd(TRIGGER) failed\n");
				LOG(D_BASIC, "\t\tdeleting the client...\n");
				deleteClient(client);
				return;
			}

			// 3.2.2 register the mapping fd -> sock:from but with status = BLOCKED
			_FDs[fd].type = UDP;
			_FDs[fd].fdTo = sock;
			memcpy(&(_FDs[fd].sinTo), &from, sizeof(from));
			_FDs[fd].status = BLOCKED;
			_FDs[fd].client = client;
			// 3.2.3 buffer the message
			_FDs[fd].head = (struct pendingMsg *)malloc(sizeof(struct pendingMsg));
			if (_FDs[fd].head == NULL) {
				ABORT("malloc failed");
			}
			_FDs[fd].head->msg = (char *) malloc(bytes);
			if (_FDs[fd].head->msg == NULL) {
				ABORT("malloc failed");
			}
			memcpy(_FDs[fd].head->msg, buf, bytes);
			_FDs[fd].head->size = bytes;
			_FDs[fd].head->next = NULL;
			_FDs[fd].tail = _FDs[fd].head;
			_FDs[fd].noMsgs = 1;
			strcpy(str, _ipport2str(client->cip, client->cport));
			LOG(D_FULL, "[UDP Mapping ADDED] %s <-> fd[%d](client:%s)[BLOCKED]\n", _sin2str(&from), fd, str);
			LOG(D_FULL, "\t1 msg(%d bytes) queued\n", bytes);
			return;
		} else {
			phyTo.sin_family = AF_INET;
			phyTo.sin_addr.s_addr = client->ip;
			phyTo.sin_port = client->port;
		}

		// 3.3 connect fd to the client's phyTo
		if (connect(fd, (struct sockaddr *)&phyTo, sizeof(phyTo)) < 0) {
			LOG(D_BASIC, "Passing a UDP packet from %s\n", _sin2str(&from));
			LOG(D_BASIC, "\tAdding a mapping to %s...\n", _ipport2str(client->cip, client->cport));
			LOG(D_BASIC, "\t\tconnect: %s\n", strerror(errno));
			return;
		}
		
		// 3.4 register the mapping fd -> sock:from
		_FDs[fd].type = UDP;
		_FDs[fd].fdTo = sock;
		memcpy(&(_FDs[fd].sinTo), &from, sizeof(from));
		_FDs[fd].status = CONNECTED;
		_FDs[fd].client = client;
		strcpy(str, _ipport2str(client->cip, client->cport));
		LOG(D_FULL, "[UDP Mapping ADDED] %s <-> fd[%d](client side)\n", _sin2str(&from), fd, str);
	}

	if (_FDs[fd].status == BLOCKED) {
		if (_FDs[fd].noMsgs > MAX_PENDING_MSG) {
			LOG(D_FULL, "\tToo many msgs buffered. The current msg get ignored\n");
			return;
		}
		// buffer the message
		struct pendingMsg *msg;
		msg = (struct pendingMsg *)malloc(sizeof(struct pendingMsg));
		if (msg == NULL) {
			ABORT("malloc failed");
		}
		msg->msg = (char *) malloc(bytes);
		if (msg->msg == NULL) {
			ABORT("malloc failed");
		}
		memcpy(msg->msg, buf, bytes);
		msg->size = bytes;
		_FDs[fd].tail->next = msg;
		msg->next = NULL;
		_FDs[fd].tail = msg;
		_FDs[fd].noMsgs++;
		LOG(D_FULL, "\t%d msg(%d bytes) queued\n", _FDs[fd].noMsgs, bytes);
	} else {
		// 4. Relay the packet
		_FDs[fd].skipped = -1;
		if (send(fd, buf, bytes, 0) < 0) {
			LOG(D_BASIC, "Passing a UDP packet from %s to fd[%d]\n", _sin2str(&from), fd);
			LOG(D_BASIC, "\tsend: %s\n", strerror(errno));
			LOG(D_BASIC, "\tDeleting fd[%d](client side) <-> %s\n", fd, _sin2str(&from));
			client->rmMap(fd);
			close(fd);
			_FDs[fd].type = NOT_DEFINED;
			if (_FDs[fd].head) {
				struct pendingMsg *ptr, *tptr;
				ptr = _FDs[fd].head;
				while(ptr) {
					tptr = ptr;
					ptr = ptr->next;
					free(tptr->msg);
					free(tptr);
				}
				_FDs[fd].head = _FDs[fd].tail = NULL;
				_FDs[fd].noMsgs = 0;
			}
			return;
		}
		LOG(D_FULL, "SENT to [fd=%d] %d bytes\n", fd, bytes);
	}
}

void
RelayServer::handlePktFrClient(int sock)
{
	Client *client = _FDs[sock].client;
	_FDs[sock].skipped = -1;
	// receive data
	char buf[65536];
	int bytes;
	struct sockaddr_in from;
	socklen_t fromlen = sizeof(from);
	if (_FDs[sock].status == CONNECTED) {
		bytes = recv(sock, buf, sizeof(buf), 0);
	} else {
		bytes = recvfrom(sock,  buf, sizeof(buf), 0, (struct sockaddr *)&from, &fromlen);
	}
	if (bytes < 0) {
		LOG(D_BASIC, "Relaying UDP packet from client: fd[%d]\n", sock);
		LOG(D_BASIC, "\treceiving from the client: %s\n", strerror(errno));
		goto deleterelay;
	}

	// unstripe the msg and do either of:
		// reply DUMMY_ACK, if msg = DUMMY. If status = BLOCKED, finish the UDP tunnel setup
		// relay the packet, if msg is data packet
	uint32_t cip, dip, ip;
	uint16_t cport, dport, port;
	char opcode, active, errcode;
	if (!_CB_unstripeMsg(bytes, buf, &opcode, &cip, &cport,
				&active, &ip, &port, &dip, &dport, &errcode)) {
		if (opcode == DUMMY) {
			// check if DUMMY is really from the client and for this socket
			struct sockaddr_in my_addr;
			socklen_t laddr = sizeof(struct sockaddr_in);
			bzero(&my_addr, laddr);
			if (getsockname(sock, (struct sockaddr *) &my_addr, &laddr) < 0) {
				LOG(D_BASIC, "Relaying UDP packet from client: fd[%d]\n", sock);
				ABORT("getsockname() failed");
			}
			if (client->cip != cip || client->cport != cport ||
				my_addr.sin_addr.s_addr != dip || my_addr.sin_port != dport) {
				LOG(D_BASIC, "Relaying UDP packet from client: fd[%d]\n", sock);
				LOG(D_BASIC, "\tcip:cport:dip:dport mismatch\n");
				return;
			}
			// reply DUMMY_ACK
			bytes = _CB_stripeMsg(buf, DUMMY_ACK, dip, dport, 0, 0, 0, cip, cport, 0);
			if (bytes == 0) {
				LOG(D_BASIC, "Relaying UDP packet from client: fd[%d]\n", sock);
				LOG(D_BASIC, "\tDUMMY packet received\n");
				LOG(D_BASIC, "\t_CB_stripMsg(DUMMY_ACK) failed\n");
				return;
			}
			// mark fd connected
			if (_FDs[sock].status == BLOCKED) {
				// connect the socket to the client
				LOG(D_FULL, "connect(%d, %s, ...)\n", sock, _sin2str(&from));
				if (connect(sock, (struct sockaddr *)&from, sizeof(from)) < 0) {
					LOG(D_BASIC, "Relaying UDP packet from client: fd[%d]\n", sock);
					LOG(D_BASIC, "\t\tconnect: %s\n", strerror(errno));
					goto deleterelay;
				}
				_FDs[sock].status = CONNECTED;
				LOG(D_FULL, "[UDP Mapping] ... <-> fd[%d](CONNECTED)\n", sock);
			}
			if (send(sock, buf, bytes, 0) != bytes) {
				LOG(D_BASIC, "Relaying UDP packet from client: fd[%d]\n", sock);
				LOG(D_BASIC, "\tDUMMY packet received\n");
				LOG(D_BASIC, "\tsend(DUMMY_ACK) failed - %s\n", strerror(errno));
				goto deleterelay;
			}
			if (_FDs[sock].head) {
				// send the pending messages to the client
				struct pendingMsg *msg = _FDs[sock].head;
				struct pendingMsg *nmsg;
				while(msg) {
					nmsg = msg->next;
					bytes = send(sock, msg->msg, msg->size, 0);
					if (bytes < 0) {
						LOG(D_BASIC, "Relaying UDP packet from client: fd[%d]\n", sock);
						LOG(D_BASIC, "\tsend: %s\n", strerror(errno));
						goto deleterelay;
					}
					LOG(D_FULL, "SENT to [fd=%d] %d bytes\n", sock, bytes);
					free(msg->msg);
					free(msg);
					msg = nmsg;
				}
				_FDs[sock].head = _FDs[sock].tail = NULL;
				_FDs[sock].noMsgs = 0;
			}
			_FDs[sock].skipped = -1;
			return;
		} else {
			LOG(D_BASIC, "Relaying UDP packet from client: fd[%d]\n", sock);
			LOG(D_BASIC, "\tunexpected CB type msg received: opcode=%d\n", opcode);
			return;
		}
	} else {
		LOG(D_FULL, "RECV [fd=%d] %d bytes\n", sock, bytes);
		int sent = sendto(_FDs[sock].fdTo, buf, bytes, 0,
						(struct sockaddr *) &(_FDs[sock].sinTo),
						sizeof(struct sockaddr_in));
		if (sent < 0) {
			LOG(D_BASIC, "Relaying UDP packet from client: fd[%d]\n", sock);
			LOG(D_BASIC, "\tsendto %s through fd[%d] failed: %s\n",
					_sin2str(&(_FDs[sock].sinTo)), _FDs[sock].fdTo, strerror(errno));
			return;
		}
		LOG(D_FULL, "SENT to %s %d bytes\n", _sin2str(&(_FDs[sock].sinTo)), sent);
		return;
	}

deleterelay:
	LOG(D_BASIC, "Deleting fd[%d](client side) <-> %s\n", sock, _sin2str(&(_FDs[sock].sinTo)));
	client->rmMap(sock);
	close(sock);
	_FDs[sock].type = NOT_DEFINED;
	if (_FDs[sock].head) {
		struct pendingMsg *ptr, *tptr;
		ptr = _FDs[sock].head;
		while(ptr) {
			tptr = ptr;
			ptr = ptr->next;
			free(tptr->msg);
			free(tptr);
		}
		_FDs[sock].head = _FDs[sock].tail = NULL;
		_FDs[sock].noMsgs = 0;
	}
	return;
}

void
RelayServer::flush(void)
{
    LOG(D_FULL, "Entered RelayServer::flush\n");
	for (int i=3; i < _GCB_max_no_fds; i++) {
		if (_FDs[i].type == TCP) {
			if (_FDs[i].status == LISTENING && time(NULL) - _FDs[i].blkedAt > MAX_STAY_BLKED) {
				// passive connection to the client has been timedout
                // We don't need to worry about timeout of active connections because
                // when timeout happens "select" will return write ready and
                // "getsockopt" will report error
				int spouse = _FDs[i].fdTo;
				_FDs[spouse].type = NOT_DEFINED;
				close(spouse);
				_FDs[i].type = NOT_DEFINED;
				close(i);
				LOG(D_BASIC, "\t[flush] relay[%d <-> %d] timedout\n", i, spouse);
			}
		} else if (_FDs[i].type == UDP) {
			// for UDP socket, we don't have to worry about blocked passive connection
			// because if the connection is timeout 'skipped' will become bigger than
			// 1 eventually
			_FDs[i].skipped++;
			LOG(D_FULL, "\t[%d] skipped = %d\n", i, _FDs[i].skipped);
			if (_FDs[i].skipped > 1) {
				_FDs[i].client->rmMap(i);
				_FDs[i].type = NOT_DEFINED;
				if (_FDs[i].head) {
					struct pendingMsg *ptr, *tptr;
					ptr = _FDs[i].head;
					while(ptr) {
						tptr = ptr;
						ptr = ptr->next;
						free(tptr->msg);
						free(tptr);
					}
					_FDs[i].head = _FDs[i].tail = NULL;
					_FDs[i].noMsgs = 0;
				}
				close(i);
			}
		}
	}
}

void
RelayServer::deleteClient(Client *client)
{
	LOG(D_FULL, "Deleting Client: %s\n", client->tostr());
	if (client->type == TCP) {
		// send TCP_DEL to the Broker
		if (_CBsend(_wrfd, TCP_DEL, client->cip, client->cport, 0, 0, 0, 0, 0, 0) < 0) {
			ABORT("deleteClient: _CBsend failed");
		}
		// close CLIENT, MANAGER
		_FDs[client->owner].type = NOT_DEFINED;
		close(client->owner);
		_FDs[client->manager].type = NOT_DEFINED;
		close(client->manager);
		// for fds owned by the client
		for (int i = 3; i < _GCB_max_no_fds; i++) {
			if (_FDs[i].type == TCP && (_FDs[i].client) && _FDs[i].client == client) {
				// unlink to the client
				_FDs[i].client = NULL;
				// unfinished active connection must be handled by doing getsockopt later
				// unfinished passive connection must be handled at flush
                // tear down of established connection will be detected by relaying routine.
			}
		}
	} else {
		// send UDP_DEL to the Broker
		if (_CBsend(_wrfd, UDP_DEL, client->cip, client->cport, 0, 0, 0, 0, 0, 0) < 0) {
			ABORT("deleteClient: _CBsend failed");
		}
		// for UDP client, we close all fds owned by this client
		for (int i = 3; i < _GCB_max_no_fds; i++) {
			if (_FDs[i].type == UDP && (_FDs[i].client) && _FDs[i].client == client) {
				close(i);
				_FDs[i].type = NOT_DEFINED;
				if (_FDs[i].head) {
					struct pendingMsg *ptr, *tptr;
					ptr = _FDs[i].head;
					while(ptr) {
						tptr = ptr;
						ptr = ptr->next;
						free(tptr->msg);
						free(tptr);
					}
					_FDs[i].head = _FDs[i].tail = NULL;
					_FDs[i].noMsgs = 0;
				}
			}
		}
		// close CLIENT, MANAGER
		_FDs[client->owner].type = NOT_DEFINED;
		close(client->owner);
		_FDs[client->manager].type = NOT_DEFINED;
		close(client->manager);
	}

	// delete the client from _clients[]
	if (client->pre) {
		client->pre->next = client->next;
	} else {
		int index = (client->cip + client->cport) % RLY_CLNT_HASH_SIZE;
		_clients[index] = client->next;
	}
	if (client->next) {
		client->next->pre = client->pre;
	}

	delete client;
}

void
RelayServer::handleBrokerEvent()
{
	// receive command from the Broker
	char opcode, active; uint32_t cip; uint16_t cport; uint32_t dip; uint16_t dport;
	uint32_t ip; uint16_t port; char errcode;
	int rst = _CBrecv(_rdfd, &opcode, &cip, &cport, &active, &ip, &port, &dip, &dport, &errcode);
	if ( rst <= 0) {
		ABORT("handleBrokerEvent: failed to receive command from the Broker");
	}

	// for each type of command received
	Client *cptr;
	int index;
	switch (opcode) {
		case CONTACT_FWD:
			// find the client with cip:cport and type = TCP
			index = (cip + cport) % RLY_CLNT_HASH_SIZE;
			cptr = _clients[index];
			while(cptr) {
				if (cptr->cport == cport && cptr->cip == cip && cptr->type == TCP) {
					// forward CONTACT
					rst = _CBsend(cptr->manager, CONTACT, 0, 0, 0, ip, port, 0, 0, 0);
					if (rst < 0) {
						LOG(D_BASIC, "handleBrokerEvent: CONTACT_FWD(%s) received\n", _ipport2str(ip, port));
						LOG(D_BASIC, "\tcan't forward CONTACT\n");
						LOG(D_BASIC, "\tdeleting the client\n"); 
						deleteClient(cptr);
					}
					return;
				}
				cptr = cptr->next;
			}
			LOG(D_BASIC, "handleBrokerEvent: CONTACT_FWD(%s) received\n", _ipport2str(ip, port));
			LOG(D_BASIC, "\tcan't find client with cip:cport=%s\n", _ipport2str(cip, cport));
			return;
		case TRIGGER_FWD:
			// find the client with cip:cport and type = UDP
			index = (dip + dport) % RLY_CLNT_HASH_SIZE;
			cptr = _clients[index];
			while(cptr) {
				if (cptr->cport == dport && cptr->cip == dip && cptr->type == UDP) {
					// forward CONTACT
					rst = _CBsend(cptr->manager, TRIGGER, cip, cport, 0, ip, port, 0, 0, 0);
					if (rst < 0) {
						LOG(D_BASIC, "handleBrokerEvent: TRIGGER_FWD(%s) received\n", _ipport2str(ip, port));
						LOG(D_BASIC, "\tcan't forward TRIGGER\n");
						LOG(D_BASIC, "\tdeleting the client\n"); 
						deleteClient(cptr);
					}
					return;
				}
				cptr = cptr->next;
			}
			LOG(D_BASIC, "handleBrokerEvent: TRIGGER_FWD(%s) received\n", _ipport2str(ip, port));
			LOG(D_BASIC, "\tcan't find client with cip:cport=%s\n", _ipport2str(dip, dport));
			return;
		default:
			LOG(D_BASIC, "handleBrokerEvent: unexpected msg received\n");
            return;
	}
}

void
RelayServer::handleNewClient()
{
	// accept a connection
	struct sockaddr_in from;
	socklen_t addrlen = sizeof(from);
	int manager = accept(_mfd, (struct sockaddr *)&from, &addrlen);
	if (manager < 0) {
		LOG(D_BASIC, "handleNewClient: manager = accept(sock, ...): %s\n", strerror(errno));
		return;
	}

	// receive command from the client
	char opcode, active; uint32_t cip; uint16_t cport; uint32_t dip; uint16_t dport;
	uint32_t ip; uint16_t port; char errcode;
	int rst = _CBrecv(manager, &opcode, &cip, &cport, &active, &ip, &port, &dip, &dport, &errcode);
	if ( rst <= 0) {
		LOG(D_BASIC, "handleNewClient: _CBrecv failed\n");
		close(manager);
		return;
	}

	// the command must be (TCP|UDP)_BIND
	if (opcode != TCP_BIND && opcode != UDP_BIND) {
		LOG(D_BASIC, "handleNewClient: Neither TCP_BIND nor UDP_BIND received\n");
		close(manager);
		return;
	}

	// create a proxy socket of the corresponding type
	int proxy;
	if (opcode == TCP_BIND) {
		proxy = socket(AF_INET, SOCK_STREAM, 0);
	} else {
		proxy = socket(AF_INET, SOCK_DGRAM, 0);
	}
	if (proxy < 0) {
		LOG(D_BASIC, "handleNewClient: socket failed - %s\n", strerror(errno));
		ABORT("Dying");
	}

	// bind the proxy socket to _cip:cport
	struct sockaddr_in c_addr;
	socklen_t laddr = sizeof(struct sockaddr_in);
	bzero(&c_addr, laddr);
	c_addr.sin_family = AF_INET;
	c_addr.sin_addr.s_addr = _cip;
	c_addr.sin_port = cport;
	if (bind(proxy, (struct sockaddr *)&c_addr, laddr) < 0) {
		LOG(D_BASIC, "handleNewClient: bind(proxy) to %s\n", _ipport2str(_cip, cport));
		LOG(D_BASIC, "\tbind - %s\n", strerror(errno));
		// Answer NAK with error code
		(void) _CBsend(manager, NAK, 0, 0, 0, 0, 0, 0, 0, errno);
		close(proxy);
		close(manager);
		return;
	}

	// get cip:cport to reply to the client
	bzero(&c_addr, laddr);
	if (getsockname(proxy, (struct sockaddr *) &c_addr, &laddr) < 0) {
		LOG(D_BASIC, "handleNewClient: getsockname(proxy)- %s\n", strerror(errno));
		// Answer NAK with error code
		(void) _CBsend(manager, NAK, 0, 0, 0, 0, 0, 0, 0, errno);
		close(proxy);
		close(manager);
		return;
	}
	cip = c_addr.sin_addr.s_addr;
	cport = c_addr.sin_port;
	// If the client sends us a 0 for its local port, that means it plans
	// to bind to the same port as the canonical address. Thus, we need
	// to set our copy of the client's local port appropriately.
	if ( port == 0 ) {
		port = cport;
	}

	// create a client
	Client *client;
	if (opcode == TCP_BIND) {
		client = new Client(TCP, proxy, manager, cip, cport, ip, port);
	} else {
		client = new Client(UDP, proxy, manager, cip, cport, ip, port);
	}
	if (!client) {
		ABORT("handleNewClient: new Client failed");
	}

	// Answer BIND_OK
	if (_CBsend(manager, BIND_OK, cip, cport, 0, 0, 0, 0, 0, 0) < 0) {
		LOG(D_BASIC, "handleNewClient: can't response BIND_OK\n");
		close(proxy);
		close(manager);
        delete client;
		return;
	}

	// send (TCP|UDP)_CLIENT to the Broker
	if (opcode == TCP_BIND) {
		if (_CBsend(_wrfd, TCP_CLIENT, cip, cport, active, ip, port, 0, 0, 0) < 0) {
			ABORT("can't send TCP_CLIENT to the Broker");
		}
	} else {
		if (_CBsend(_wrfd, UDP_CLIENT, cip, cport, active, ip, port, 0, 0, 0) < 0) {
			ABORT("can't send UDP_CLIENT to the Broker");
		}
	}

	// Register the client to _FDs and _clients
	_FDs[proxy].type = CLIENT;
	_FDs[proxy].status = (opcode == TCP_BIND) ? BOUND : LISTENING;
	_FDs[proxy].client = client;
	_FDs[manager].type = MANAGER;
	_FDs[manager].status = CONNECTED;
	_FDs[manager].client = client;
	int index = (cip + cport) % RLY_CLNT_HASH_SIZE;
	Client *cptr = _clients[index];
	if (cptr) {
		cptr->pre = client;
		client->next = cptr;
	}
	_clients[index] = client;
	LOG(D_FULL, "Added [Client %s] _FDs[%d] & _clients[%d]\n", client->tostr(), proxy, index);

	return;
}

void
RelayServer::handleClientReq(int manager)
{
	// Receive command from the client
	char opcode, active; uint32_t cip; uint16_t cport; uint32_t dip; uint16_t dport;
	uint32_t ip; uint16_t port; char errcode;
	int rst = _CBrecv(manager, &opcode, &cip, &cport, &active, &ip, &port, &dip, &dport, &errcode);
	if ( rst == 0) {
		LOG(D_FULL, "manager connection to the client closed\n");
		LOG(D_FULL, "\tdeleting the client...\n");
		deleteClient(_FDs[manager].client);
		return;
	} else if (rst < 0) {
		LOG(D_BASIC, "handleClientReq: _CBrecv failed\n");
		LOG(D_BASIC, "\tdeleting the client...\n");
		deleteClient(_FDs[manager].client);
		return;
	}

	switch(opcode) {

		case HEARTBEAT: break; // No action nor reply necessary.

		case LISTEN: handleListen(manager, cip, cport); break;

		case REMOTE_REPORT: handleRemoteReport(manager, cip, cport, active, ip, port, dip, dport, errcode);

		default:
			LOG(D_FULL, "Received unexpected opcode %d from a MANAGER socket. Ignoring.\n", (int)opcode);
			// TODO: Log ip:port of source here.
			break;
	}
}

void 
RelayServer::handleListen(int manager, uint32_t cip, uint16_t cport)
{
	// Make the owner passive and tag it as LISTENING
	int owner = _FDs[manager].client->owner;
	if (listen(owner, 200) < 0) {
		LOG(D_BASIC, "handleClientReq: listen(owner) failed: %s\n", strerror(errno));
		if (_CBsend(manager, NAK, 0, 0, 0, 0, 0, 0, 0, errno) < 0) {
			LOG(D_BASIC, "\tcan't answer NAK\n");
			LOG(D_BASIC, "\tdeleting the client...\n");
			deleteClient(_FDs[manager].client);
		}
		return;
	}

	// Send LISTEN to the Broker
	if (_CBsend(_wrfd, LISTEN, cip, cport, 0, 0, 0, 0, 0, 0) < 0) {
		ABORT("handleClientReq: can't send LISTEN to the Broker");
	}

	// Answer ACK to the client
	if (_CBsend(manager, ACK, 0, 0, 0, 0, 0, 0, 0, 0) < 0) {
		LOG(D_BASIC, "\tcan't answer ACK\n");
		LOG(D_BASIC, "\tdeleting the client...\n");
		deleteClient(_FDs[manager].client);
	}

	// Tag the client as LISTENING
	_FDs[owner].status = LISTENING;

	return;
}


void
RelayServer::handleRemoteReport(int manager, uint32_t cip, uint16_t cport, char logtype, uint32_t ip, uint16_t port, uint32_t dip, uint16_t dport, char errcode)
{
	const char * PREFIX = "Remote report: ";
	Client * client = _FDs[manager].client;
	char * infosrc = "UNKNOWN";
	if(client) { infosrc = client->tostr(); }
	switch(logtype) {
		case REMOTE_REPORT_PASSIVE:
			{
				char ADDRC[22];
				_ipport2strbuf(cip, cport, ADDRC);
				// TODO: Log ip:port of source here.
				LOG(D_BASIC, "%sConnection from %s to %s reversed %s\n",
					PREFIX,
					ADDRC,
					infosrc,
					ip ? "unsuccessfully" : "successfully");
			}
			break;
		default:
			{
				char ADDRC[22], ADDR[22], ADDRD[22];
				_ipport2strbuf(cip, cport, ADDRC);
				_ipport2strbuf(ip, port, ADDR);
				_ipport2strbuf(dip, dport, ADDRD);
				// TODO: Log ip:port of source here.
				// TODO: This should be D_FULL, not BASIC; just here for testing.
				LOG(D_BASIC, "%sReceived unexpected remote report request from %s of type %d."
					"  Data: c %u:%u (%s), "
					"  base %u:%u (%s), "
					"  d %u:%u (%s), "
					"\n",
					PREFIX,
					infosrc,
					(int)logtype,
					(int)cip, (int)cip, ADDRC,
					(int)ip, (int)ip, ADDR,
					(int)dip, (int)dip, ADDRD
				);
			}
			break;
	}
}

void
RelayServer::run()
{
	while(true) {
		time_t current = time(NULL);
		if (current - _lastFlush > FLUSH_INTERVAL) {
			flush();
			_lastFlush = current;
		}

		// initialize rdfds with _rdfd
		fd_set rdfds, wrfds;
		FD_ZERO(&rdfds);
		FD_ZERO(&wrfds);
		int maxfd = 3;
		FD_SET(_rdfd, &rdfds);
		FD_SET(_mfd, &rdfds);
		// add fds of TCP relay, UDP relay socket, and listening clients to rdfds
		for (int i = 3; i < _GCB_max_no_fds; i++) {
			if (// MANAGER
				(_FDs[i].type == MANAGER)
				||
				// LISTENING CLIENT
				(_FDs[i].type == CLIENT && _FDs[i].status == LISTENING)
				||
				// half-connected passive TCP connection
				(_FDs[i].type == TCP && _FDs[i].status == LISTENING)
				||
				// connected TCP w/o pending msg for relaying
				(_FDs[i].type == TCP &&
				 _FDs[i].status == CONNECTED
				 && _FDs[_FDs[i].fdTo].tcp_bytes == 0
				) ||
				// connected or blocked UDP
				(_FDs[i].type == UDP)
			   ) {
				FD_SET(i, &rdfds);
				maxfd = i + 1;
			}
			// connected TCP with pending msg to relay
			if (_FDs[i].type == TCP && _FDs[i].status == CONNECTED && _FDs[i].tcp_bytes) {
				FD_SET(i, &wrfds);
				maxfd = i + 1;
			// half-connected active TCP connection
			} else if (_FDs[i].type == TCP && _FDs[i].status == BLOCKED) {
				FD_SET(i, &wrfds);
				maxfd = i + 1;
			}
		}

		// do select until FLUSH_INTERVAL
		struct timeval tv;
		tv.tv_sec = FLUSH_INTERVAL;
		tv.tv_usec = 0;
		char tfds[50];
		snprintf(tfds, sizeof(tfds), "%s", _prtfds(maxfd,&wrfds));
		LOG(D_FULL, "select(%d, %s, %s, NULL, FLUSH_INTERVAL)...\n", maxfd, _prtfds(maxfd, &rdfds), tfds);
		int ready = select(maxfd, &rdfds, &wrfds, NULL, &tv);
		if (ready < 0) {
			if (errno == EINTR) { continue; }
			LOG(D_BASIC, "run - select: %s\n", strerror(errno));
			return;
		} else if (ready == 0) {
			continue;
		}

		LOG(D_FULL, "\tread ready: %s\n", _prtfds(maxfd, &rdfds));
		LOG(D_FULL, "\twrite ready: %s\n", _prtfds(maxfd, &wrfds));

		if (FD_ISSET(_rdfd, &rdfds)) {
			handleBrokerEvent();
		}

		if (FD_ISSET(_mfd, &rdfds)) {
			handleNewClient();
		}

		for (int i = 3; i < maxfd; i++) {
			if (FD_ISSET(i, &rdfds)) {
				if (_FDs[i].type == MANAGER) {
					handleClientReq(i);
				} else if (_FDs[i].type == CLIENT) {
					Client *client = _FDs[i].client;
					if (client->type == TCP) {
						startAddTCPrelay(i);
					} else {
						handlePktToClient(i);
					}
				} else if (_FDs[i].type == TCP) {
					if (_FDs[i].status == LISTENING) {
						finishPassiveAdd(i);
					} else if (_FDs[i].status == CONNECTED) {
						relayTCPsock(i);
					}
				} else if (_FDs[i].type == UDP) {
					handlePktFrClient(i);
				} else {
					LOG(D_BASIC, "invalid fd[%d] is ready\n", i);
				}
			}
			if (FD_ISSET(i, &wrfds)) {
				if (_FDs[i].status == BLOCKED) {
					finishActiveAdd(i);
				} else { // must be TCP with pending msg for relay
					pushTCPdata(i);
				}
			}
		} // of for
	} // end of while(true)
}

int
main(int argc, char *argv[])
{
	char logfile[PATH_MAX], ipaddr[30], port[10], *namestr, *levelstr, *sizestr;
	int debug_level = D_BASIC;
	long maxsize;

	/* get command line arguments */
	int c;
	while ( (c = getopt(argc, argv, "i:p:")) != -1 ) {
		switch (c) {
			case 'i':
				snprintf(ipaddr, sizeof(ipaddr), "%s", optarg);
				break;
			case 'p':
				snprintf(port, sizeof(port), "%s", optarg);
				break;
			default:
				return -1;
		}
	}

	namestr = getenv("GCB_RELAY_SERVER_LOG");
	if (namestr) {
		snprintf(logfile, PATH_MAX, "%s.%ld", namestr, (long)getpid());
	} else {
		namestr = getenv("GCB_LOG_DIR");
		if (namestr) {
			snprintf(logfile, PATH_MAX, "%s/GCB_RelayServerLog.%ld",
					  namestr, (long)getpid());
		} else {
			snprintf(logfile, PATH_MAX, "GCB_RelayServerLog.%ld",
					 (long)getpid());
		}
	}

	levelstr = getenv("GCB_RELAY_SERVER_DEBUG");
	if (!levelstr) {
		levelstr = getenv("GCB_DEBUG_LEVEL"); 
	}
	if (levelstr) {
		if (!strcasecmp(levelstr, "fulldebug")) {
			debug_level = D_FULL;
		}
	}

	sizestr = getenv("GCB_RELAY_SERVER_MAX_LOG");
	if( ! sizestr ) {
		sizestr = getenv("GCB_MAX_LOG");
	}
	if (sizestr) {
		maxsize = atol(sizestr);
	} else { 
		maxsize = 640000;
	}

	Generic_logInit(logfile, debug_level, maxsize);

	// Ignore SIGPIPE
	struct sigaction act;
	act.sa_handler = SIG_IGN;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	if (sigaction(SIGPIPE, &act, NULL) < 0) {
		ABORT("sigaction(SIGPIPE) failed, errno %d (%s)", errno,
			  strerror(errno));
	}

		/*
		  we should also ignore SIGHUP, since there's no config file
		  to re-read, but if we're running under a condor_master, it
		  will send us a SIGHUP if it's told to reconfig its children.
		  admins might try to send a SIGHUP, too.  so, for now, we
		  just ignore it so we don't die in either case.
		*/
	if (sigaction(SIGHUP, &act, NULL) < 0) {
		ABORT("sigaction(SIGHUP) failed, errno %d (%s)", errno,
			  strerror(errno));
	}
	
	// Create a RelayServer and let it run
	RelayServer *server = new RelayServer(ipaddr, port);
	if (server == NULL) {
		ABORT("Can't create RelayServer");
	}
	server->run();

	ABORT("Error: RelayServer exited");
	return -1;
}
