/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#ifndef __RELAY_SERVER
#define __RELAY_SERVER

// socket type
#define NOT_DEFINED 200
#define CLIENT 201
#define MANAGER 202
#define TCP 203
#define UDP 204

// socket status
#define NONE 123
#define BOUND 124
#define BLOCKED 125
#define LISTENING 126
#define HALF_CONNECTED 127
#define CONNECTED 128

// The following data structure is used to relay UDP packets.
// Packets sent from "from" and received at "owner" socket (see the comments right
// before class Client) will be passed to the client. Relaying in this direction is
// done simply by sending UDP packets from the socket "fd", which is connected to the
// client. How is relaying done in the other direction?
// When a UDP message is received from the client at "fd" (again note that "fd" is
// connected to the client), it will be passed to the remote peer by doing:
// sendto("_FDs[fd].fdTo", ..., "&(_FDs[fd].sinTo", ...) Note that "_FDs[fd].fdTo" is
// set to "owner".
struct Mapping {
	struct sockaddr_in from;
	int to;
	struct Mapping *next;
};

struct pendingMsg {
	int size;
	char *msg;
	struct pendingMsg *next;
};

// This structure represents clients (i.e. server socket inside a private network)
// being served by this relay server.
// When a new client is created, the relay server creates a proxy socket at the
// official address that it rents to the client. "cip:cport" is this official address
// and "owner" is the fd of the proxy socket. A client also has a TCP connection to
// its relay server and send/recv various GCB commands. We call this management
// connection. "manager" is the fd of the management connection.
class Client {
	friend class RelayServer;
	public:
		Client(
			int type,	// TCP or UDP
			int owner,
			int manager,
			uint32_t cip,
			uint16_t cport,
			uint32_t ip,
			uint16_t port);
		// Free all maps and destruct
		~Client();

		// Given "from", find a mapping ("from", "to"). If a mapping is found,
		// return "to" by setting the 2nd arg with the value. If not found, return
		// error and "to" is unchanged
		// @in-args: "from"
		// @out-args: "to"
		// @return:	0, if a mapping ("from", "to") is found
		//			-1, if error occurred
		int getMapping(struct sockaddr_in *from, int *to);

		// Create a mapping ("from", "to") and add it to "maps"
		// @in-args: "from", "to"
		// @return:	0, if succeed
		//			-1, if error occurred
		// Note: the client shall not have the mapping ("from", "to") already
		int createMapping(struct sockaddr_in *from, int to);

		// free the map with the given "to", i.e. the map ("from", "to")
		void rmMap(int fd);

		char *tostr();

	private:
		// Data members
		int type;	// TCP or UDP
		int owner;
		int manager;
		uint32_t cip;
		uint16_t cport;
        // actual private address that this client is bound to
		uint32_t ip;
		uint16_t port;
		struct Mapping *maps;
		Client *pre;
		Client *next;

		// Functions
		// free all maps
		void rmMaps();
};

#define TCP_RELAY_BUF_SIZE 1024
#define MAX_PENDING_MSG 20
class FDinfo {
	friend class RelayServer;
	private:
		int type;	// NOT_DEFINED, TCP, UDP, CLIENT, MANAGER
		int fdTo;	// fd to relay packets
		struct sockaddr_in sinTo;	// for UDP socket, (fdTo, sinTo) pair defines (socket, sinaddr) to which
									// packets from client should be relayed
		int status;	// NONE, BLOCKED, HALF_CONNECTED, CONNECTED, LISTENING, BOUND
		time_t blkedAt; // the time when this fd becomes blocked for passive connection
		// For BLOCKED UDP socket, this buffer holds data that are supposed to be
		// passed through this fd when the connection is complete
		int noMsgs;
		struct pendingMsg *head;
		struct pendingMsg *tail;
		// For CONNECTED TCP socket, this holds data that are remained after nonblocking
		// send. Because RelayServer should not be blocked for a super slow client, we
		// use nonblocking send to relay data and queue whatever remaining to send it
		// later
		int tcp_bytes;
		//char tcp_buf[RELAY_BUF_SIZE];
        char *tcp_buf;
		int skipped;

		Client *client; // when type is either TCP or UDP, this field indicates the owner of the relay rule
};


#define RLY_CLNT_HASH_SIZE 173 // some prime number
#define MAX_STAY_BLKED 50
class RelayServer {
	public:
		RelayServer(char *ipaddr, char *port);
		void run();
	private:
		int _rdfd;			// read side of pipe to communicate with Broker
		int _wrfd;			// write side of pipe to communicate with Broker
		int _mfd;			// fd for local clients to connect and send GCB request
		char _canonAddr[30]; // public ip address
		uint32_t _cip;
        // "_mfd" is listening on "_mngIP:_mngPort"
		uint32_t _mngIP;
		uint16_t _mngPort;
		time_t _lastFlush;	// time when the last flush of mappings was done
        // Clients are indexed both by "owner" and "manager". "_FDs[]" is the array
        // that they are indexed into. Clients are also stored in a hashed bucket,
        // "_clients[]". The hash key is "cip" + "cport".
		FDinfo *_FDs;
		Client *_clients[RLY_CLNT_HASH_SIZE];

		void flush(void);
		void deleteClient(Client *client);
		void handleBrokerEvent();
		void handleNewClient();
		void handleClientReq(int manager);
		void handleListen(int manager, uint32_t cip, uint16_t cport);
		void handleRemoteReport(int manager, uint32_t cip, uint16_t cport, char logtype, uint32_t ip, uint16_t port, uint32_t dip, uint16_t dport, char errcode);
		void startAddTCPrelay(int sock);
		void finishActiveAdd(int fd_b);
		void finishPassiveAdd(int fd_b);
		void pushTCPdata(int fd);
		void relayTCPsock(int sock);
		void handlePktToClient(int sock);
		void handlePktFrClient(int sock);
};

#endif
