/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include <iostream>
#include "GCB.h"
#include "GCB_errno.h"

#define TCP_BUF_SIZE 100000

extern "C" {
char *_sin2str(const struct sockaddr_in *sin);
}

using namespace std;

static bool
sendData(int sock, char *buf, int size)
{
	int total, sent;

	cout << "[sendData] sending " << size << " bytes through fd[" << sock <<"]... ";
	total = 0;
	while(total < size) {
		sent = GCB_send(sock, &buf[total], size-total, 0);
		if (sent < 0) {
			if (errno != EINTR && errno != EAGAIN) {
				cerr << "send failed: " << strerror(errno) << endl;
				return false;
			} else {
				continue;
			}
		}
		cout << sent << ", ";
		total += sent;
	}
	cout << endl;
	return true;
}

static int
recvData(int sock, char *buf, int size)
{
	int total, recvd;

	cout << "[recvData] waiting for " << size << " bytes[fd=" << sock << "]... ";
	fflush(NULL);
	total = 0;
	while(total < size) {
		recvd = GCB_read(sock, &buf[total], size-total);
		if (recvd < 0) {
			if (errno != EINTR && errno != EAGAIN) {
				cerr << "read failed: " << strerror(errno) << endl;
				return -1;
			} else {
				continue;
			}
		} else if (recvd == 0) {
			return 0;
		}
		cout << recvd << ", ";
		fflush(NULL);
		total += recvd;
	}
	cout << endl;
	return total;
}

void echoUDP(int fd)
{
	char buf[1040];
	struct sockaddr_in from;
	socklen_t fromlen = sizeof(from);

	int read = GCB_recvfrom(fd,  buf, sizeof(buf), 0, (struct sockaddr *)&from, &fromlen);
	if (read < 0) {
		cerr << "_CB_recvfrom failed\n";
		return;
	}
	cout << "GCB_recvfrom(" << fd << ", " << _sin2str(&from) << ") = " << read << "\n";
	cout << "\t" << buf << endl;

	int sent = GCB_sendto(fd, buf, read, 0, (struct sockaddr *)&from, fromlen);
	cout << "GCB_sendto(" << fd << ", " << _sin2str(&from) << ") = " << sent << "\n";
	cout << sent << " bytes echoed\n";
	cout << "\t" << buf << endl << endl;
	return;
}

void server(uint32_t ip, uint16_t port)
{
	int tcpsock = GCB_socket(AF_INET, SOCK_STREAM, 0);
	if (tcpsock < 0) {
		perror("tcp socket creation");
		return;
	}
	cout << "Listening TCP socket: " << tcpsock << endl;

	struct sockaddr_in my_addr;
	bzero(&my_addr, sizeof(my_addr));
	my_addr.sin_family = AF_INET;
	my_addr.sin_addr.s_addr = ip;
	my_addr.sin_port = port;
	if(GCB_bind(tcpsock, (struct sockaddr *)&my_addr, sizeof(my_addr)) < 0) {
		cerr << "bind failed: " << strerror(errno) << endl;
        cerr << "GCB_error msg: " << GCB_strerror(_GCB_errno) << endl;
		return;
	}
	if(GCB_listen(tcpsock, 5) < 0) {
		cerr << "listen failed: " << strerror(errno) << endl;
		return;
	}

	int udpsock;
	udpsock = GCB_socket(AF_INET, SOCK_DGRAM, 0);
	if (udpsock < 0) {
		perror("udp socket creation");
		return;
	}
	cout << "UDP socket: " << udpsock << endl;
	if(GCB_bind(udpsock, (struct sockaddr *)&my_addr, sizeof(my_addr)) < 0) {
		cerr << "bind failed: " << strerror(errno) << endl;
		return;
	}

	// Here we just fork and execve to test GCB's ability to inherit GCB sockets
	// To test inheritance of duplicated sockets, we dup2 here
	GCB_dup2(tcpsock, 20);
	GCB_dup2(udpsock, 30);
	pid_t pid = fork();
	if (pid < 0) {
		cerr << "fork error\n";
		exit(-1);
	} else if (pid > 0) { // parent
		exit(0);
	}
	// we do not close tcpsock and udpsock for testing
	char arg0[20];
	char *args[2];
	sprintf(arg0, "./EchoServer");
	args[0] = arg0;
	args[1] = NULL;
	cout << "About to start EchoServer\n";
	GCB_execv("./EchoServer", args);
	cout << "FAILED to start EchoServer\n";
	exit(-1);
}

int nblkConnect(int sock, struct sockaddr_in *serv_addr)
{
	int succ;
	fd_set wrfds;
	int maxfd = 0;

	FD_ZERO(&wrfds);

	succ = GCB_connect(sock, (struct sockaddr *)serv_addr, sizeof(*serv_addr));
	if (succ < 0) {
		if (errno == EINPROGRESS) {
			cout << "connect returned EINPROGRESS\n";
		} else {
			cerr << "connect failed: " << strerror(errno) << endl;
			return -1;
		}
		FD_SET(sock, &wrfds);
		maxfd = sock + 1;
	} else if (succ == 0) {
		return 0;
	}

	int ready = GCB_select(maxfd, NULL, &wrfds, NULL, NULL);
	if (ready <= 0 ) {
		perror("select");
		return -1;
	}
	// do sockopt
	int error;
	socklen_t len = sizeof(error);
	if (succ < 0 && FD_ISSET(sock, &wrfds)) {
		if (GCB_getsockopt(sock, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
			perror("getsockopt");
			return -1;
		}
		// return result
		if (error) {
			cerr << "Sock1 Connection failed\n";
			return -1;
		} else {
			cout << "Sock1 Connection succ\n";
		}
	}

	return 0;
}

void tcpclient(uint32_t serv_ip, uint16_t serv_port)
{
	char ch;
	struct sockaddr_in serv_addr;
	bzero(&serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = serv_ip;
	serv_addr.sin_port = serv_port;

	cout << "Char to send: ";
	cin >> ch;
	char buf[TCP_BUF_SIZE], rcv_buf[TCP_BUF_SIZE];
	for (int i=0; i<TCP_BUF_SIZE; i++) {
		buf[i] = ch;
	}

	while (true) {
		int sock = GCB_socket(AF_INET, SOCK_STREAM, 0);
		int number = rand() % 2;
		if (number == 0) {
			cout << "Non-blocking\n";
			int val;
			while(true) {
				val = fcntl(sock, F_GETFL, 0);
				if (val < 0) {
					if (errno == EINTR || errno == EAGAIN || errno == EINPROGRESS) continue;
					cerr << "fcntl failed: " << strerror(errno) << endl;
					return;
				}
				break;
			}
			while(true) {
				val = fcntl(sock, F_SETFL, val | O_NONBLOCK);
				if (val < 0) {
					if (errno == EINTR || errno == EAGAIN || errno == EINPROGRESS) continue;
					cerr << "fcntl failed: " << strerror(errno) << endl;
					return;
				}
				break;
			}
			if (nblkConnect(sock, &serv_addr) < 0) {
				cerr << "fcntl failed: " << strerror(errno) << endl;
				return;
			}
		} else {
			cout << "Blocking\n";
			if (GCB_connect(sock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
				cerr << "connect failed: " << strerror(errno) << endl;
				return;
			}
		}

		for (int i = 0; i < 100; i++) {
			if (sendData(sock, buf, sizeof(buf)) != true) {
				cerr << "sendData failed: " << strerror(errno) << endl;
				return;
			}
			int read = recvData(sock, rcv_buf, sizeof(rcv_buf));
			if (read < 0) {
				cerr << "recvData failed: " << strerror(errno) << endl;
				return;
			} else if (read == 0) {
				cerr << "Oops! peer socket has been closed\n";
				return;
			}
			for (int i=0; i<TCP_BUF_SIZE; i++) {
				if (buf[i] != rcv_buf[i]) {
					cerr << "compare failed at " << i << "th byte\n";
					return;
				}
			}
		}
		GCB_close(sock);
		cout << "\n\n\n\n\n\n";
		cout << "\n*********************************************************\n";
		cout << "\t\tEchoed correctly for 100 sends\n";
		cout << "*********************************************************\n";
		cout << "\n\n\n\n\n\n";
	}
}

void udpclient(uint32_t serv_ip, uint16_t serv_port)
{
	bool connected = false;
	struct sockaddr_in serv_addr, from;
	bzero(&serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = serv_ip;
	serv_addr.sin_port = serv_port;
	socklen_t addrlen = sizeof(from);

	while(true) {
		int sock;
		sock = GCB_socket(AF_INET, SOCK_DGRAM, 0);
		if (sock < 0) {
			cerr << "Socket creation failed\n";
			return;
		}

		char ch = rand() % 32 + 33;
		if (rand() % 2) {
			if (GCB_connect(sock, (struct sockaddr *)&serv_addr, addrlen) < 0) {
				cerr << "connect to the server failed: " << strerror(errno) << endl;
				return;
			}
			cout << "GCB_connect(" << sock << ", " << _sin2str(&serv_addr) << ")\n";
			connected = true;
		} else {
			connected = false;
		}
		char buf[100], rcv_buf[100];
		buf[99] = '\0';
		int sent, rcved;
		for (int i=0; i<100; i++) {
			for (int j=0; j<99; j++) {
				buf[j] = ch;
			}
			if (connected) {
				sent = GCB_send(sock, buf, sizeof(buf), 0);
				cout << "GCB_send(" << sock << ") = " << sent << "\n";
			} else {
				sent = GCB_sendto(sock, buf, sizeof(buf), 0, (struct sockaddr *)&serv_addr, addrlen);
				cout << "GCB_sendto(" << sock << ", " << _sin2str(&serv_addr) << ") = " << sent << "\n";
			}
			if (sent < 0) {
				cerr << "sendto failed: " << strerror(errno) << endl;
				break;
			}
			cout << sent << " bytes sent\n";
			cout << "sent:\t" << buf << endl;
			fd_set rdfds;
			int maxfd, ready;
			struct timeval timer;
			FD_ZERO(&rdfds);
			FD_SET(sock, &rdfds);
			maxfd = sock + 1;
			timer.tv_sec = 10;
			timer.tv_usec = 0;
			ready = GCB_select(maxfd, &rdfds, NULL, NULL, &timer);
			if (ready <= 0) {
				break;
			}
			if (connected) {
				rcved = GCB_recv(sock,  rcv_buf, sent, 0);
				cout << "GCB_recv(" << sock << ") = " << rcved << "\n";
			} else {
				rcved = GCB_recvfrom(sock,  rcv_buf, sent, 0, (struct sockaddr *)&from, &addrlen);
				cout << "GCB_recvfrom(" << sock << ", " << _sin2str(&from) << ") = " << rcved << "\n";
			}
			if (rcved < 0) {
				cerr << "recv failed: " << strerror(errno) << endl;
				break;
			}
			cout << rcved << " bytes received\n";
			cout << "received:\t" << rcv_buf << endl;
			for (int j=0; j<rcved; j++) {
				if (buf[j] != rcv_buf[j]) {
					cerr << "compare failed\n";
					return;
				}
			}
			cout << "\n";
			sleep(1);
			ch = ++ch % 32 + 33;
		}
		(void)GCB_close(sock);
		cout << endl << endl;
		cout << "Fresh start\n";
		cout << endl << endl;
	}
}

int main(int argc, char *argv[])
{
	int op;
	char buf[50];
	uint32_t ip;
	uint16_t port;

	cout << "1) server\n";
	cout << "2) TCP client\n";
	cout << "3) UDP client\n";
	cout << "Select: ";
	cin >> op;
	cout << "Server ip: ";
	cin >> buf;
	ip = inet_addr(buf);
	cout << "Server port: ";
	cin >> port;
	port = htons(port);
	switch(op) {
		case 1:
			server(ip, port);
			break;
		case 2:
			tcpclient(ip, port);
			break;
		case 3:
			udpclient(ip, port);
			break;
		default:
			return -1;
	}
}
