/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include "gcb_common.h"
#include "GCB.h"

// run as "echo_client_1 server_fqdn port"
int
main(int argc, char *argv[])
{
    if (argc != 3) {
        printf("usage: echo_client_1 server_fqdn port\n");
        return -1;
    }

    // create a socket
    int fd = GCB_socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
        printf("GCB_socket failed: %s\n", strerror(errno));
        return -1;
    }
    printf("socket(fd=%d) created\n", fd);

    // make the socket nonblocking
    int val;
    if ((val = fcntl(fd, F_GETFL, 0)) < 0) {
        LOG(D_BASIC, "_make_nonblock: fcntl failed - %s\n", strerror(errno));
        return -1;
    }
    if (fcntl(fd, F_SETFL, (void *)(val | O_NONBLOCK)) < 0) {
        LOG(D_BASIC, "_make_nonblock: fcntl failed - %s\n", strerror(errno));
        return -1;
    }

    struct sockaddr_in peer;
    peer.sin_family = AF_INET;
    if (inet_aton(argv[1], &peer.sin_addr) < 0) {
        printf("ip address is in invalid form: %s\n", strerror(errno));
        return -1;
    }
    short port = atoi(argv[2]);
    peer.sin_port = htons(port);
    socklen_t addrlen = sizeof(struct sockaddr_in);
    if (GCB_connect(fd, (struct sockaddr *)&peer, addrlen) < 0) {
        if (errno != EINPROGRESS) {
            printf("<tcp_client_2> GCB_connect: failed - %s\n", strerror(errno));
            return -1;
        }
        LOG(D_FULL, "GCB_connect returned EINPROGRESS\n");
        while(1) {
            fd_set wrfds;
            FD_ZERO(&wrfds);
            FD_SET(fd, &wrfds);
            int ready = GCB_select(fd+1, NULL, &wrfds, NULL, NULL);
            if (ready < 0) {
                if (errno == EINTR) continue;
                printf("GCB_select failed: %s\n", strerror(errno));
                return -1;
            }
            LOG(D_FULL, "<tcp_client_2> GCB_select returned write ready\n");
            break;
        }
    }

    int error;
    socklen_t len = sizeof(error);
    if (GCB_getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
        printf("<tcp_client_2> GCB_getsockopt failed: %s\n", strerror(errno));
        return -1;
    }
    if (error) {
        printf("<tcp_client_2> GCB_getsockopt returned: %s\n", strerror(error));
        return -1;
    }
    printf("<tcp_client_2> connected to the server\n");

    while(1) {
        // generate a random printable character
        struct timeval tv;
        gettimeofday(&tv, NULL);
        char ch = tv.tv_usec % 24 + 40;
        char sbuf[5001], rbuf[5001];
        for (int i = 0; i < 5000; i++) {
            sbuf[i] = ch;
        }
        sbuf[5000] = '\0';
        while(1) {
            fd_set wrfds;
            FD_ZERO(&wrfds);
            FD_SET(fd, &wrfds);
            int ready = GCB_select(fd+1, NULL, &wrfds, NULL, NULL);
            if (ready < 0) {
                if (errno == EINTR) continue;
                printf("<tcp_client_2> GCB_select failed: %s\n", strerror(errno));
                return -1;
            }
            break;
        }
        int sent = GCB_send(fd, sbuf, 5000, 0);
        if (sent < 0) {
            printf("<tcp_client_2> GCB_send: %s\n", strerror(errno));
            return -1;
        }
        printf("***%d bytes sent\n", sent);
        int total = 0;
        while(total < sent) {
            while(1) {
                fd_set rdfds;
                FD_ZERO(&rdfds);
                FD_SET(fd, &rdfds);
                int ready = GCB_select(fd+1, &rdfds, NULL, NULL, NULL);
                if (ready < 0) {
                    if (errno == EINTR) continue;
                    printf("GCB_select failed: %s\n", strerror(errno));
                    return -1;
                }
                break;
            }
            int rcved = GCB_recv(fd, &rbuf[total], sent - total, 0);
            if (rcved < 0) {
                printf("GCB_recv failed: %s\n", strerror(errno));
                return -1;
            } else if (rcved == 0) {
                printf("server closed\n");
                return 0;
            }
            total += rcved;
        }
        printf("%d bytes received\n", total);
        for (int i=0; i<total; i++) {
            if (sbuf[i] != rbuf[i]) {
                printf("compare failed\n");
                printf("\n\nsent: %s\n", sbuf);
                printf("\n\nrcved: %s\n", rbuf);
                return -1;
            }
        }
    }
}
