/***************************************************************
 *
 * Copyright (C) 1990-2007, Condor Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 * 
 *    http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ***************************************************************/

#include <stdio.h>
#include <sys/socket.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include "gcb_common.h"
#include "GCB_query.h"

const char * VAR_FREE_SOCKS = "freesockets";

bool do_query(const char * broker_addr_string, const char * variable_str)
{
	int brokerfd;
	struct in_addr brokerinaddr;
	struct sockaddr_in brokeraddr;
	uint32_t variable;
	int received;

	if( ! inet_aton(broker_addr_string, &brokerinaddr) ) {
		fprintf(stderr, "Failed to parse broker's IP address. \"%s\" may not be a valid address.\n", broker_addr_string);
		return false;
	}

	brokeraddr.sin_family = AF_INET;
	brokeraddr.sin_addr = brokerinaddr;
	brokeraddr.sin_port = htons(BROKER_LOCAL_PORT);

	if( strcasecmp(variable_str, VAR_FREE_SOCKS) == 0 ) {
		variable = GCB_DATA_QUERY_FREE_SOCKS;
	} else {
		fprintf(stderr, "Unknown variable \"%s\".\n", variable_str);
		return false;
	}
	
	brokerfd = socket(AF_INET, SOCK_STREAM, 0);
	if(brokerfd < 0) {
		fprintf(stderr, "Unexpected error: failed to create a socket. %s (%d)\n", strerror(errno), errno);
		return false;
	}

	if( connect(brokerfd, (sockaddr*)&brokeraddr, sizeof(struct sockaddr_in)) < 0 ) {
		fprintf(stderr, "Failed to connect to the broker at %s.  %s (%d)\n", broker_addr_string, strerror(errno), errno);
		return false;
	}

	if( _CBsend(brokerfd, DATA_QUERY, variable, 0, 0, 0, 0, 0, 0, 0) < 0 ) {
		fprintf(stderr, "Failed to send query to the broker at %s.  %s (%d)\n", broker_addr_string, strerror(errno), errno);
		return false;
	}

	uint32_t data_result;
	{
		char opcode;
		uint32_t cip;
		uint16_t cport;
		char active;
		uint16_t port;
		uint32_t dip;
		uint16_t dport;
		char errcode;
		received = _CBrecv(brokerfd, &opcode, &cip, &cport, &active, &data_result, &port, &dip, &dport, &errcode);
		if(received <= 0) {
			fprintf(stderr, "Failed to receive response from the broker at %s.  %s (%d)\n", broker_addr_string, strerror(errno), errno);
			return false;
		}
		if (opcode == NAK && errcode == EINVAL) {
			fprintf(stderr, "Broker %s did not understand request.  Requested variable may not be supported by that version of the broker.\n", broker_addr_string); 
			return false;
		}
		if (opcode != DATA_RESPONSE ) {
			fprintf(stderr, "Received unexpected response of %d from the broker at %s.\n", opcode, broker_addr_string);
			return false;
		}
		if( cip != variable ) {
			fprintf(stderr, "Received unexpected response from the broker at %s. Requested variable %d, received %d. \n", broker_addr_string, (int)variable, (int)cip);
			return false;
		}
	}
	printf("%s = %d\n", variable_str, (int)data_result);
	return true;
}

void show_usage(const char * progname) {
	printf("Usage: %s [IP address of broker] [variable]\n"
		"Valid variables are:\n"
		"  freesockets\n"
		, progname);
}

int main(int argc, char ** argv)
{
	if(argc != 3) {
		show_usage(argv[0]);
		exit(1);
	}
	return do_query(argv[1], argv[2]) == false;
}
